//= require lib/AppKit.js
//= require JLV.js
//= require ./DetailsWindow.js
//= require ./WidgetWindow.js
//= require ./MoreFilter.js

Date.prototype.isValid = function () {
    // An invalid date object returns NaN for getTime() and NaN is the only
    // object not strictly equal to itself.
    return this.getTime() === this.getTime();
};

JLV.WidgetDataModel = Backbone.Model.extend({
    dateFields: [],
    parse: function (data) {
        var count = this.dateFields.length,
            i = 0;

        for (; i < count; i++) {
            if (data[this.dateFields[i]]) {
                data[this.dateFields[i]] = new Date(data[this.dateFields[i]]);
                if (!data[this.dateFields[i]].isValid()) {
                    data[this.dateFields[i]] = null;
                }
            }
            else {
                data[this.dateFields[i]] = null;
            }
        }

        if (data.provider) {
            data.taxonomy = data.provider.providerType;
            data.taxonomyCode = data.provider.taxonomyCode;
            data.providerName = data.provider.name;
        }

        if (data.orderingProvider) {
            if (data.orderingProvider.providerType) {
                data.taxonomy = data.orderingProvider.providerType;
            }

            if (data.orderingProvider.taxonomyCode) {
                data.taxonomyCode = data.orderingProvider.taxonomyCode;
            }

            data.providerName = data.orderingProvider.name;
        }

        return data;
    }
});


JLV.DateFilteredDataCollection = Backbone.Collection.extend({
    doFetch: function (callback, error) {
        this.fetch({
            data: {
                startDate: moment(this.startDate).format("M/D/YY"),
                endDate: moment(this.endDate).format("M/D/YY")
            },
            reset: true,
            type: 'POST',
            success: callback,
            error: error
        });
    }
});




/**
 *  Widget base class
 *
 **/

JLV.Widget = AppKit.View.extend({
    classID: 'jlv-widget',
    normCodeSystems: [],
    portal: null,
    options: {
        name: '',
        data: null,
        closable: true,
        settings: true,
        dateFilter: false,
        vlerDomain: false,
        print: false,
        refresh: false,
        status: false,
        clipboard: false,
        plus: false,
        moreWindowClass: null,
        columns: undefined,
        defaultSort: null,
        height: 300,
        startDay: -120,
        endDay: 0,
        gridHeaderHeight: 24,
        groupBy: undefined,
        groupFormatter: undefined,
        groupSorter: undefined
    },
    initialize: function (options) {

        var self = this;

        this.dodAdapterSourceReport = [];
        this.dodAdapterStatusRec = [];

        this.$el.attr('tabindex', 0);


        this._header = $('<div role="heading">').addClass('jlv-widget-header');
        this._title = $('<span>').addClass('jlv-widget-title').appendTo(this._header).attr("unselectable", "on");

        this._tools = $('<div role="presentation">').addClass('jlv-widget-tools').prependTo(this._header);

        this._closeBtn = $('<div title="Close" tabindex="0" class="jlv-widget-toolbutton close">')
            .append('<div class="icon-close">')
            .button()
            .appendTo(this._tools);


        this._settingsBtn = $('<div title="Column Settings" tabindex="0" class="jlv-widget-toolbutton settings">')
            .append('<div class="icon-columns">')
            .button()
            .appendTo(this._tools);

        this._setupColumnSettings();


        this._refreshBtn = $('<div title="Refresh" tabindex="0" class="jlv-widget-toolbutton refresh">')
            .append('<div class="icon-refresh">')
            .button()
            .appendTo(this._tools).on('click', function (evt) {
                if (_.isFunction(self.loadData)) {
                    self.loadData();
                }
            });

        this._printBtn = $('<div title="Print" tabindex="0" class="jlv-widget-toolbutton print">')
            .append('<div class="icon-print">')
            .button()
            .on('click', function () {
                if (_.isFunction(self.doPrint)) {
                    self.doPrint();
                }
            }).appendTo(this._tools);

        this._clipboardBtn = $('<div title="Copy to clipboard" tabindex="0" class="jlv-widget-toolbutton clipboard">')
            .append('<div class="icon-clipboard2">')
            .button()
            .on('click', function () {
                self.doClipboard();
            }).appendTo(this._tools);

        this._interfaceStatusBtn = $('<div title="Interface Status" tabindex="0" class="jlv-widget-toolbutton status">')
            .append('<div class="icon-info">')
            .button()
            .on('click', function () {
                self.showInterfaceStatus();
            }).appendTo(this._tools);

        this._dateFilterBtn = $('<div title="Date Filter" tabindex="0" class="jlv-widget-toolbutton filter">')
            .append('<div class="icon-filter">')
            .button()
            .on('click', function () {
                self.toggleDateFilterPanel();
            }).appendTo(this._tools);


        if (JLV.Session.dodVLER) {
            this._vlerLoadBtn = $('<div title="Force VLER Data Refresh - This can take a while" tabindex="0" class="jlv-widget-toolbutton vler">')
                .append('<div class="icon-refresh-alt">')
                .button()
                .on('click', function () {
                    //self.vlerLoadMessage("Community partner data retrieval in progress...");

                    AppKit.AlertDialog.yesNo('Confirm VLER Refresh',
                        'A forced refresh of community partner data may take a few minutes. Click Yes to continue. Click No to cancel.',
                        function (yes) {
                            if (yes) {
                                JLV.Session.forceRefresh = true;
                                if (_.isFunction(JLV.LoadVlerData)) {
                                    JLV.LoadVlerData();
                                }
                            }
                        }
                    );


                }).appendTo(this._tools);
        }



        this._addBtn = $('<div title="Add" tabindex="0" class="jlv-widget-toolbutton add">')
            .append('<div class="icon-plus-big">')
            .button()
            .on('click', function () {
                if (_.isFunction(self.onAdd)) {
                    self.onAdd();
                }
            }).appendTo(this._tools).hide();


        this._closeBtn.on('click', function (e) {
            e.stopPropagation();
            if (self.portal) {
                self.portal.removeWidget(self);
            }
        });

        this.$el.append(this._header);

        this._body = $('<div role="presentation">').addClass('jlv-widget-body').appendTo(this.$el);
        this._footer = $('<div role="presentation">').addClass('jlv-widget-footer').appendTo(this.$el);

        if (this.options.columns || options.columns) {

            var tablePager = new AppKit.Pager();

            this._datagrid = new AppKit.PagedTableView({
                headerHeight: this.options.gridHeaderHeight,
                fitColumns: true,
                controller: this,
                pager: tablePager
            });

            if (_.isFunction(this.itemClicked)) {

                this._datagrid.on('afterRender.click', function () {

                    this.$el.find('.jlv-clickable-link').on('click', function () {
                        var modelId = $(this).data('model-id'),
                            model = self._datagrid.get('source').get(modelId);
                        self.itemClicked(model, $(this));
                    });
                });
            }


            this._datagrid.$el.addClass('jlv-widget-table').appendTo(this._body);

            this._datagrid.on('onSort', _.bind(function (evt, sortInfo) {

                this._defaultSort = {
                    asc: sortInfo.asc,
                    id: sortInfo.column.id
                };

                if (_.isFunction(this.savePortalUI)) {
                    this.savePortalUI();
                }
            }, this));


            tablePager.$el.appendTo(this._footer);

            tablePager.on('selection', function () {
                self._updatePageInfoDisplay();
            });

            this._tablePageInfo = $('<span role="presentation">')
                .addClass('jlv-table-page-info')
                .appendTo(this._footer)
                .hide();
        }

        this._moreBtn = $('<div title="More" tabindex="0" class="jlv-more-button">')
            .text("More >>")
            .button({text: false})
            .appendTo(this._footer)
            .on('click', function () {

                if (!self._moreWindow) {
                    if (self.get('moreWindowClass')) {
                        self._createMoreWindow();
                    }
                }

                if (self._moreWindow) {

                    self._moreWindow.center();
                    if (self._moreWindow.get('maximized')) {
                        self._moreWindow.set('maximized', false);
                    }
                    self._moreWindow.set('defaultSort', self._defaultSort);

                    self._moreWindow.open();
                    self._moreWindow.focus();
                }
            }).
            on('keydown', function (evt) {
                if (evt.which == $.ui.keyCode.TAB) {
                    evt.preventDefault();
                    //loop tab
                    self.$el.find('.jlv-widget-toolbutton').first().focus();
                }
            })
            .hide();

        this.$el.resizable({
            handles: "s",
            stop: function () {
                if (_.isFunction(self.savePortalUI)) {
                    self.savePortalUI();
                }
            }
        });

        //508 - tab nav and hot keys
        this.$el.on('keydown', function (evt) {

            if (evt.keyCode === $.ui.keyCode.ESCAPE) {
                self.$el.focus();
            }

            //block bubbling keydown events from child elements
            if (!$(evt.target).hasClass('jlv-widget'))
                return;

            if (evt.keyCode === $.ui.keyCode.TAB) {
                if (!evt.shiftKey) {
                    var next = self.$el.nextInDOM('.jlv-widget');
                    evt.preventDefault();
                    if (next.length) {
                        if (!next.is(":visible")) {
                            $('#AddWidgetButton').focus();
                        } else {
                            next.focus();
                        }
                    }
                    else {
                        $('#AddWidgetButton').focus();
                    }
                }
            }
            else if (evt.keyCode === $.ui.keyCode.RIGHT) {

                var pager = self._datagrid.get('pager'),
                    current = pager.get('selectedPage'),
                    nextPage = Math.min(current + 1, pager.get('pageCount'));

                if (nextPage > current)
                    pager.set('selectedPage', nextPage);
            }
            else if (evt.keyCode === $.ui.keyCode.LEFT) {

                var pager = self._datagrid.get('pager'),
                    current = pager.get('selectedPage'),
                    nextPage = Math.max(current - 1, 1);

                if (nextPage < current)
                    pager.set('selectedPage', nextPage);
            }
            else if (evt.keyCode === $.ui.keyCode.ENTER) {
                evt.preventDefault();
                //focus on toolbar buttons
                self.$el.find('.jlv-widget-toolbutton').first().focus();
            }


        });

        if (this.options.data) {
            this.set('_data', this.options.data);
        }

        this._super(options);

        this.set('title', this.get('name'));


        this.$el.addClass(this.get('widgetId'));


        //make it draggable
        this.$el.draggable({
            revert: 'invalid',
            helper: 'clone',
            handle: '.jlv-widget-title',
            start: function (event, ui) {
                //for performance reasons in IE, do not drag along the table

                $(this).hide();

                ui.helper.parent().css('zIndex', 1000);
                ui.helper.css({
                    opacity: 0.88,
                    zIndex: 1000
                });

                ui.helper.find('.jlv-widget-body, .jlv-widget-footer, .jlv-vler-loading').hide();
            },
            stop: function (event, ui) {

                ui.helper.css({
                    opacity: 1,
                    zIndex: 0
                });
                ui.helper.find('.jlv-widget-body, .jlv-widget-footer, .jlv-vler-loading').show();
                ui.helper.parent().css('zIndex', 0);

                var portal = $(this).data('portal');

                if (portal) {
                    portal.portal('removePlaceholder');
                }

                $(this).show();

            },
            drag: function (event, ui) {

                var portal = $(this).data('portal');

                if (portal) {
                    portal.portal('computeDrop', event, ui, false);
                }
            }
        });


    },

    _setupColumnSettings: function () {

        this._settingsBtn.on('click', _.bind(function (evt) {

            evt.stopPropagation();

            if (this._settingsPopover) {
                this._settingsPopover.remove();
                this._settingsPopover = null;
                $(document).off('click.columnSettings, keydown.columnSettings');
                return;
            }

            this._settingsPopover = $('<div>')
                .addClass('ui-tooltip bottom arrow').hide().appendTo('body');


            var content = $('<div>').addClass('content').on('click', function (evt) {
                evt.stopPropagation();
            });

            var columns = this._datagrid.get('columns'),
                count = columns.length,
                i = 0;

            var table = $('<table>'),
                tr = null;
            for (; i < count; i++) {

                if (i % 2 == 0) {

                    if (tr !== null) {
                        table.append(tr);
                    }

                    tr = $('<tr>')
                }

                var col = columns[i];

                var td = $('<td>');

                td.append(
                    $(String.format('<input type="checkbox" value="{0}" {1} />',
                        col.id, (col.isHidden ? '' : 'checked')))
                );

                td.append($(String.format('<span>&nbsp;{0}</span>', col.title)));

                tr.append(td);
            }

            if (tr !== null) {
                table.append(tr);
            }


            content.append(table);

            content.append(
                $('<div class="primary" style="margin-top:14px;;height:25px;width:80px;">Apply</div>')
                    .button().on('click', _.bind(function () {

                        this._settingsPopover.remove();
                        this._settingsPopover = null;

                        var notChecked = content.find('input:checkbox:not(:checked)');

                        var toHide = [];
                        $.each(notChecked, function (idx, el) {

                            var id = $(el).val();
                            toHide.push(id);

                        });

                        this.set('hiddenColumnsMinimized', toHide);

                        this.savePortalUI();

                        $(document).off('click.columnSettings, keydown.columnSettings');
                    }, this))
            );

            this._settingsPopover.append(content);

            this._settingsPopover.css({
                left: (this._settingsBtn.offset().left -
                this._settingsPopover.outerWidth() / 2 + 12),
                top: this._settingsBtn.offset().top + 24
            });

            this._settingsPopover.show();

            var self = this;
            $(document).on('click.columnSettings, keydown.columnSettings', function (evt) {
                if (self._settingsPopover) {
                    self._settingsPopover.remove();
                }
                self._settingsPopover = null;
                $(document).off('click.columnSettings, keydown.columnSettings');
            });

        }, this));


    },

    /**
     * creates the more window
     * @private
     */
    _createMoreWindow: function () {
        var moreWindowClassName = this.get('moreWindowClass');
        if (moreWindowClassName) {
            if (!this._moreWindow) {
                var Class = AppKit.get(moreWindowClassName);
                this._moreWindow = new Class({
                    data: this.get('data'),
                    widget: this,
                    hiddenColumns: this.get('hiddenColumnsExpanded')

                });
            }
            else {
                this._moreWindow.renderData();
            }
        }
    },

    close: function () {

        this.destroy();

        if (_.isFunction(this.savePortalUI)) {
            this.savePortalUI();
        }


        this.$el.trigger('close');

    },


    destroy: function () {
        //remove all notifications if they exist
        JLV.NotificationCenter.removeObserverForNotificationName(this, JLV.VlerDataFetchedNotification);
        JLV.NotificationCenter.removeObserverForNotificationName(this, JLV.VlerDataRefreshNotification);

        if (this._moreWindow) {
            this._moreWindow.destroy();
            this._moreWindow = null;
        }

        if (this._settingsPopover) {
            this._settingsPopover.remove();
            this._settingsPopover = null;
        }

        setTimeout(_.bind(function () {
            this.$el.remove();
        }, this), 0);
    },


    adjustHeight: function(){
        var newHeight = this._header.outerHeight()
        if(this._vlerLoadPanel)
            newHeight += this._vlerLoadPanel.el.offsetHeight

        this._body.css({
            top: newHeight,
            bottom: this._footer.outerHeight()
        });
    },
    /**
     * Update the pages displayed information in the widget footer
     */
    _updatePageInfoDisplay: function () {

        if (this._datagrid) {
            var range = this._datagrid.displayedRange(),
                count = this._datagrid.get('rowCount');


            if (range && count) {
                this._tablePageInfo.html(_.template('Displaying {{start}}-{{end}} of {{total}}')({
                    start: range.start + 1,
                    end: range.end,
                    total: count
                })).show();
            }
            else {
                this._tablePageInfo.hide();
            }

        }

        setTimeout(_.bind(function () {
            this.adjustHeight();
        }, this), 0);

    },
    loadData: function (callback) {

        this.set('title', this.get('name'));
        this.dodAdapterSourceReport = [];
        this.dodAdapterStatusRec = [];
        this._moreBtn.hide();

        if (this._datagrid)
            this._datagrid.showMessage('<div class="jlv-table-data-loading" alt="loading indicator"></div>');

        if (this._moreWindow) {
            if (this._moreWindow._datagrid) {
                this._moreWindow._datagrid.showMessage('<div class="jlv-table-data-loading" alt="loading indicator"></div>');
            }
        }

        JLV.VLERDataAvailable();

    },
    doClipboard: function () {
        if (this._datagrid) {

            var clone = this._datagrid.$el.clone();
            clone.find('.appkit-table-column, .ui-datatable-loading-rows, .dummy-row-1, .dummy-column').remove();
            JLV.common.copyGridPanelHTML(clone.html());

        }
    },
    vlerLoadMessage: function (mesg){
        if (!this._vlerLoadPanel) {
            var data = this.get('data');
            if (data) {

                this._vlerLoadPanel = new JLV.Widget.VlerLoadPanel({
                    widget: this
                });

                this._vlerLoadPanel.$el.insertAfter(this._header);
                this._vlerLoadPanel.$el.hide();
            }
        }

        if (this._vlerLoadPanel) {

            var $loading = this.$el.find('.jlv-vler-loading');

            $loading.html(mesg);
            $loading.removeClass('jlv-vler-loading-img');

            if (this._vlerLoadPanel.get('hidden')) {
                this._vlerLoadPanel.set('hidden', false);
            }

            this.adjustHeight();
        }
    },


    hideVlerLoadMessage: function (){
        if (this._vlerLoadPanel) {


            this._vlerLoadPanel.set('hidden', true);
            this.adjustHeight();
        }
    },


    toggleDateFilterPanel: function () {

        if (!this._dateFilterPanel) {
            var data = this.get('data');
            if (data) {
                if (data.startDate !== undefined && data.endDate !== undefined) {
                    this._dateFilterPanel = new JLV.Widget.DateFilterPanel({
                        widget: this,
                        startDate: data.startDate.toDate(),
                        endDate: data.endDate.toDate()
                    });

                    this._dateFilterPanel.$el.prependTo(this._body);
                    this._dateFilterPanel.$el.hide();
                }
            }
        }

        if (this._dateFilterPanel) {
            if (this._dateFilterPanel.get('hidden')) {
                this._dateFilterPanel.set('hidden', false);
                if (this._datagrid)
                    this._datagrid.$el.css('top', this._datagrid.el.offsetTop + this._dateFilterPanel.el.offsetHeight);

            }
            else {

                if (this._datagrid) {
                    if (this._datagrid.el.offsetTop - this._dateFilterPanel.el.offsetHeight < 0) {
                        this._datagrid.$el.css('top', 0);
                    } else {
                        this._datagrid.$el.css('top', this._datagrid.el.offsetTop - this._dateFilterPanel.el.offsetHeight);
                    }
                }


                this._dateFilterPanel.set('hidden', true);
            }


        }
    },

    showInterfaceStatus: function (aBOOL) {
        if (!this._interfaceStatusWindow) {
            this._interfaceStatusWindow = new AppKit.Dialog({
                title: 'Interface Status ~ ' + this.get('name'),
                width: 600,
                height: 500,
                maximizable: true,
                minimizable: true,
                resizable: true,
                modal: false,
                className: 'jlv-interface-status'
            });

            JLV.Desktop.registerWindow(this._interfaceStatusWindow);
        }

        var sourceStatus = JLV.common.DODADAPTER_STATUS_NO_REPORT_MSG;

        var data = this.get('data');

        if (data.dodAdapterSourceReport.length) {
            sourceStatus = '';

            for (var i = 0; i < data.dodAdapterSourceReport.length; i++) {
                sourceStatus += data.dodAdapterSourceReport[i] + '</br></br>';
            }
        }

        this._interfaceStatusWindow.$el.html('<div class="jlv-dodadapter-report">' + sourceStatus + '</div>');
        if (aBOOL === false) {
            this._interfaceStatusWindow.close();
        }
        else {
            this._interfaceStatusWindow.center();
            this._interfaceStatusWindow.open();

        }
    },
    /**
     * Columns setter/getter
     * @param anArray - an Array of column config objects
     * @returns anArray of column configurations
     */
    columns: function (anArray) {
        if (anArray !== undefined) {

            var colcount = anArray.length,
                i = 0;

            if (this._datagrid)
                this._datagrid.removeAllColumns();

            for (; i < colcount; i++) {

                var colInfo = anArray[i];
                if (this._datagrid)
                    this._datagrid.addColumn(colInfo);
            }

            this._columns = anArray;
        }

        return this._columns;
    },

    hiddenColumnsMinimized: function (anArrayOfColumnIds) {

        if (this._datagrid) {
            if (anArrayOfColumnIds !== undefined) {
                this._datagrid.set('hiddenColumns', anArrayOfColumnIds);
            }

            return this._datagrid.get('hiddenColumns');
        }

        return [];

    },

    hiddenColumnsExpanded: function (anArrayOfColumnIds) {

        if (anArrayOfColumnIds !== undefined) {
            this._hiddenColumnsExpanded = anArrayOfColumnIds;
        }

        return this._hiddenColumnsExpanded;
    },


    /** public getter **/
    data: function () {
        if (this._datagrid)
            return this._datagrid.get('source');
    },
    /**
     * private setter/getter Collection for data shown in the widget's table
     * @param aBBCollection - a Backbone collection
     * @returns a Backbone collection
     */
    _data: function (aBBCollection) {
        if (this._datagrid) {
            if (_.isFunction(aBBCollection)) {
                this._datagrid.set('source', aBBCollection ? new aBBCollection() : null);
            }
            else {
                this._datagrid.set('source', aBBCollection);
            }

            return this._datagrid.get('source');
        }

    },
    /**
     * widget title getter/setter shown in widget header
     * @param aString
     * @returns {*}
     */
    title: function (aString) {
        if (aString) {
            this._title.html(aString);
            this.$el.attr('aria-label', aString);
            this.adjustHeight();
        }

        return this._title.text();
    },
    /**
     * is widget closable getter/setter
     * @param aBOOL
     * @returns {*}
     */
    closeable: function (aBOOL) {
        if (aBOOL === true) {
            this._closeBtn.show();
        }
        else if (aBOOL === false) {
            this._closeBtn.hide();
        }

        return this._closeBtn.is(':visible');

    },


    /**
     * does widget have column settings getter/setter
     * @param aBOOL
     * @returns {*}
     */
    settings: function (aBOOL) {
        if (aBOOL === true) {
            this._settingsBtn.show();
        }
        else if (aBOOL === false) {
            this._settingsBtn.hide();
        }

        return this._settingsBtn.is(':visible');

    },


    /**
     * widget clipboard tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    clipboard: function (aBOOL) {
        if (!JLV.common.OUTSIDE_PROVIDER_FLAG) {
            if (aBOOL === true) {
                this._clipboardBtn.show();
            }
            else if (aBOOL === false) {
                this._clipboardBtn.hide();
            }
        }
        else {
            // Do not show clipboard for outside providers (VACV)
            this._clipboardBtn.hide();
        }

        return this._clipboardBtn.is(':visible');
    },
    /**
     * widget date filter tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    dateFilter: function (aBOOL) {

        if (aBOOL === true) {
            this._dateFilterBtn.show();
        }
        else if (aBOOL === false) {
            this._dateFilterBtn.hide();
        }
        return this._dateFilterBtn.is(':visible');

    },
    /**
     * widget date filter tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    vlerDomain: function (aBOOL) {

        if (this._vlerLoadBtn) {

            if (aBOOL === true) {

                this._vlerDomain = true;
                this._vlerLoadBtn.show();
                //add vler-related notifications
                JLV.NotificationCenter.addObserverForNotification({
                    observer: this,
                    name: JLV.VlerDataFetchedNotification,
                    action: _.bind(function () {
                        this.vlerLoadMessage(JLV.common.DOD_VLER_LOADING_MSG);
                        this.$el.find('.jlv-vler-loading').addClass('jlv-vler-loading-img');
                        if (this._moreWindow) {
                            this._moreWindow.vlerLoadMessage(JLV.common.DOD_VLER_LOADING_MSG);
                            this._moreWindow.$el.find('.jlv-vler-loading').addClass('jlv-vler-loading-img');
                        }
                    }, this)
                });

                JLV.NotificationCenter.addObserverForNotification({
                    observer: this,
                    name: JLV.VlerDataRefreshNotification,
                    action: _.bind(function () {

                        if (this._moreWindow) {
                            this.loadData(_.bind(function () {
                                this._moreWindow.renderData();
                            }, this));
                            this._moreWindow.hideVlerLoadMessage();
                        } else {
                            this.loadData();
                        }

                        this.hideVlerLoadMessage();

                    }, this)
                });


            }
            else if (aBOOL === false) {
                this._vlerDomain = false;
                this._vlerLoadBtn.hide();

            }

            if (!this._vlerDomain) {
                //remove all vler notifications if they exist
                JLV.NotificationCenter.removeObserverForNotificationName(this, JLV.VlerDataFetchedNotification);
                JLV.NotificationCenter.removeObserverForNotificationName(this, JLV.VlerDataRefreshNotification);
            }

            return this._vlerDomain;
        }

        return false;

    },

    defaultSort: function (sort) {
        if (sort !== undefined) {
            this._defaultSort = sort;
        }

        return this._defaultSort;
    },

    /**
     * widget print tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    print: function (aBOOL) {

        if (aBOOL === true) {
            this._printBtn.show();
        }
        else if (aBOOL === false) {
            this._printBtn.hide();
        }

        return this._printBtn.is(':visible');

    },
    /**
     * widget refresh tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    refresh: function (aBOOL) {
        if (aBOOL === true) {
            this._refreshBtn.show();
        }
        else if (aBOOL === false) {
            this._refreshBtn.hide();
        }

        return this._refreshBtn.is(':visible');
    },
    /**
     * widget status tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    status: function (aBOOL) {

        if (aBOOL === true) {
            this._interfaceStatusBtn.show();
        }
        else if (aBOOL === false) {
            this._interfaceStatusBtn.hide();
        }
        return this._interfaceStatusBtn.is(':visible');
    },
    /**
     * widget plus/add tool available setter/getter
     * @param aBOOL
     * @returns {*}
     */
    plus: function (aBOOL) {
        if (aBOOL === true) {
            this._addBtn.show();
        }
        else if (aBOOL === false) {
            this._addBtn.hide();
        }

        return this._addBtn.is(':visible');
    },
    groupBy: function (aString) {
        if (this._datagrid)
            return this._datagrid.set('groupBy', aString);
    },
    groupFormatter: function (aFunction) {
        if (this._datagrid)
            return this._datagrid.set('groupFormatter', aFunction);
    },
    groupSorter: function (aFunction) {
        if (this._datagrid)
            return this._datagrid.set('groupSorter', aFunction);
    },
    /**
     * called when widget is resized
     */
    sizer: function () {
        if (this._datagrid) {
            this._datagrid.sizer();
        }

        this.adjustHeight();
    },
    height: function (aNumber) {
        if (aNumber !== undefined) {
            this.$el.height(aNumber);
        }

        return this.$el.height();
    },
    startDay: function (days) {

        var collection = this.get('data');
        if (days !== undefined) {
            if (collection) {
                collection.startDate = moment(moment().add(days, 'days'), "M/D/YY");
                this._startDay = days;

                if (this._dateFilterPanel) {
                    this._dateFilterPanel.set('startDate', collection.startDate);
                    this._dateFilterPanel.updateDatePickers();
                }
            }
            return days;

        }
        if (collection) {
            if (!collection.startDate) {
                collection.startDate = moment(moment().add(this._startDay, 'days'), "M/D/YY");
            }

            return this._startDay;
        }

    },
    endDay: function (days) {

        var collection = this.get('data');
        if (days !== undefined) {
            if (collection) {
                collection.endDate = moment(moment().add(days, 'days'), "M/D/YY");
                this._endDay = days;

                if (this._dateFilterPanel) {
                    this._dateFilterPanel.set('endDate', collection.endDate);
                    this._dateFilterPanel.updateDatePickers();
                }
            }
            return days;
        }

        if (collection) {

            if (!collection.endDate) {
                collection.endDate = moment(moment().add(this._endDay, 'days'), "M/D/YY");
            }

            return this._endDay;
        }
    },
    datagrid: function () {
        return this._datagrid;
    },
    gridHeaderHeight: function (aNumber) {
        if (this._datagrid) {
            if (_.isNumber(aNumber)) {
                this._datagrid.set('headerHeight', aNumber);
            }

            return this._datagrid.get('headerHeight');
        }
    },
    /**
     * widget id getter/setter
     * @param aString
     * @returns {*}
     */
    widgetId: function (aString) {
        if (aString !== undefined)
            this._widgetId = aString;

        return this._widgetId;
    }
});

$(function () {
    JLV.Widget.DateFilterPanel = AppKit.View.extend({
        className: 'jlv-date-filter-panel',
        template: $('#WidgetDateFilterTemplate').template(),
        widget: null,
        startDate: null,
        endDate: null,
        initialize: function (options) {
            this._super(options);

            var self = this;
            this.subview('.jlv-date-filter-close').on('click', function () {
                if (self.widget) {
                    self.widget.toggleDateFilterPanel();
                }
            }).on('keydown', function (evt) {
                if (evt.which === $.ui.keyCode.SPACE) {
                    if (self.widget) {
                        self.widget.toggleDateFilterPanel();
                        setTimeout(function () {
                            self.widget.$el.find('.filter').focus();
                        }, 0);

                    }
                }
            });

            this.subview('.jlv-date-filter-apply').on('click', function () {
                if (self.widget) {
                    var data = self.widget.get('data');
                    if (data) {

                        data.startDate = moment(self._startDateInput.get('date'));
                        data.endDate = moment(self._endDateInput.get('date'));

                        var startDay = data.startDate.diff(moment(), 'days');
                        if (startDay > 0) {//fix diff rounding error
                            startDay++;
                        }

                        self.widget.startDay(startDay);

                        var endDay = data.endDate.diff(moment(), 'days');
                        if (endDay > 0) {
                            endDay++;
                        }

                        self.widget.endDay(endDay);

                        self.widget.loadData();
                    }
                    if (self.widget.savePortalUI) {
                        self.widget.savePortalUI();
                    }
                }
            });

            this.$el.find('.preset').on('click', function () {
                self.applyPreset($(this).data('value'));

            });

            this._startDateInput = this.subview('.jlv-start-date-input');
            this._endDateInput = this.subview('.jlv-end-date-input');

            this.updateDatePickers();

            if (this._startDateInput) {
                this._startDateInput.on('dateSet', function () {
                    self.startDate = moment(this.get('date'));
                    if (self._startDateInput)
                        self._endDateInput.set('minDate', self.startDate.toDate());
                });

                if (JLV.common.OUTSIDE_PROVIDER_FLAG && JLV.common.ASSIGNED_PATIENT_DATE_FILTER) {
                    self._startDateInput.set('minDate', JLV.common.ASSIGNED_PATIENT_DATE_FILTER.toDate());
                }
            }

            if (this._endDateInput) {
                this._endDateInput.on('dateSet', function () {
                    self.endDate = moment(this.get('date'));
                    if (self._startDateInput)
                        self._startDateInput.set('maxDate', self.endDate.toDate());
                });
            }

        },
        applyPreset: function (value) {
            if (this.widget) {
                var data = this.widget.get('data');
                if (data) {
                    data.endDate = moment();
                    if (value === '1w') {
                        data.startDate = moment().subtract(7, 'days');
                    }
                    else if (value === '2w') {
                        data.startDate = moment().subtract(14, 'days');
                    }
                    else if (value === '1m') {
                        data.startDate = moment().subtract(1, 'months');
                    }
                    else if (value === '3m') {
                        data.startDate = moment().subtract(3, 'months');
                    }
                    else if (value === '6m') {
                        data.startDate = moment().subtract(6, 'months');
                    }
                    else if (value === '1y') {
                        data.startDate = moment().subtract(1, 'years');
                    }
                    else if (value === '2y') {
                        data.startDate = moment().subtract(2, 'years');
                    }
                    else if (value === '3y') {
                        data.startDate = moment().subtract(3, 'years');
                    }


                    this.widget.startDay(data.startDate.diff(moment(),'days'));
                    this.widget.endDay(data.endDate.diff(moment(),'days'));

                    this.startDate = data.startDate;
                    this.endDate = data.endDate;

                    this.updateDatePickers();

                    this.widget.loadData();

                    if (this.widget.savePortalUI) {
                        this.widget.savePortalUI();
                    }
                }
            }
        },
        updateDatePickers: function () {
            if (this._startDateInput) {
                if (this.startDate.toDate) {
                    this._startDateInput.set('date', this.startDate.toDate());
                } else {
                    this._startDateInput.set('date', this.startDate);
                }
            }

            if (this._endDateInput) {
                if (this.endDate.toDate) {
                    this._endDateInput.set('date', this.endDate.toDate());
                } else {
                    this._endDateInput.set('date', this.endDate);
                }
            }
        }
    });
});

$(function () {
    JLV.Widget.VlerLoadPanel = AppKit.View.extend({
        className: 'jlv-vler-loading-panel',
        template: $('#WidgetVlerLoadingTemplate').template(),
        widget: null,
        initialize: function (options) {
            this._super(options);

            var self = this;
            this.subview('.jlv-vler-loading').on('click', function () {
                if (self.widget) {
                    //self.widget.hideVlerLoadMessage();
                }
            });

        }
    });
});
