/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package jlv

import gov.va.med.dao.DaoException
import gov.va.med.foundations.utilities.EncryptUtil
import gov.va.med.foundations.utilities.VistaKernelHash
import gov.va.med.groovy.user.AppUser
import gov.va.med.guibeans.GUIAuditEntry
import gov.va.med.guibeans.GUIOutsideProvider
import gov.va.med.guibeans.GUISite
import gov.va.med.guibeans.GUIUser
import gov.va.med.jmeadows.webservice.IehrUserProfile
import grails.converters.JSON
import groovy.json.JsonBuilder
import org.apache.commons.lang3.StringUtils
import org.codehaus.groovy.grails.web.json.JSONObject

class AuthenticateController extends BaseController {
    def appConfig
    def userDao
    def heartbeatDao
    def tokenUtils

    def community = { CommunityLoginCommand loginCmd ->
        session.setAttribute("acceptedAgreement", true)
        def String validLoginDate = ""
        def String validLoginIP = ""
        def String badLoginDate = ""
        def String badLoginIP = ""

        def csrfToken = UUID.randomUUID().toString()
        session.setAttribute("X-CSRF-Token", csrfToken)

        session.guiOutsideProviderFlag = false

        // Attempt to login outside provider
        if (loginCmd.getUserName() != null && loginCmd.getPassword() != null) {
            String errorText = ""

            try {
                GUIOutsideProvider guiOutsideProvider = userDao.loginOutsideProvider(loginCmd.getUserName(), loginCmd.getPassword().encodeAsMD5())
                if (guiOutsideProvider == null) {
                    errorText = "Username and/or Password is incorrect."
                    log.error(errorText)
                }
                else {
                    String cardId = guiOutsideProvider.getId()
                    String loginAgency = 'OP'

                    IehrUserProfile profile = userDao.getIehrUserProfile(cardId, loginAgency)

                    // Load user profile into session
                    try {
                        // Profile exists in database
                        if ((profile != null) && StringUtils.isNotBlank(profile.getDbUserID())) {
                            if (profile.getFlags() == null) {
                                profile.setFlags("")
                            }
                            if (profile.getLocID() == null) {
                                profile.setLocID("")
                            }
                            if (profile.getCfg() == null) {
                                profile.setCfg("")
                            }
                            if (profile.getEmailAddress() == null) {
                                profile.setEmailAddress("")
                            }
                            if (profile.getSubjectDN() == null) {
                                profile.setSubjectDN("");
                            }
                        }
                        // Profile was not in database
                        else {
                            profile = new IehrUserProfile()
                            profile.setDbUserID("")
                            profile.setSmartCardID(cardId)
                            profile.setSmartCardAgency(loginAgency)
                            profile.setEmailAddress(guiOutsideProvider.getEmail());
                            profile.setSubjectDN("");
                            profile.setFlags("")
                            profile.setLocID("")
                            profile.setCfg("")
                        }

                        session.userProfile = profile
                        session.profileName = guiOutsideProvider.getName()

                        // Create GUISite
                        GUISite site = new GUISite()
                        site.setAgency(loginAgency)
                        site.setDmisId("OP")
                        site.setMoniker("OP")
                        site.setName("Outside Provider")
                        site.setSiteCode("OP")
                        site.setStatus("active")

                        // Create GUIUser
                        GUIUser user = new GUIUser()
                        user.setHostSite(site)
                        user.setAgency(loginAgency)
                        user.setName(StringUtils.upperCase(guiOutsideProvider.getName()))
                        user.setUserIen(guiOutsideProvider.getId())
                        user.setEncryptedAvCode(VistaKernelHash.encrypt(guiOutsideProvider.getUsername() + ";" + guiOutsideProvider.getPassword(), false))
                        user.setEncryptedFederatedUid(EncryptUtil.encrypt("123456789"))

                        AppUser appUser = new AppUser(user)

                        // Save user profile
                        if (userDao.setIehrUserProfile(profile)) {
                            if (StringUtils.isNotBlank(profile.getDbUserID())) {
                                profile = userDao.getIehrUserProfile(profile.getSmartCardID(), profile.getSmartCardAgency())
                            }
                        }

                        appUser.setSecurityKey(tokenUtils.generateSecurityKey())
                        appUser.setUserId(profile.getDbUserID())
                        appUser.setCfg(profile.getCfg())

                        session.appUser = appUser
                        session.userProfile = profile
                        session.outsideProviderFlag = true
                        session.guiOutsideProvider = guiOutsideProvider

                        // Get outside provider data type access flags
                        if ((guiOutsideProvider != null) && (guiOutsideProvider.dataTypeAccess != null)
                                && (guiOutsideProvider.dataTypeAccess.length() > 0)) {
                            def flags = JSON.parse(guiOutsideProvider.dataTypeAccess)
                            if (flags[0] != null) {
                                session.vaDataTypeAccessFlag = flags[0]
                            }
                            if (flags[1] != null) {
                                session.dodDataTypeAccessFlag = flags[1]
                            }
                            if (flags[2] != null) {
                                session.vaVlerDataTypeAccessFlag = flags[2]
                            }
                            if (flags[3] != null) {
                                session.dodVlerDataTypeAccessFlag = flags[3]
                            }
                        }

                        redirect(controller: 'app')
                        return true;
                    }
                    catch (Exception e) {
                        errorText = "System Error: Failed to login."
                        log.error(e.getMessage())
                        return false
                    }
                }
            }
            catch (DaoException e) {
                errorText = "System Error: Failed to login."
                log.error(e.getMessage())
            }

            // If an error was present, render login page with error message
            if (StringUtils.isNotBlank(errorText)) {
                render(view: "community", model: [errorText : errorText, grailsConfig  : grailsApplication.config.grails, "userProfile": session.userProfile,
                                                  "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": heartbeatDao.getServicesStatus(), "csrfToken": csrfToken])
            }
        }
        else {
            // Normal rendering of login page before user enters username and password
            render(view: "community", model: [grailsConfig  : grailsApplication.config.grails, "userProfile": session.userProfile,
                                              "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": heartbeatDao.getServicesStatus(), "csrfToken": csrfToken])
        }
    }


    def login = { LoginCommand loginCmd ->

        //populate stuff
        def IehrUserProfile userProfile
        def List<GUISite> sites
        def List<GUIAuditEntry> loginEntries
        def String validLoginDate = ""
        def String validLoginIP = ""
        def String badLoginDate = ""
        def String badLoginIP = ""
        def String cfg = ""
        def String theme = "default"
        def String dodVLER = "false"
        def String loginAgency = ""
        def String loginSite = ""
        def jsonCfg = null
        def String agreement = "false"
        def String clientIP = request.remoteAddr;
        def subjectDN = ""
        def certificate = ""

        //generate and set csrf token
        def csrfToken = UUID.randomUUID().toString()
        session.setAttribute("X-CSRF-Token", csrfToken)

        // Service Statuses
        def serviceStatus = heartbeatDao.getServicesStatus();

        if (session.subjectDN != null) {
            subjectDN = session.subjectDN
        }
        if (session.certificate != null) {
            certificate = session.certificate
        }

        if (session.userProfile != null) {
            userProfile = session.userProfile
            //prep the profile config items for profile window
            cfg = userProfile.cfg
            if (cfg != null && cfg.length() > 0) {
                jsonCfg = JSON.parse(cfg)
                for (item in jsonCfg) {
                    if (item.cfgItem == "UI_CONF") {
                        for (option in item.options) {
                            if (option.theme) {
                                theme = option.theme
                            } else if (option.loginAgency) {
                                loginAgency = option.loginAgency
                            } else if (option.loginSite) {
                                loginSite = option.loginSite
                            } else if (option.dodVLER) {
                                dodVLER = option.dodVLER
                            }

                        }
                    }
                }
            }
            if (loginAgency == "" && userProfile.getSmartCardAgency() != "")
                loginAgency = userProfile.getSmartCardAgency()
            if (loginSite == "" && userProfile.getLocID() != null)
                loginSite = userProfile.getLocID()
        }

        if (session.sites == null) {
            try {
                sites = userDao.getSites()
                if (sites != null) {
                    def tempSites
                    Iterator<GUISite> i = sites.iterator()
                    while (i.hasNext()) {
                        GUISite site = i.next()
                        if (site.endpoints == null || site.endpoints.size() == 0 || site.status != "active") {
                            i.remove()
                        }
                    }
                    session.sites = sites
                }
            }
            catch (Exception e) {
                log.error("AuthenticateController::login - Unable to retrieve site list", e)
                String errorText = "There's an error retrieving the sites list.<br>Please notify the system administrator";
                render(view: "login", model: [errorText: errorText, grailsConfig: grailsApplication.config.grails, "userProfile": userProfile, "theme": theme, "loginAgency": loginAgency, "loginSite": loginSite, "serviceStatus": serviceStatus, "csrfToken": csrfToken])
                return
            }
        } else {
            sites = session.sites
        }

        //DEBUG data sent to login page
        log.debug(new JsonBuilder(userProfile).toPrettyString())
        log.debug(new JsonBuilder(sites).toPrettyString())

        if (request.method == "POST") {

            if (!loginCmd.hasErrors()) {
                String errorText = "Failed to Login - Error contacting the server.<br>Please notify the system administrator";

                loginAgency = loginCmd.loginAgency
                loginSite = loginCmd.siteCode
                agreement = loginCmd.acceptedAgreement



                if (session.userProfile != null) {
                    userProfile = session.userProfile;
                    //userProfile.setSmartCardAgency(loginCmd.loginAgency);
                    userProfile.setLocID(loginCmd.siteCode);

                    //update userCfg with loginAgency and loginSiteCode
                    if (loginCmd.userCfg != null && loginCmd.userCfg.length() > 0) {
                        boolean foundUIConf = false
                        boolean foundLoginAgency = false
                        boolean foundLoginSite = false

                        jsonCfg = JSON.parse(loginCmd.userCfg)

                        for (item in jsonCfg) {
                            if (item.cfgItem == "UI_CONF") {
                                foundUIConf = true
                                for (option in item.options) {
                                    if (option.theme) {
                                        theme = option.theme
                                    }
                                    if (option.loginAgency) {
                                        option.loginAgency = loginAgency
                                        foundLoginAgency = true
                                    } else if (option.loginSite) {
                                        option.loginSite = loginSite
                                        foundLoginSite = true
                                    }
                                }
                                if (!foundLoginAgency)
                                    item.options.add(new JSONObject("{loginAgency:" + loginAgency + "}"))
                                if (!foundLoginSite)
                                    item.options.add(new JSONObject("{loginSite:" + loginSite + "}"))
                            }
                        }
                        if (!foundUIConf) {
                            jsonCfg.add(new JSONObject("{cfgItem:UI_CONF,options:[{dodVLER : " + dodVLER + "},{theme:" + theme + "},{loginSite:" + loginSite + "},{loginAgency:" + loginAgency + "}]}"))
                        }
                        userProfile.setCfg(jsonCfg.toString())
                    }

                    if (userProfile.getFlags() == null) {
                        userProfile.setFlags("");
                    }
                }

                try {
                    def user = userDao.login(userProfile.smartCardID, loginCmd.loginAgency, loginCmd.siteCode, loginCmd.accessCode, loginCmd.verifyCode, clientIP, subjectDN, userProfile.emailAddress)

                    //check if there was a problem logging in user
                    if (user == null ||
                            user.getUserIen() == null ||
                            user.getUserIen().isEmpty() ||
                            user.getUserIen().contains("^")) {
                        if (user != null && user.getUserIen() != null) {
                            String userIen = user.getUserIen();

                            if (userIen.contains("^")) {
                                errorText = userIen.substring(userIen.indexOf("^") + 1)
                            }
                        }

                        try {
                            loginEntries = userDao.getLoginInfo(userProfile.smartCardID, userProfile.emailAddress);
                            if (loginEntries != null) {
                                Iterator<GUIAuditEntry> i = loginEntries.iterator()
                                while (i.hasNext()) {
                                    GUIAuditEntry loginEntry = i.next()
                                    if (loginEntry.userID.startsWith("^")) {
                                        badLoginDate = loginEntry.entryDate;
                                        badLoginIP = loginEntry.ipAddress;
                                    } else {
                                        validLoginDate = loginEntry.entryDate;
                                        validLoginIP = loginEntry.ipAddress;
                                    }
                                }
                            }
                        }
                        catch (Exception e) {
                            log.error("AuthenticateController::login - Unable to retrieve previous login information", e)
                        }

                        render(view: "login", model: [errorText: errorText, grailsConfig: grailsApplication.config.grails, "dodVLER": dodVLER, "sites": sites, "userProfile": userProfile, "theme": theme, "loginAgency": loginAgency, "loginSite": loginSite, "agreement": agreement,
                                                      "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": serviceStatus, "csrfToken": csrfToken])
                        return
                    } else {
                        AppUser appUser = new AppUser(user)

                        //save profile
                        try {
                            if (userDao.setIehrUserProfile(userProfile)) {
                                if (userProfile.dbUserID == "") {
                                    userProfile = userDao.getIehrUserProfile(userProfile.smartCardID, userProfile.smartCardAgency)
                                }
                            }
                        }
                        catch (Exception e) {
                            errorText = "Could not save the User Profile."
                            log.error("AuthenticateController::setIehrUserProfile", e)
                            render(view: "login", model: [errorText: errorText, grailsConfig: grailsApplication.config.grails, "sites": sites, "userProfile": userProfile, "theme": theme, "dodVLER": dodVLER, "loginAgency": loginAgency, "loginSite": loginSite, "agreement": agreement,
                                                          "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": serviceStatus, "csrfToken": csrfToken])
                            return
                        }

                        appUser.setSecurityKey(tokenUtils.generateSecurityKey())
                        appUser.setUserId(userProfile.dbUserID);
                        appUser.setCfg(userProfile.cfg);

                        //**melcolm remove this if/when jimi fixes login mock service
                        if (appUser.getHostSite() == null && appConfig.getSmartCardBypass()) {
                            GUISite hostSite = new GUISite()
                            hostSite.setAgency(loginCmd.loginAgency);
                            hostSite.setPermissions(null);
                            appUser.setHostSite(hostSite);
                            appUser.setName("McLuvin,Dood");
                            appUser.setAgency(userProfile.getSmartCardAgency());
                            appUser.setCardId(userProfile.getSmartCardID());
                            appUser.setEmail(userProfile.getEmailAddress());
                        }

                        session.appUser = appUser
                        session.userProfile = userProfile
                        session.setAttribute("acceptedAgreement", agreement)

                        // User JAM widget access flag
                        if ((userProfile != null) && (userProfile.flags != null) && (userProfile.flags.length() > 0)) {
                            def flags = JSON.parse(userProfile.flags)
                            for (item in flags) {
                                if (item['JAM'] != null) {
                                    session.jamWidgetAccessFlag = item['JAM']
                                }
                            }
                        }

                        redirect(controller: 'app')
                    }
                }
                catch (Exception e) {
                    log.error("AuthenticateController::login", e)

                    render(view: "login", model: [errorText: errorText, grailsConfig: grailsApplication.config.grails, "sites": sites, "userProfile": userProfile, "theme": theme, "dodVLER": dodVLER, "loginAgency": loginAgency, "loginSite": loginSite, "agreement": agreement,
                                                  "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": serviceStatus, "csrfToken": csrfToken])
                    return
                }
            } else {
                render(view: "login", model: [loginCmd: loginCmd, grailsConfig: grailsApplication.config.grails, "sites": sites, "theme": theme, "dodVLER": dodVLER, "userProfile": userProfile, "loginAgency": loginAgency, "loginSite": loginSite, "agreement": agreement,
                                              "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": serviceStatus, "csrfToken": csrfToken])
                return
            }
        }

        if (userProfile.smartCardID != null) {
            try {
                loginEntries = userDao.getLoginInfo(userProfile.smartCardID, userProfile.emailAddress);
                if (loginEntries != null) {
                    Iterator<GUIAuditEntry> i = loginEntries.iterator()
                    while (i.hasNext()) {
                        GUIAuditEntry loginEntry = i.next()
                        if (loginEntry.userID.startsWith("^")) {
                            badLoginDate = loginEntry.entryDate;
                            badLoginIP = loginEntry.ipAddress;
                        } else {
                            validLoginDate = loginEntry.entryDate;
                            validLoginIP = loginEntry.ipAddress;
                        }
                    }
                    session.loginEntries = loginEntries
                }
            }
            catch (Exception e) {
                log.error("AuthenticateController::login - Unable to retrieve previous login information", e)
//                String errorText = "There's an error retrieving your previous login information<br>Please notify the system administrator";
//                render(view:"login", model:[errorText:errorText, grailsConfig:grailsApplication.config.grails,"userProfile":userProfile,"theme":theme, "loginAgency":loginAgency, "loginSite":loginSite, "csrfToken" : csrfToken])
//                return
            }
        }

        [grailsConfig: grailsApplication.config.grails, "sites": sites, "userProfile": userProfile, "theme": theme, "dodVLER": dodVLER, "loginAgency": loginAgency, "loginSite": loginSite, "agreement": agreement,
         "badLoginDate": badLoginDate, "badLoginIP": badLoginIP, "validLoginDate": validLoginDate, "validLoginIP": validLoginIP, "serviceStatus": serviceStatus, "csrfToken": csrfToken]
    }

    def logout = {
        session.setAttribute("acceptedAgreement", false);
        session.invalidate()
    }

    def disclaimer = {
        if (request.method == 'GET') {
            //generate and set csrf token
            def csrfToken = UUID.randomUUID().toString()
            session.setAttribute("X-CSRF-Token", csrfToken)
            render(view: 'disclaimer', model: [csrfToken: csrfToken])
        } else if (request.method == 'POST') {
            session.setAttribute("acceptedAgreement", params['acceptAgreement']);
            redirect(controller: 'app')
        }
    }


}

class LoginCommand {
    String siteCode
    String loginAgency
    String accessCode
    String verifyCode
    String userCfg
    String acceptedAgreement

    static constraints = {
        siteCode blank: false
        accessCode blank: false
        verifyCode blank: false
        loginAgency blank: false
        userCfg blank: false
        acceptedAgreement blank: false
    }
}

class CommunityLoginCommand {
    String userName
    String password
    String acceptedAgreement

    static constraints = {
        userName blank: false
        password blank: false
        acceptedAgreement blank: false
    }
}