/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */

package jlv

import gov.va.med.common.HttpKeys
import gov.va.med.dao.DaoException
import gov.va.med.jmeadows.webservice.VlerDocument
import org.springframework.context.ApplicationContextException
import org.w3c.dom.Document

import javax.servlet.ServletContext
import javax.xml.transform.*
import javax.xml.transform.stream.StreamResult
import javax.xml.transform.stream.StreamSource

class VLERController extends BaseController {
    def VLERDao;
    def auditDao;

    static allowedMethods = [list: 'POST', details: 'POST']

    private static
    final String XPATH_CCDA_TEMPLATEID = "/ClinicalDocument/templateId[@root='2.16.840.1.113883.10.20.22.1.1']";
    private static final String CCDA_NODE_ID = "templateId root=\"2.16.840.1.113883.10.20.22.1.1\"";
    private static final String XSL_PATH_C32 = "/c32styles/nhin/cda2detail.xsl";
    private static final String XSL_PATH_CCDA = "/c32styles/nhin/cdaHFG.xsl";

    def beforeInterceptor = {

        if (request.patient == null) {
            response.sendError 500
            return false;
        }

        if (actionName == 'details') {
            if (request.recordQuery == null) {
                response.sendError 500
                return false;
            }
        }
    }

    def index = {

        PrintWriter out = response.getWriter();
        // setting some response headers
        response.setHeader("Expires", "0");
        response.setHeader("Cache-Control", "must-revalidate, post-check=0, pre-check=0");
        response.setHeader("Pragma", "public");
        // setting the content type
        response.setContentType("text/html");

        def auditQueryName = "GetVLERDocument"
        auditDao.auditDetailsQuery(auditQueryName, session.appUser, request.patient,
                "retrievingC32:" + request.recordQuery[HttpKeys.VLER_DOCUMENT_ID],
                request.recordQuery[HttpKeys.RECORD_SITE_MONIKER],
                request.recordQuery[HttpKeys.RECORD_SITE_AGENCY])

        VlerDocument vDoc = new VlerDocument();
        vDoc.setDocumentUniqueId(request.recordQuery[HttpKeys.VLER_DOCUMENT_ID]);
        vDoc.setHomeCommunityId(request.recordQuery[HttpKeys.VLER_HOMECOMMUNITY_ID]);
        vDoc.setRepositoryUniqueId(request.recordQuery[HttpKeys.VLER_REPOSITORY_ID]);
        String titleString = "JLV - " + request.recordQuery[HttpKeys.VLER_DOCUMENT_INSTITUTION] + " : " + request.recordQuery[HttpKeys.VLER_DOCUMENT_NAME];

        Document vlerDoc = null;

        try {
            vDoc = VLERDao.getVLERDocument(session.appUser, request.patient, vDoc);
        }
        catch (DaoException e) {
            log.error(e);
        }

        if (vDoc != null && vDoc.getDocument() != null) {
            try {

                String dataXML = new String(vDoc.getDocument(), "UTF-8");

                String xslTemplate = XSL_PATH_C32;
                if (isCcdaDoc(dataXML)) {
                    xslTemplate = XSL_PATH_CCDA
                }

                String outputHTML = "";

                outputHTML = transform(dataXML, xslTemplate);

                outputHTML = outputHTML.replace("</head>", "<title>" + titleString + "</title>" +
                        "<script type='text/javascript' src='resources/scripts/js/nhin.js' ></script>" +
                        "</head>");
                out.println(outputHTML);

                return true;
            }
            catch (TransformerConfigurationException e) {
                log.error(e);
            }
            catch (TransformerException e) {
                log.error(e);
            }
        }

        String errorMsg = "An error occurred while attempting to retrieve the document.";

        out.println(errorMsg);

        return true;

    }

    public String transform(String xmlString, String xslTemplate) throws TransformerConfigurationException, TransformerException {
        try {

            ServletContext context = session.getServletContext();
            InputStream xslInput = context.getResourceAsStream(xslTemplate);

            String xslInputPath = context.getResource(xslTemplate).toExternalForm();
            String xslSystemId = xslInputPath.substring(0, xslInputPath.lastIndexOf('/') + 1);

            StreamSource xmlInput = new StreamSource(new StringReader(xmlString));

            Source xslStreamSource = null;
            if (XSL_PATH_C32.equalsIgnoreCase(xslTemplate)) {
                xslStreamSource = new StreamSource(xslInput, xslSystemId);
            } else {
                xslStreamSource = new StreamSource(xslInput, xslInputPath);
            }

            TransformerFactory factory = TransformerFactory.newInstance();
            Transformer transformer = factory.newTransformer(xslStreamSource);

            StringWriter outputWriter = new StringWriter();
            Result outputResult = new StreamResult(outputWriter);
            transformer.transform(xmlInput, outputResult);
            return outputWriter.toString();

        } catch (Exception e) {
            e.printStackTrace(System.out);
        }

        return "An error occurred while attempting to retrieve the document: Error during transform";
    }

    protected Source getStylesheetSource(String xslTemplate) throws ApplicationContextException {

        try {

            URL requestURL = new URL(request.getRequestURL().toString())
            URL xslURL = new URL(requestURL, xslTemplate);
            URL url = xslURL;

            String urlPath = url.toString();
            if (XSL_PATH_CCDA.equals(xslTemplate)) {
                return new StreamSource(url.openStream(), urlPath);
            } else {
                String systemId = urlPath.substring(0, urlPath.lastIndexOf('/') + 1);
                return new StreamSource(url.openStream(), systemId);
            }
        } catch (IOException ex) {
            throw new ApplicationContextException("Can't load XSLT stylesheet from " + xslTemplate, ex);
        }
    }

    /*
 * Return true if XML document is CCDA document.
 *
 * @param ccdDoc VLER CCD XML document.
 * @return true if root CCDA template ID is present.  Otherwise, false.
 * @throws VlerDocumentUtilException If an error occurs while parsing the CCD.
 */

    public static boolean isCcdaDoc(String ccdXML) throws ApplicationContextException {
        if (ccdXML.indexOf(CCDA_NODE_ID) > 0) {
            return true;
        }
        return false;
    }


}
