/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.dao;

import gov.va.med.guibeans.*;
import gov.va.med.jmeadows.webservice.Code;
import gov.va.med.jmeadows.webservice.FreeTextReport;
import gov.va.med.jmeadows.webservice.Order;
import gov.va.med.DNS   .common.Utils;
import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

@Component
public class OrderDao extends DaoBase {
    private static final Logger logger = Logger.getLogger(OrderDao.class.getName());

    public List<GUIOrder> getPatientOrders(GUIUser user, GUIPatient patient, Calendar startDate, Calendar endDate) {
        try {
            Order[] orders = getServiceHandler().getPatientOrders(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUIOrder> guiOrders = new ArrayList<GUIOrder>();

            for (Order order : orders) {
                String orderDate = "";
                String orderId = "";
                String type = "";
                String providerName = "";
                String specialty = "";
                String description = "";
                String status = "";
                String priority = "";

                if (order.getOrderDate() != null) {
                    orderDate = Utils.formatDisplayDate(order.getOrderDate());
                }

                if (order.getOrderid() != null) {
                    orderId = order.getOrderid();
                }

                if (order.getType() != null) {
                    type = order.getType();
                }

                if (order.getOrderingProvider() != null) {
                    providerName = order.getOrderingProvider().getName();
                    specialty = order.getOrderingProvider().getSpecialization();
                }

                if (order.getDescription() != null) {
                    description = order.getDescription();
                }

                if (order.getStatus() != null) {
                    status = order.getStatus();
                }

                if(order.getPriority() != null){
                    priority = order.getPriority();
                }

                GUIOrder guiOrder = new GUIOrder();
                guiOrder.setDescription(description);
                guiOrder.setOrderDate(orderDate);
                guiOrder.setOrderId(orderId);
                guiOrder.setProviderName(providerName);
                if (order.getOrderingProvider() != null) {
                    guiOrder.setProvider(dozerBeanMapper.map(order.getOrderingProvider(), GUIProvider.class));
                }
                guiOrder.setStatus(status);
                guiOrder.setType(type);
                guiOrder.setPriority(priority);

                List<GUICode> guiOrderCodes = new ArrayList<GUICode>();
                for (Code orderCode : order.getCodes()) {
                    guiOrderCodes.add(dozerBeanMapper.map(orderCode, GUICode.class));
                }
                guiOrder.setCodes(guiOrderCodes);

                addSiteDataToGUIBean(guiOrder, order);

                guiOrders.add(guiOrder);
            }

            return guiOrders;
        } catch (RuntimeException e) {
            throw new DaoException(e);
        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public GUIOrderDetail getOrderDetail(GUIUser user, GUIPatient patient, String orderId,
                                         String recordSiteCode) {
        try {
            FreeTextReport report = getServiceHandler().getOrderDetail(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .patient(patient)
                            .itemId(orderId)
                            .recordSiteCode(recordSiteCode)
                            .build());

            GUIOrderDetail orderDetail = new GUIOrderDetail();

            if (report != null) {
                orderDetail.setOrderReport(report.getReportText());
                addSiteDataToGUIBean(orderDetail, report);
            }

            return orderDetail;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUIOrder> getProviderOrdersPending(
            GUIUser user,
            Calendar startDate,
            Calendar endDate,
            byte[] encryptSecurityKey) {
        try {

            Order[] orders = getServiceHandler().getProviderOrdersPending(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUIOrder> guiOrders = new ArrayList<GUIOrder>();

            for (Order order : orders) {
                String orderDate = "";
                String orderId = "";
                String type = "";
                String providerName = "";
                String description = "";
                String status = "";
                String patToken = "";
                String patientName = "";

                if (order.getOrderDate() != null) {
                    orderDate = Utils.formatDisplayDate(order.getOrderDate());
                }

                if (order.getOrderid() != null) {
                    orderId = order.getOrderid();
                }

                if (order.getType() != null) {
                    type = order.getType();
                }

                if (order.getOrderingProvider() != null) {
                    providerName = order.getOrderingProvider().getName();
                }

                if (order.getDescription() != null) {
                    description = order.getDescription();
                }

                if (order.getStatus() != null) {
                    status = order.getStatus();
                }

                if (order.getPatientName() != null) {
                    patientName = order.getPatientName();
                }

                patToken = getProvQueryPatientToken(encryptSecurityKey, order);

                GUIOrder guiOrder = new GUIOrder();
                guiOrder.setDescription(description);
                guiOrder.setOrderDate(orderDate);
                guiOrder.setOrderId(orderId);
                guiOrder.setProviderName(providerName);
                guiOrder.setStatus(status);
                guiOrder.setType(type);

                addSiteDataToGUIBean(guiOrder, order);

                guiOrder.setPatientName(patientName);
                guiOrder.setPatientToken(patToken);

                guiOrders.add(guiOrder);
            }

            return guiOrders;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }

    public List<GUIOrder> getProviderOrdersResulted(
            GUIUser user,
            Calendar startDate,
            Calendar endDate,
            byte[] encryptSecurityKey) {
        try {
            Order[] orders = getServiceHandler().getProviderOrdersResulted(
                    jMeadowsQueryBuilderFactory.getBuilder()
                            .user(user)
                            .startDate(startDate)
                            .endDate(endDate)
                            .build());

            List<GUIOrder> guiOrders = new ArrayList<GUIOrder>();

            for (Order order : orders) {
                String orderDate = "";
                String orderId = "";
                String type = "";
                String providerName = "";
                String description = "";
                String status = "";
                String patientName = "";
                String patToken = "";

                if (order.getOrderDate() != null) {
                    orderDate = Utils.formatDisplayDate(order.getOrderDate());
                }

                if (order.getOrderid() != null) {
                    orderId = order.getOrderid();
                }

                if (order.getType() != null) {
                    type = order.getType();
                }

                if (order.getOrderingProvider() != null) {
                    providerName = order.getOrderingProvider().getName();
                }

                if (order.getDescription() != null) {
                    description = order.getDescription();
                }

                if (order.getStatus() != null) {
                    status = order.getStatus();
                }

                if (order.getPatientName() != null) {
                    patientName = order.getPatientName();
                }

                patToken = getProvQueryPatientToken(encryptSecurityKey, order);


                GUIOrder guiOrder = new GUIOrder();
                guiOrder.setDescription(description);
                guiOrder.setOrderDate(orderDate);
                guiOrder.setOrderId(orderId);
                guiOrder.setProviderName(providerName);
                guiOrder.setStatus(status);
                guiOrder.setType(type);

                addSiteDataToGUIBean(guiOrder, order);

                guiOrder.setPatientName(patientName);
                guiOrder.setPatientToken(patToken);

                guiOrders.add(guiOrder);
            }

            return guiOrders;

        } catch (Exception e) {
            throw new DaoException(e);
        }
    }
}
