/*
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *             Honorable Senator Daniel K. Inouye
 *             VA Pacific Islands Health Care System
 *             Tripler Army Medical Center
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 *
 *            http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 */


package gov.va.med.jmeadows;

import gov.va.med.common.AppConfig;
import gov.va.med.jmeadows.webservice.*;
import gov.va.med.DNS   .common.Utils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPException;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.soap.MTOMFeature;
import java.util.ArrayList;
import java.util.List;

public class JMeadowsServiceHandlerImpl implements JMeadowsServiceHandler {

    @Autowired
    protected AppConfig appConfig;

    private static final Logger logger = Logger.getLogger(JMeadowsServiceHandlerImpl.class.getName());

    private String serviceURL;
    private String m_ServiceName;
    private String m_namespace;
    private QName m_QserviceName;

    private String appName;

    private String username;
    private String password;

    private int connectionTimeoutMS;
    private int requestTimeoutMS;

    private JMeadowsDataService m_jmeadowsDataService;

    public JMeadowsServiceHandlerImpl(String url) {
        serviceURL = url;
        init();
    }

    public JMeadowsServiceHandlerImpl() {
        init();
    }

    public void init() {
        m_ServiceName = "JMeadowsDataService";
        m_namespace = "http://webservice.jmeadows.med.DNS   /";
        m_QserviceName = new QName(m_namespace, m_ServiceName);

        connectionTimeoutMS = 30000;
        requestTimeoutMS = 30000;

    }

    public int getConnectionTimeoutMS() {
        return connectionTimeoutMS;
    }

    public void setConnectionTimeoutMS(int connectionTimeoutMS) {
        this.connectionTimeoutMS = connectionTimeoutMS;
    }

    public int getRequestTimeoutMS() {
        return requestTimeoutMS;
    }

    public void setRequestTimeoutMS(int requestTimeoutMS) {
        this.requestTimeoutMS = requestTimeoutMS;
    }

    public String getServiceURL() {
        return serviceURL;
    }

    public void setServiceURL(String serviceURL) {
        this.serviceURL = serviceURL;
    }

    private JMeadowsData getJMeadowsDataPort() {
        return getJMeadowsDataPort(false);
    }

    public String getUsername() {
        return username;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    private JMeadowsData getJMeadowsDataPort(boolean includeMTOMFeature) {
        if (m_jmeadowsDataService == null) {
            createService();
            if (appConfig.getEnableWSSecurityHeader()) {
                m_jmeadowsDataService.setHandlerResolver(new HeaderHandlerResolver());
            }
        }

        JMeadowsData port = null;
        if (includeMTOMFeature) {
            port = m_jmeadowsDataService.getJMeadowsDataPort(new MTOMFeature());
        } else port = m_jmeadowsDataService.getJMeadowsDataPort();

        //set connection timeout
        BindingProvider bp = (BindingProvider) port;
        bp.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, serviceURL);
        bp.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", connectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", requestTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.connect.timeout", connectionTimeoutMS);
        bp.getRequestContext().put("com.sun.xml.ws.request.timeout", requestTimeoutMS);

        return port;
    }

    private synchronized void createService() {
        if (m_jmeadowsDataService == null) {
            try {
                m_jmeadowsDataService = new JMeadowsDataService();
            } catch (Exception e) {
                throw new RuntimeException(e);
            }

        }
    }

    private void log(String methodName) {
        log(methodName, null);
    }

    private void log(String methodName, JMeadowsQuery queryBean) {
        if (logger.isDebugEnabled()) {
            logger.debug("Invoking method: " + methodName);
            if (queryBean != null) {
                logQueryBeanParams(methodName, queryBean);
            }
        }
    }

    private void logQueryBeanParams(String methodName, JMeadowsQuery queryBean) {
        if (!logger.isDebugEnabled()) return;

        String userId;
        Site hostSite;
        String patVAIcn;
        String patDoDEdipi;
        List<Site> vistaSites = new ArrayList<Site>();
        String itemId;
        String recordSiteCode;
        String status;
        Integer active;
        String startDate;
        String endDate;

        StringBuilder sb = new StringBuilder();
        sb.append(methodName);
        sb.append(" query bean params: ");

        if (queryBean != null) {
            User user = queryBean.getUser();
            Patient patient = queryBean.getPatient();

            if (user != null) {
                userId = user.getUserId();
                hostSite = user.getHostSite();

                sb.append("User - userId: ");
                sb.append(userId);
                sb.append(" , hostSite: ");
                if (hostSite != null) sb.append(hostSite.getSiteCode());
            } else {
                sb.append("User - null");
            }

            if (patient != null) {

                patVAIcn = patient.getICN();
                patDoDEdipi = patient.getEDIPI();

                sb.append("; Patient - ");
                sb.append(", ICN: ");
                sb.append(patVAIcn);
                sb.append(", EDIPI: ");
                sb.append(patDoDEdipi);

                sb.append(", vistaSites: ");

                if (vistaSites.size() < 1) sb.append("none");

                for (Site site : vistaSites) {
                    String siteCode = site.getSiteCode();
                    sb.append(siteCode);
                    sb.append(", ");
                }
            } else {
                sb.append("; Patient - null");
            }

            itemId = queryBean.getItemId();
            active = queryBean.getActive();
            recordSiteCode = queryBean.getRecordSiteCode();
            startDate = Utils.formatDisplayDate(queryBean.getStartDate());
            endDate = Utils.formatDisplayDate(queryBean.getEndDate());
            status = queryBean.getStatus();

            sb.append("; startDate: ");
            sb.append(startDate);
            sb.append(", endDate: ");
            sb.append(endDate);
            sb.append(", itemId: ");
            sb.append(itemId);
            sb.append(", recordSiteCode: ");
            sb.append(recordSiteCode);
            sb.append(", status: ");
            sb.append(status);
            sb.append(", active: ");
            sb.append(active);

        }

        logger.debug(sb.toString());
    }

    @Override
    public User loginEnterprise(String cardID, String siteCode, String accessCode, String verifyCode, String ipAddress, String subjectDN, String emailAddress) throws SOAPException {

        log("loginEnterprise");

        User result = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            result = port.loginEnterprise(cardID, siteCode, accessCode, verifyCode, appName, ipAddress, subjectDN, emailAddress);

        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return result;
    }

//    @Override
//    public User loginEnterprise(String cardID, String siteCode, String accessCode, String verifyCode, String ipAddress, String subjectDN) throws SOAPException {
//
//        log("loginEnterprise");
//
//        User result = null;
//
//        try {
//
//            JMeadowsData port = getJMeadowsDataPort();
//            result = port.loginEnterprise(cardID, siteCode, accessCode, verifyCode, appName, ipAddress, subjectDN, "");
//
//        } catch (Exception e) {
//            throw new SOAPException(e);
//        }
//
//        return result;
//
//        //code to mock DoD user
////        siteCode = "200";
////
////        User user = new User();
////        user.setEncryptedFederatedUid(EncryptUtil.encrypt("123456789"));
////        user.setName("MockUser,DoD");
////        user.setUserIen("1234");
////        user.setProviderIen("4321");
////        user.setAgency("DOD");
////
////        JMeadowsData port = getJMeadowsDataPort();
////        List<Site> sites = port.getSites();
////        java.util.Iterator<Site> it = sites.iterator();
////        while(it.hasNext())
////        {
////            Site site = it.next();
////            if(siteCode.equalsIgnoreCase(site.getSiteCode()))
////            {
////                user.setHostSite(site);
////                break;
////            }
////        }
////        return user;
//
//    }

    @Override
    public AdmissionDetail getAdmissionDetails(JMeadowsQuery queryBean) throws SOAPException {

        log("getAdmissionDetails", queryBean);

        AdmissionDetail returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getAdmissionDetails(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }


        return returnVal;
    }

    @Override
    public Diagnosis[] getAdmissionDiagnosis(JMeadowsQuery queryBean) throws SOAPException {

        log("getAdmissionDiagnosis", queryBean);

        Diagnosis[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getAdmissionDiagnosis(queryBean).toArray(new Diagnosis[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }


    @Override
    public Procedure[] getAdmissionProcedures(JMeadowsQuery queryBean) throws SOAPException {

        log("getAdmissionProcedures", queryBean);

        Procedure[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getAdmissionProcedures(queryBean).toArray(new Procedure[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public FreeTextReport getClinicalReminderDetail(JMeadowsQuery queryBean) throws SOAPException {

        log("getClinicalReminderDetail", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getClinicalReminderDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public FreeTextReport[] getConsultReport(JMeadowsQuery queryBean) throws SOAPException {

        log("getConsultReport", queryBean);

        FreeTextReport[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getConsultReport(queryBean).toArray(new FreeTextReport[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public FreeTextReport getMedicationDetails(JMeadowsQuery queryBean) throws SOAPException {

        log("getMedicationDetails", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getMedicationDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public LabResult[] getLabOrderResult(JMeadowsQuery queryBean) throws SOAPException {
        log("getLabOrderResult", queryBean);

        LabResult[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getLabOrderResult(queryBean).toArray(new LabResult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public FreeTextReport getOrderDetail(JMeadowsQuery queryBean) throws SOAPException {

        log("getOrderDetail", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getOrderDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public PatientAdmission[] getPatientAdmissions(JMeadowsQuery queryBean) throws SOAPException {

        log("getPatientAdmissions", queryBean);

        PatientAdmission[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientAdmissions(queryBean).toArray(new PatientAdmission[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Allergy[] getPatientAllergies(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientAllergies", queryBean);

        Allergy[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientAllergies(queryBean).toArray(new Allergy[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public AllergyDetail getAllergyDetail(JMeadowsQuery queryBean) throws SOAPException {
        log("getAllergyDetail", queryBean);

        AllergyDetail returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getAllergyDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public PatientAppointments[] getPatientAppointments(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientAppointments", queryBean);

        PatientAppointments[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientAppointments(queryBean).toArray(new PatientAppointments[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Encounter[] getPatientEncounters(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientEncounter", queryBean);

        Encounter[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientEncounters(queryBean).toArray(new Encounter[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public ClinicalReminder[] getPatientClinicalReminders(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientClinicalReminders", queryBean);

        ClinicalReminder[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientClinicalReminders(queryBean).toArray(new ClinicalReminder[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Consult[] getPatientConsultRequests(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientConsultRequests", queryBean);

        Consult[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientConsultRequests(queryBean).toArray(new Consult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Vitals[] getPatientCurrentVitals(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientCurrentVitals", queryBean);

        Vitals[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientCurrentVitals(queryBean).toArray(new Vitals[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public List<PatientDemographicsDetail> getAllPatientDemographicDetails(JMeadowsQuery queryBean) throws SOAPException {
        log("getAllPatientDemographicDetails", queryBean);

        List<PatientDemographicsDetail> returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientDemographicsDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public List<PatientDemographics> getPatientDemographics(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientDemographics", queryBean);

        List<PatientDemographics> returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientDemographics(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Procedure[] getPatientProcedures(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientProcedures", queryBean);

        Procedure[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getOutpatientProcedures(queryBean).toArray(new Procedure[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }


//    @Override
//    public QuestionnaireBean[] getPatientQuestionnaires(JMeadowsQuery queryBean) throws SOAPException {
//        log("getPatientQuestionnaires", queryBean);
//
//        QuestionnaireBean[] returnVal = null;
//
//        try {
//
//            JMeadowsData port = getJMeadowsDataPort();
//            returnVal = port.getPatientQuestionnaires(queryBean).toArray(new QuestionnaireBean[0]);
//        } catch (Exception e) {
//            throw new SOAPException(e);
//        }
//
//        return returnVal;
//    }

    @Override
    public Form[] getPatientForms(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientForms", queryBean);

        Form[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientForms(queryBean).toArray(new Form[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public ProgressNote[] getPatientDischargeSummaries(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientDischargeSummaries", queryBean);

        ProgressNote[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientDischargeSummaries(queryBean).toArray(new ProgressNote[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Immunization[] getPatientImmunizations(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientImmunizations", queryBean);

        Immunization[] returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientImmunizations(queryBean).toArray(new Immunization[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public InsuranceBean[] getPatientInsurances(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientInsurances", queryBean);

        InsuranceBean[] returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientInsurances(queryBean).toArray(new InsuranceBean[0]);

        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }


    @Override
    public LabOrder[] getPatientLabs(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientLabs", queryBean);

        LabOrder[] returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientLabs(queryBean).toArray(new LabOrder[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public FreeTextReport getPatientLabReport(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientLabReport", queryBean);

        FreeTextReport returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientLabReport(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public LabResult[] getPatientLabTestResults(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientLabTestResults", queryBean);

        LabResult[] returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientLabTestResults(queryBean).toArray(new LabResult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public LabResult[] getPatientLabResults(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientLabResults", queryBean);

        LabResult[] returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientLabResults(queryBean).toArray(new LabResult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Medication[] getPatientMedications(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientMedications", queryBean);

        Medication[] returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientMedications(queryBean).toArray(new Medication[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

//    public Medication[] getPatientMedicationsNVA(JMeadowsQuery queryBean) throws SOAPException {
//        log("getPatientMedicationsNVA", queryBean);
//
//        Medication[] returnVal = null;
//        try {
//
//            JMeadowsData port = getJMeadowsDataPort();
//            returnVal = port.getPatientMedicationsNVA(queryBean).toArray(new Medication[0]);
//        } catch (Exception e) {
//            throw new SOAPException(e);
//        }
//        return returnVal;
//
//    }

    @Override
    public Order[] getPatientOrders(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientOrders", queryBean);

        Order[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientOrders(queryBean).toArray(new Order[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Problem[] getPatientProblemList(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientProblemList", queryBean);

        Problem[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientProblemList(queryBean).toArray(new Problem[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

//      public Procedure[] getPatientProcedures(JMeadowsQuery queryBean) throws SOAPException
//      {
//              Procedure[] returnVal = null;
//
//        try {
//
//                    JMeadowsData port = getJMeadowsDataPort();
//                    returnVal = port.getPatientProcedures(queryBean).toArray(new Procedure[0]);
//            }
//            catch (Exception e)
//            {
//                    throw new SOAPException(e);
//            }
//
//        return returnVal;
//
//      }

    @Override
    public ProgressNote[] getPatientProgressNotes(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientProgressNotes", queryBean);

        ProgressNote[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientProgressNotes(queryBean).toArray(new ProgressNote[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    /**
     * Fetches a BHIE note image.
     *
     * @param queryBean required params: provider, itemid (containing bhie note event id)
     * @return Note image
     * @throws javax.xml.soap.SOAPException if an error occurs.
     */
    @Override
    public NoteImage getBHIENoteImage(JMeadowsQuery queryBean) throws SOAPException {
        log("getBHIENoteImage", queryBean);

        try {
            return getJMeadowsDataPort(true).getBHIENoteImage(queryBean);

        } catch (Exception e) {
            throw new SOAPException(e);
        }

    }

    @Override
    public RadiologyReport[] getPatientRads(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientRads", queryBean);

        RadiologyReport[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientRads(queryBean).toArray(new RadiologyReport[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Vitals[] getPatientVitals(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientVitals", queryBean);

        Vitals[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientVitals(queryBean).toArray(new Vitals[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    //    public Prescription getPrescriptionFills(JMeadowsQuery queryBean) throws SOAPException {
//        log("getPrescriptionFills", queryBean);
//
//        Prescription returnVal = null;
//
//        try {
//
//            JMeadowsData port = getJMeadowsDataPort();
//            returnVal = port.getPrescriptionFills(queryBean);
//        } catch (Exception e) {
//            throw new SOAPException(e);
//        }
//
//        return returnVal;
//
//    }
    @Override
    public PatientHistory[] getPatientHistories(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientHistories", queryBean);

        PatientHistory[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientHistory(queryBean).toArray(new PatientHistory[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public VlerDocument[] getVLERDocumentList(JMeadowsQuery queryBean) throws SOAPException {
        log("getVLERDocumentList", queryBean);

        VlerDocument[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getVLERDocumentList(queryBean).toArray(new VlerDocument[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public VlerDocument getVLERDocument(JMeadowsQuery queryBean, VlerDocument vDoc) throws SOAPException {
        log("getVLERDocument", queryBean);

        VlerDocument returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getVLERDocument(vDoc, queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Vlerccda[] getDODVLERPatientCCDA(JMeadowsQuery queryBean) throws SOAPException {
        log("getDODVLERPatientCCDA", queryBean);

        Vlerccda[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getDODVLERPatientCCDA(queryBean).toArray(new Vlerccda[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public List<OutsideProvider> getOutsideProviderList(JMeadowsQuery queryBean, String searchName) throws SOAPException {

        List<OutsideProvider> returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getOutsideProviderList(queryBean, searchName);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public OutsideProvider getOutsideProvider(JMeadowsQuery queryBean, String username, String password) throws SOAPException {

        OutsideProvider returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getOutsideProvider(queryBean, username, password);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public void addOutsideProvider(JMeadowsQuery queryBean, OutsideProvider outsideProvider) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.addOutsideProvider(queryBean, outsideProvider);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }


    @Override
    public void deleteOutsideProvider(JMeadowsQuery queryBean, String id) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.deleteOutsideProvider(queryBean, id);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public List<AssignedPatient> getAssignedPatientList(JMeadowsQuery queryBean, String providerId) throws SOAPException {

        List<AssignedPatient> returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getAssignedPatientList(queryBean, providerId);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public void addAssignedPatient(JMeadowsQuery queryBean, AssignedPatient assignedPatient) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.addAssignedPatient(queryBean, assignedPatient);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

    }

    @Override
    public void deleteAssignedPatient(JMeadowsQuery queryBean, String assignedPatientId) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.deleteAssignedPatient(queryBean, assignedPatientId);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public void addAssignedPatientUsage(JMeadowsQuery queryBean, String providerId, String assignedPatientId) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.addAssignedPatientUsage(queryBean, providerId, assignedPatientId);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public List<Consult> getPatientActiveConsultRequest(JMeadowsQuery queryBean) throws SOAPException {
        List<Consult> returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientActiveConsultRequests(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public void updateOutsideProvider(JMeadowsQuery queryBean, OutsideProvider provider) throws SOAPException {

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.updateOutsideProvider(queryBean, provider);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

    }

    @Override
    public OutsideProvider getOutsideProviderExist(JMeadowsQuery queryBean, String username) throws SOAPException {

        OutsideProvider returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getOutsideProviderExist(queryBean, username);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }
    

    @Override
    public ProblemDetail getProblemDetail(JMeadowsQuery queryBean) throws SOAPException {
        log("getProblemDetail", queryBean);

        ProblemDetail returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProblemDetail(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public FreeTextReport getProgressNote(JMeadowsQuery queryBean) throws SOAPException {
        log("getProgressNote", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProgressNote(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }


    @Override
    public User getProviderProfile(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderProfile", queryBean);

        User returnVal = null;
        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderProfile(queryBean);

        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public RadiologyReport getRadiologyReport(JMeadowsQuery queryBean) throws SOAPException {
        log("getRadiologyReport", queryBean);

        RadiologyReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getRadiologyReport(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public Alert[] getUserAlerts(JMeadowsQuery queryBean) throws SOAPException {
        log("getUserAlerts", queryBean);

        Alert[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getUserAlerts(queryBean).toArray(new Alert[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;

    }

    @Override
    public FreeTextReport getVisitNotes(JMeadowsQuery queryBean) throws SOAPException {
        log("getVisitNotes", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getVisitNotes(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public FreeTextReport getPatientEncountersReport(JMeadowsQuery queryBean) throws SOAPException {
        log("getPatientEncountersReport", queryBean);

        FreeTextReport returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getPatientEncountersReport(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Patient[] lookupPatient(JMeadowsQuery queryBean) throws SOAPException {
        log("lookupPatient", queryBean);

        Patient[] returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.lookupPatient(queryBean).toArray(new Patient[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public ResponsePatientQuery lookupPatientPDWS(PdwsQueryBean queryBean) throws SOAPException {
        log("lookupPatientPDWS");

        ResponsePatientQuery returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.lookupPatientPDWS(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public Patient selectPatient(JMeadowsQuery queryBean) throws SOAPException {
        log("selectPatient", queryBean);

        Patient returnVal = null;

        try {

            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.selectPatient(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;

    }

    @Override
    public ResponsePatientSelect selectPatientMVI(JMeadowsQuery queryBean) throws SOAPException {
        log("selectPatientMVI");

        ResponsePatientSelect returnVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.selectPatientMVI(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

        return returnVal;
    }

    @Override
    public void selectPatientForVASensitive(JMeadowsQuery queryBean) throws SOAPException {
        log("selectPatientForVASensitive");

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.selectPatientForVASensitive(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public void auditVARestrictedAccess(JMeadowsQuery queryBean) throws SOAPException {
        log("auditVARestrictedAccess");

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.auditVARestrictedAccess(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public void auditSensitiveRecordAccess(JMeadowsQuery queryBean) throws SOAPException {
        log("auditSensitiveRecordAccess");

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.auditSensitiveRecordAccess(queryBean);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
    }

    @Override
    public void logAudit(String siteId, String userId, String providerId, String userNPI,
                         String userName, String patId, String category,
                         String startDate, String endDate, String info, String cardId, String emailAddress) throws SOAPException {

        log("logAudit");

        try {
            JMeadowsData port = getJMeadowsDataPort();
            port.logAudit(siteId, userId, providerId, userNPI, userName, patId, category, appName, startDate, endDate, info, cardId, emailAddress);
        } catch (Exception e) {
            throw new SOAPException(e);
        }

    }

//    @Override
//    public void logAudit(String siteId, String userId, String providerId, String userNPI,
//                         String userName, String patId, String category,
//                         String startDate, String endDate) throws SOAPException {
//
//        log("logAudit");
//
//        try {
//            JMeadowsData port = getJMeadowsDataPort();
//            port.logAudit(siteId, userId, providerId, userNPI, userName, patId, category, appName, startDate, endDate);
//        } catch (Exception e) {
//            throw new SOAPException(e);
//        }
//
//    }

    @Override
    public PatientAdmission[] getProviderAdmissions(JMeadowsQuery queryBean) throws SOAPException {

        log("getProviderAdmissions", queryBean);

        PatientAdmission[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderAdmissions(queryBean).toArray(new PatientAdmission[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public PatientAppointments[] getProviderAppointments(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderAppointments", queryBean);

        PatientAppointments[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderAppointments(queryBean).toArray(new PatientAppointments[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Consult[] getProviderConsultsReceived(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderConsultsReceived", queryBean);

        Consult[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderConsultsReceived(queryBean).toArray(new Consult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Consult[] getProviderConsultsRequested(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderConsultsRequested", queryBean);

        Consult[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderConsultsRequested(queryBean).toArray(new Consult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public LabResult[] getProviderLabAbnormalResults(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderLabAbnormalResults", queryBean);

        LabResult[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderLabAbnormalResults(queryBean).toArray(new LabResult[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Order[] getProviderOrdersPending(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderOrdersPending", queryBean);

        Order[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderOrdersPending(queryBean).toArray(new Order[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public Order[] getProviderOrdersResulted(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderOrdersResulted", queryBean);

        Order[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderOrdersResulted(queryBean).toArray(new Order[0]);
        } catch (Exception e) {
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public ProgressNote[] getProviderUnsignedNotes(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderUnsignedNotes", queryBean);

        ProgressNote[] returnVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            returnVal = port.getProviderUnsignedNotes(queryBean).toArray(new ProgressNote[0]);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return returnVal;
    }

    @Override
    public boolean setJanusGUIConfig(JMeadowsQuery queryBean, String cfg) throws SOAPException {
        log("setJanusGUIConfig", queryBean);

        boolean bRet = false;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            bRet = port.setJanusGUIConfig(queryBean, cfg);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return bRet;
    }

    @Override
    public String getJanusGUIConfig(JMeadowsQuery queryBean) throws SOAPException {
        log("getJanusGUIConfig", queryBean);

        String retVal = "";
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getJanusGUIConfig(queryBean);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

    @Override
    public boolean setProviderFlags(JMeadowsQuery queryBean, String flags) throws SOAPException {
        log("setProviderFlags", queryBean);

        boolean bRet = false;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            bRet = port.setProviderFlags(queryBean, flags);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return bRet;
    }

    @Override
    public String getProviderFlags(JMeadowsQuery queryBean) throws SOAPException {
        log("getProviderFlags", queryBean);

        String retVal = "";
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getProviderFlags(queryBean);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

//    @Override
//    public IehrUserProfile getIehrUserProfile(String smartCardID, String smartCardAgency) throws SOAPException {
//        return getIehrUserProfile(smartCardID, smartCardAgency, "");
//    }

    @Override
    public IehrUserProfile getIehrUserProfile(String smartCardID, String smartCardAgency, String emailAddress) throws SOAPException {
        IehrUserProfile retVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getIehrUserProfile(smartCardID, smartCardAgency, emailAddress);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

    @Override
    public boolean setIehrUserProfile(IehrUserProfile userProfile) throws SOAPException {
        boolean bRet = false;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            bRet = port.setIehrUserProfile(userProfile);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return bRet;
    }

    @Override
    public String getAppName() {
        return appName;
    }

    @Override
    public void setAppName(String appName) {
        this.appName = appName;
    }

    @Override
    public List<Site> getSites() throws SOAPException {
        List<Site> retVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getSites();
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

    @Override
    public List<AuditEntry> getLoginInfo(String cardID, String emailAddress) throws SOAPException {
        List<AuditEntry> retVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getLoginInfo(cardID, emailAddress);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

    @Override
    public String getServiceErrors() throws SOAPException {
        String retVal = null;
        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getServiceErrors();
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }
        return retVal;
    }

    @Override
    public AuthUserInfo getAuthUser(AuthUserInfo authUserInfoQuery) throws SOAPException {
        AuthUserInfo retVal = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            retVal = port.getAuthUser(authUserInfoQuery);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }

        return retVal;
    }

    @Override
    public List<Patient> getRecentlyViewedPatients(String userId) throws SOAPException {
        List<Patient> patients = null;

        try {
            JMeadowsData port = getJMeadowsDataPort();
            patients = port.getRecentlyViewedPatients(userId);
        }
        catch (Exception e) {
            logger.error(e.getMessage(), e);
            throw new SOAPException(e);
        }

        return patients;
    }
}
