/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */
 
 
package gov.va.med.rpc;

import gov.va.med.foundations.utilities.ARMLog;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 *
 *
 */
public class Timestamp {
    private static final int YRIDX = 0;
    private static final int YRLTH = 3;
    private static final int MOIDX = 3;
    private static final int MOLTH = 2;
    private static final int DYIDX = 5;
    private static final int DYLTH = 2;
    private static final int DOTIDX = 7;
    private static final int DOTLTH = 1;
    private static final int HRIDX = 8;
    private static final int HRLTH = 2;
    private static final int MNIDX = 10;
    private static final int MNLTH = 2;
    private static final int SCIDX = 12;
//    private static final int SCLTH = 2;

    private static final String JULIAN_DATE_REGEX_CYYMMDD_GROUP = "^(\\d\\d\\d(?:0[1-9]|1[0-2])(?:0[1-9]|[12][0-9]|3[01]))";


    private static final Logger LOGGER = LoggerFactory.getLogger(Timestamp.class);


    public static String getDatePart(String tstamp) {
        return StringUtils.CSubstring(padZeroes(tstamp), YRIDX, YRLTH);
    }

    public static String getTimePart(String tstamp) {
        return padZeroes(tstamp).substring(HRIDX);
    }

    public static int getYear(String tstamp) {
        return Integer.parseInt(StringUtils.CSubstring(padZeroes(tstamp), YRIDX, YRLTH)) + 1700;
    }

    public static int getMonth(String tstamp) {
        return Integer.parseInt(StringUtils.CSubstring(padZeroes(tstamp), MOIDX, MOLTH));
    }

    public static int getDay(String tstamp) {
        return Integer.parseInt(StringUtils.CSubstring(padZeroes(tstamp), DYIDX, DYLTH));
    }

    public static int getHour(String tstamp) {
        return Integer.parseInt(StringUtils.CSubstring(padZeroes(tstamp), HRIDX, HRLTH));
    }

    public static int getMinute(String tstamp) {
        return Integer.parseInt(StringUtils.CSubstring(padZeroes(tstamp), MNIDX, MNLTH));
    }

    public static int getSecond(String tstamp) {
        return Integer.parseInt(padZeroes(tstamp).substring(SCIDX));
    }

    public static String padZeroes(String intstamp) {
        String tstamp = intstamp;
        String rtc;
        if ("*SENSITIVE*".equals(tstamp)) {
            rtc = tstamp;
        } else {
            if (tstamp.indexOf('@') != -1) {
                tstamp = tstamp.replace('@', '.');
            }
            StringBuilder padded = new StringBuilder();
            if (tstamp.indexOf('.') != -1) {
                String[] parts = StringUtils.split(tstamp, ".");
                padded.append(parts[0]);
                padded.append(StringUtils.CSubstring("0000000", 0, 7 - parts[0].length()));
                padded.append('.');
                padded.append(parts[1]);
                padded.append(StringUtils.CSubstring("000000", 0, 6 - parts[1].length()));
            } else {
                padded.append(tstamp);
                padded.append(StringUtils.CSubstring("0000000", 0, 7 - tstamp.length()));
                padded.append(".000000");
            }
            rtc = padded.toString();
        }
        return rtc;
    }

    public static boolean isValid(String ints) {
        boolean rtc = true;
        String tstamp = ints;
        tstamp = padZeroes(tstamp);
        int year = getYear(tstamp);
        if (year < 1900) {
            rtc = false;
        }
        int mon = getMonth(tstamp);
        if (mon > 12) {
            rtc = false;
        }
        int day = getDay(tstamp);
        if (mon > 0 && day > 0 && !DateUtils.isValidDay(year, mon, day)) {
            rtc = false;
        }
        if (getHour(tstamp) > 23) {
            rtc = false;
        }
        if (getMinute(tstamp) > 59) {
            rtc = false;
        }
        if (getSecond(tstamp) > 59) {
            rtc = false;
        }
        return rtc;
    }

    public static String corrected(String ints) {
        String tstamp = ints;
        String rtc;
        if ("*SENSITIVE*".equals(tstamp))  {
            rtc = tstamp;
        } else {
            tstamp = padZeroes(tstamp);
            int year = getYear(tstamp);
            if (year < 1900) {
                rtc = "";
            } else {
                StringBuilder correctedStr = new StringBuilder();
                int mon = getMonth(tstamp);
                if (mon == -1 || mon > 12) {
                    correctedStr.append(StringUtils.CSubstring(tstamp, YRIDX, YRLTH));
                    correctedStr.append("00");
                    correctedStr.append(tstamp.substring(DYIDX));
                    tstamp = correctedStr.toString();
                    mon = 0;
                }
                int day = getDay(tstamp);
                if (day == -1 || !DateUtils.isValidDay(year, mon, day)) {
                    correctedStr.setLength(0);
                    correctedStr.append(StringUtils.CSubstring(tstamp, YRIDX, YRLTH + MOLTH));
                    correctedStr.append("00");
                    correctedStr.append(tstamp.substring(DOTIDX));
                    tstamp = correctedStr.toString();
                }
                int hour = getHour(tstamp);
                if (hour == -1 || hour > 23) {
                    correctedStr.setLength(0);
                    correctedStr.append(StringUtils.CSubstring(tstamp, YRIDX, YRLTH + MOLTH + DYLTH + DOTLTH));
                    correctedStr.append("00");
                    correctedStr.append(tstamp.substring(MNIDX));
                    tstamp = correctedStr.toString();
                }
                int min = getMinute(tstamp);
                if (min == -1 || min > 59) {
                    correctedStr.setLength(0);
                    correctedStr.append(StringUtils.CSubstring(tstamp, YRIDX, YRLTH + MOLTH + DYLTH + DOTLTH + HRLTH));
                    correctedStr.append("00");
                    correctedStr.append(tstamp.substring(SCIDX));
                    tstamp = correctedStr.toString();
                }
                int sec = getSecond(tstamp);
                if (sec == -1 || sec > 59) {
                    correctedStr.setLength(0);
                    correctedStr.append(StringUtils.CSubstring(tstamp, YRIDX, YRLTH + MOLTH + DYLTH + DOTLTH + HRLTH + MNLTH));
                    correctedStr.append("00");
                }
                rtc = tstamp;
            }
        }
        return rtc;
    }

    public static String toUtcString(String vistaTS) {
        String rtc;
        if ("*SENSITIVE*".equals(vistaTS)) {
            rtc = vistaTS;
        } else {
            String correctedTS = corrected(vistaTS);
            if ("".equals(correctedTS)) {
                rtc = "";
            } else {
                int year = getYear(correctedTS);
                StringBuilder result = new StringBuilder();
                result.append(Integer.toString(year));
                result.append(correctedTS.substring(MOIDX));
                rtc = result.toString();
            }
        }
        return rtc;
    }

    public static Calendar toDateTimeFromRdv(String rdvTS) throws RPCException {
        Calendar rtc;
        String[] parts = StringUtils.split(rdvTS, StringUtils.SPACE);
        String[] dateFlds = StringUtils.split(parts[0], StringUtils.SLASH);

        // TBD - should we do this? Technically '00' is an invalid value for the day and should be fixed at that VistA
        // sometimes we get '00' as the day value from VistA - we change this to '01' to create our DateTime object
        if ("0".equals(dateFlds[1]) || "00".equals(dateFlds[1])) {
            dateFlds[1] = "01";
        }

        if (parts.length == 1) {
            rtc = new GregorianCalendar(
                //year
                Integer.parseInt(dateFlds[2]),
                //month
                Integer.parseInt(dateFlds[0]) - 1,
                //day
                Integer.parseInt(dateFlds[1]),
                //hour
                0,
                //minute
                0,
                //second
                0);
        } else {
            String[] timeFlds = StringUtils.split(parts[1], StringUtils.COLON);
            try {
                rtc = new GregorianCalendar(
                    //year
                    Integer.parseInt(dateFlds[2]),
                    //month
                    Integer.parseInt(dateFlds[0]) - 1,
                    //day
                    Integer.parseInt(dateFlds[1]),
                    //hour
                    Integer.parseInt(timeFlds[0]),
                    //minute
                    Integer.parseInt(timeFlds[1]),
                    //second
                    0);
            } catch (Exception e) {
                if ("24".equals(timeFlds[0]) && "00".equals(timeFlds[1])) {
                    timeFlds[0] = "00";

                    rtc = new GregorianCalendar(
                            //year
                            Integer.parseInt(dateFlds[2]),
                            //month
                            Integer.parseInt(dateFlds[0]) - 1,
                            //day
                            Integer.parseInt(dateFlds[1]),
                            //hour
                            Integer.parseInt(timeFlds[0]),
                            //minute
                            Integer.parseInt(timeFlds[1]),
                            //second
                            0);
                } else {
                    throw new RPCException(e);
                }
            }
        }
        return rtc;
    }

    /// <summary>
    /// Pad a month or day string with zeroes
    /// </summary>
    /// <param name="part">Month or day as string (e.g. "03" or "2" or "28" etc.)</param>
    /// <returns>Padded string using zero. Always returns a string of length 2</returns>
    private static String padDatePartWithZeroes(String part) {
        String rtc;
        if (part == null || "".equals(part)) {
            rtc = "00";
        } else {
            StringBuilder padStr = new StringBuilder();
            padStr.append(part.trim());
            if (padStr.length() == 1) {
                padStr.insert(0, '0');
            }
            rtc = padStr.toString();
        }
        return rtc;
    }

    public static String toUtcFromRdv(String rdvTS) {
        String rtc;
        if (StringUtils.isEmpty(rdvTS)) {
            rtc = "";
        } else {
            String[] parts = StringUtils.split(rdvTS, StringUtils.SPACE);
            String[] dateFlds = null;

            if (parts != null && parts.length > 0) {
                dateFlds = StringUtils.split(parts[0], StringUtils.SLASH);
            }

            if (dateFlds == null || dateFlds.length < 3) {
                rtc = "";
            } else {
    
                StringBuilder datePartyyyyMMdd = new StringBuilder();
                datePartyyyyMMdd.append(dateFlds[2]);
                datePartyyyyMMdd.append(padDatePartWithZeroes(dateFlds[0]));
                datePartyyyyMMdd.append(padDatePartWithZeroes(dateFlds[1]));
                if (parts.length == 1) {
                    datePartyyyyMMdd.append(".000000");
                    rtc = datePartyyyyMMdd.toString();
                } else {
                    String[] timeFlds = StringUtils.split(parts[1], StringUtils.COLON);
    
                    if (timeFlds == null || timeFlds.length < 2) {
                        rtc = "";
                    } else {
                        datePartyyyyMMdd.append(".");
                        datePartyyyyMMdd.append(timeFlds[0]);
                        datePartyyyyMMdd.append(timeFlds[1]);
                        datePartyyyyMMdd.append("00");
                        rtc = datePartyyyyMMdd.toString();
                        //return new DateTime(Convert.ToInt16(dateFlds[2]),
                        //                    Convert.ToInt16(dateFlds[0]),
                        //                    Convert.ToInt16(dateFlds[1]),
                        //                    Convert.ToInt16(timeFlds[0]),
                        //                    Convert.ToInt16(timeFlds[1]),
                        //                    0);
                    }
                }
            }
        }
        return rtc;
    }

    public static Calendar toDateTimeFromTextTS(String textTS) {
        return new GregorianCalendar(
                //year
                Integer.parseInt(StringUtils.CSubstring(textTS, 0, 4)),
                //month
                Integer.parseInt(StringUtils.CSubstring(textTS, 4, 2)) - 1,
                //day
                Integer.parseInt(StringUtils.CSubstring(textTS, 6, 2)),
                //hour
                Integer.parseInt(StringUtils.CSubstring(textTS, 9, 2)),
                //minute
                Integer.parseInt(StringUtils.CSubstring(textTS, 11, 2)),
                //second
                Integer.parseInt(StringUtils.CSubstring(textTS, 13, 2)));
    }

    public static Calendar toDateTime(String vistaTS) {
        Calendar rtc;
        try {
            if (vistaTS.indexOf("SENSITIVE") != -1) {
                rtc = Calendar.getInstance();
            } else  if ("TODAY".equals(vistaTS)) {
                rtc = Calendar.getInstance();
            } else if ("".equals(vistaTS)) {
                rtc = null;
            } else {
                String correctedTS = corrected(vistaTS);
    
                if (getMonth(correctedTS) == 0 || getDay(correctedTS) == 0) {
                    StringBuilder warnMsg = new StringBuilder();
                    warnMsg.append(":: Bad VistA Timestamp detected: ");
                    warnMsg.append(vistaTS);
                    warnMsg.append("\n");
                    LOGGER.warn(ARMLog.cleanString(warnMsg.toString()));
                    rtc = null;
                } else {
                    rtc = new GregorianCalendar(
                            getYear(correctedTS),
                            getMonth(correctedTS) - 1,
                            getDay(correctedTS),
                            getHour(correctedTS),
                            getMinute(correctedTS),
                            getSecond(correctedTS));
                }
            }
        } catch (NumberFormatException ex) {
            StringBuilder warnMsg = new StringBuilder();
            warnMsg.append(":: Bad VistA Timestamp detected: ");
            warnMsg.append(vistaTS);
            warnMsg.append("\n");
            warnMsg.append(ex.getMessage());
            LOGGER.warn(ARMLog.cleanString(warnMsg.toString()), ex);
            rtc = null;
        }
        return rtc;
    }

    public static String fromUtcString(String tststamp) {
        //ts = corrected(ts);
        //ts = ts.Substring(0,ts.Length-3);
        int year = Integer.parseInt(StringUtils.CSubstring(tststamp, 0, 4));
        StringBuilder result = new StringBuilder();
        result.append(Integer.toString(year - 1700));
        result.append(tststamp.substring(4));
        return result.toString();
    }

    /// <summary>Convert DateTime into VistA format
    /// </summary>
    /// <param name="t"></param>
    /// <returns>yyyMMdd.HHmmss with year - 1700</returns>
    public static String fromDateTime(Calendar time) {
        String sYear = Integer.toString(time.get(Calendar.YEAR) - 1700);
        String sMonth = StringUtils.prependChars(Integer.toString(time.get(Calendar.MONTH) + 1), '0', 2);
        String sDay = StringUtils.prependChars(Integer.toString(time.get(Calendar.DAY_OF_MONTH)), '0', 2);
        StringBuilder result = new StringBuilder();
        result.append(sYear);
        result.append(sMonth);
        result.append(sDay);
        result.append('.');
        String sHour = StringUtils.prependChars(Integer.toString(time.get(Calendar.HOUR_OF_DAY)), '0', 2);
        String sMins = StringUtils.prependChars(Integer.toString(time.get(Calendar.MINUTE)), '0', 2);
        String sSecs = StringUtils.prependChars(Integer.toString(time.get(Calendar.SECOND)), '0', 2);
        result.append(sHour);
        result.append(sMins);
        result.append(sSecs);
        return result.toString();
    }

    public static String fromDateTimeToTextTS(Calendar time) {
        String sYear = Integer.toString(time.get(Calendar.YEAR));
        String sMonth = StringUtils.prependChars(Integer.toString(time.get(Calendar.MONTH) + 1), '0', 2);
        String sDay = StringUtils.prependChars(Integer.toString(time.get(Calendar.DAY_OF_MONTH)), '0', 2);
        StringBuilder result = new StringBuilder();
        result.append(sYear);
        result.append(sMonth);
        result.append(sDay);
        result.append('.');
        String sHour = StringUtils.prependChars(Integer.toString(time.get(Calendar.HOUR_OF_DAY)), '0', 2);
        String sMins = StringUtils.prependChars(Integer.toString(time.get(Calendar.MINUTE)), '0', 2);
        String sSecs = StringUtils.prependChars(Integer.toString(time.get(Calendar.SECOND)), '0', 2);
        result.append(sHour);
        result.append(sMins);
        result.append(sSecs);
        return result.toString();
    }

    public static String noSeconds(String invistaTS) {
        String rtc;
        String vistaTS = invistaTS;
        if (vistaTS.indexOf('.') == -1) {
            rtc = vistaTS;
        } else {
            vistaTS = padTimeToSeconds(vistaTS);
            String[] flds = StringUtils.split(vistaTS, StringUtils.PERIOD);
            StringBuilder returnStr = new StringBuilder();
            returnStr.append(flds[0]);
            returnStr.append('.');
            returnStr.append(StringUtils.CSubstring(flds[1], 0, 4));
            rtc = returnStr.toString();
        }
        return rtc;
    }

    public static String padTimeToSeconds(String vistaTS) {
        StringBuilder resultSB = new StringBuilder(vistaTS);
        while (resultSB.length() < 12) {
            resultSB.append('0');
        }
        return resultSB.toString();
    }

    public static String fromUtcFromDate(String inutcFrom) {
        StringBuffer utcFrom = new StringBuffer();
        utcFrom.append(inutcFrom);
        if (inutcFrom.indexOf('.') == -1) {
            utcFrom.append(".000000");
        }
        return fromUtcString(utcFrom.toString());
    }

    public static String fromUtcToDate(String inutcTo) {
        StringBuffer utcTo = new StringBuffer();
        utcTo.append(inutcTo);
        if (inutcTo.indexOf('.') == -1) {
            utcTo.append(".235959");
        }
        return fromUtcString(utcTo.toString());
    }

    public static String invertTimestamp(String tstamp) {
        Double dTs = Double.parseDouble(tstamp);
        Double newTs = 9999999 - dTs;
        return newTs.toString();
    }

    public static String trimZeroes(String tstamp) {
        int idx = tstamp.length() - 1;
        while (tstamp.charAt(idx) == '0') {
            idx--;
        }
        return StringUtils.CSubstring(tstamp, 0, idx + 1);
    }

    public static String fromText(String textTS) {
        if (textTS == null || "".equals(textTS)) {
            throw new IllegalArgumentException();
        }
        boolean fHasTime = true;
        int validLength = 18;
        if (textTS.indexOf('@') == -1) {
            fHasTime = false;
            validLength = 12;
        }
        if (textTS.length() != validLength) {
            throw new IllegalArgumentException("Invalid text date");
        }
        String mon = txtMonthToNum(StringUtils.CSubstring(textTS, 0, 3));
        String day = StringUtils.CSubstring(textTS, 4, 2);
        String year = StringUtils.CSubstring(textTS, 8, 4);
        if (!DateUtils.isValidDay(Integer.parseInt(year), Integer.parseInt(mon), Integer.parseInt(day))) {
            throw new IllegalArgumentException("Invalid text date");
        }
        String hour = "00";
        String min = "00";
        if (fHasTime) {
            hour = StringUtils.CSubstring(textTS, 13, 2);
            min = StringUtils.CSubstring(textTS, 16, 2);
        }
        StringBuilder returnStr = new StringBuilder();
        returnStr.append(year);
        returnStr.append(mon);
        returnStr.append(day);
        returnStr.append('.');
        returnStr.append(hour);
        returnStr.append(min);
        returnStr.append("00");
        return returnStr.toString();
    }

    public static String txtMonthToNum(String insMonth) {
        String sMonth = insMonth.toUpperCase(Locale.getDefault());
        String rtc;
        if ("JAN".equals(sMonth)) {
            rtc = "01";
        } else if ("FEB".equals(sMonth)) {
            rtc = "02";
        } else if ("MAR".equals(sMonth)) {
            rtc = "03";
        } else if ("APR".equals(sMonth)) {
            rtc = "04";
        } else if ("MAY".equals(sMonth)) {
            rtc = "05";
        } else if ("JUN".equals(sMonth)) {
            rtc = "06";
        } else if ("JUL".equals(sMonth)) {
            rtc = "07";
        } else if ("AUG".equals(sMonth)) {
            rtc = "08";
        } else if ("SEP".equals(sMonth)) {
            rtc = "09";
        } else if ("OCT".equals(sMonth)) {
            rtc = "10";
        } else if ("NOV".equals(sMonth)) {
            rtc = "11";
        } else if ("DEC".equals(sMonth)) {
            rtc = "12";
        } else {
            throw new IllegalArgumentException("Invalid month");
        }
        return rtc;
    }

    public static String txtNumToMonth(int month) {
        String rtc = null;
        switch (month) {
            case 1:  rtc = "Jan";
                break;
            case 2:  rtc = "Feb";
                break;
            case 3:  rtc = "Mar";
                break;
            case 4:  rtc = "Apr";
                break;
            case 5:  rtc = "May";
                break;
            case 6:  rtc = "Jun";
                break;
            case 7:  rtc = "Jul";
                break;
            case 8:  rtc = "Aug";
                break;
            case 9:  rtc = "Sep";
                break;
            case 10: rtc = "Oct";
                break;
            case 11: rtc = "Nov";
                break;
            case 12: rtc = "Dec";
                break;
            default:
                throw new IllegalArgumentException("Invalid month");
        }
        return rtc;
    }

    /**
     * If the julianDateString starts with a Julian date in the format CYYMMDD then this method will return that date
     * and trim off the time. If the julianDateString does not start with a valid Julian date in that format, then this
     * method will return <code>null</code>.
     *
     * For example "3101204.1234" will return "3101204" while "3102132" will return null.
     *
     * @param julianDateString the string to extract the Julian date from
     * @return the Julian date or null if the parameter string does not start with a Julian date in the proper format.
     */
    public static String getJulianCYYMMDD (String julianDateString) {
        Pattern julianPattern = Pattern.compile(JULIAN_DATE_REGEX_CYYMMDD_GROUP);
        Matcher julianMatcher = julianPattern.matcher(julianDateString.trim());

        String trimmedDate = null;
        if (julianMatcher.find()) {
            trimmedDate = julianMatcher.group();
        }

        return trimmedDate;
    }

    /**
     * ***************************************************************
     * Name:	formatDisplayDateNoBadMonthOrDay
     * Purpose:	formats a Date MMM dd, yyyy. Modified from Utils.formatDisplayDate(Calendar cal)
     *          used primarily by JLV DAOs. This will format the date from a calendar to MMM dd, yyyy, or MMM, yyyy if the
     *          day is not specified (i.e. Calendar.DAY_OF_MONTH == 0), or yyyy if the month is not specified (i.e.
     *          Calendar.MONTH == -1).
     * Inputs:	vistaTS (Julian type date)
     * Returns:	formatted date
     */
    public static String formatDisplayDateNoBadMonthOrDay(String vistaTS) {
        String rtc = vistaTS;
            if (vistaTS.indexOf("SENSITIVE") != -1) {
            } else  if ("TODAY".equals(vistaTS)) {
            } else if ("".equals(vistaTS)) {
                rtc = null;
            } else {
                String correctedTS = corrected(vistaTS);

                StringBuilder formattedDateSB = new StringBuilder();
                formattedDateSB.append(Timestamp.getYear(correctedTS));
                int month = Timestamp.getMonth(correctedTS);
                if (month == 0) {
                    rtc = formattedDateSB.toString();
                } else {
                    formattedDateSB.insert(0, ", ");
                    formattedDateSB.insert(0, Timestamp.txtNumToMonth(month));
                }
                int day = Timestamp.getDay(correctedTS);
                if (day > 0) {
                    formattedDateSB.insert(3, padDatePartWithZeroes(Integer.toString(day)));
                    formattedDateSB.insert(3, " ");
                }
                rtc = formattedDateSB.toString();
            }
        return rtc;
    }


}
