/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.vistabeans.AdmissionDetail;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.Encounter;
import gov.va.med.vds.vistabeans.EncounterDocument;
import gov.va.med.vds.vistabeans.PatientAdmission;
import gov.va.med.DNS   .common.Utils;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class AdmissionsRPC extends BaseRPC {

    private static final Logger LOGGER = LoggerFactory.getLogger(AdmissionsRPC.class);
    private static final String SPECIALTY_MARKER = "Last Tr Specialty:";
    private static final String PROVIDER_MARKER = "Last Prov:";
    private static final String ADMITTING_DIAGNOSIS_MARKER = "Admitting Diagnosis:";
    private static final String DATE_AT_START_OF_INPUT_REGEX = "\\A(0[1-9]|1[012])[- /.](0[1-9]|[12][0-9]|3[01])[- /.](19|20)\\d\\d.*";
    private static final String DATE_REGEX = "(0[1-9]|1[012])[- /.](0[1-9]|[12][0-9]|3[01])[- /.](19|20)\\d\\d";
    private static final String END_MARKER = "* END";


    public AdmissionsRPC(RPCConnection cxn) {
        super(cxn);
    }

//    private void logErrorMessage(Exception e) {
//        LOGGER.error(e.getMessage());
//    }

    //SB: Missing DCDate, RegNumber, User, Diagnosis
    public PatientAdmission[] getAdmissions(Calendar startDate, Calendar endDate) throws RPCException {
        return getAdmissions(cxn.getPatientIen(), startDate, endDate);
    }

    public PatientAdmission[] getAdmissions(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetAdmissionsRequest(patientIen);
        String response = (String) cxn.query(request);
        PatientAdmission[] adms = toInpatientStays(response/*, patientIen*/);
        ArrayList<PatientAdmission> admFiltered = new ArrayList<PatientAdmission>();
        //filter by date range
        for (PatientAdmission adm : adms) {
            Calendar admDate = adm.getAdmissionDate();
            if (admDate.compareTo(startDate) >= 0 && admDate.compareTo(endDate) <= 0) {
                admFiltered.add(adm);
            }
        }

        request = buildAdmissionDetailReportRequest(patientIen, startDate, endDate);
        String admissionReport = (String) cxn.query(request);
        PatientAdmission[] admissionsArray = admFiltered.toArray(new PatientAdmission[0]);
        // add in details from details report
        parseAdmissionsDetail(admissionsArray, admissionReport);

        return admissionsArray;
    }

    public Encounter[] getAdmissionsEncounters(Calendar startDate, Calendar endDate) throws RPCException {
        return getAdmissionsEncounters(cxn.getPatientIen(), startDate, endDate);
    }

    public Encounter[] getAdmissionsEncounters(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
//        PatientAdmission[] admissionsArray = new PatientAdmission[0];
        EncountersRPC encountersRPC = new EncountersRPC(cxn);
        ArrayList<Encounter> admissionsFromEncountersList = new ArrayList<Encounter>(Arrays.asList(encountersRPC.getPatientAdmissionsFromEncounters(patientIen, startDate, endDate)));
//        addDataFromEncountersToAdmissions(admissionsArray, admissionsFromEncountersList);

        return admissionsFromEncountersList.toArray(new Encounter[0]);
    }


    private String buildGetAdmissionsRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWPT ADMITLST");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        return vquery.buildMessage();
    }

    private PatientAdmission[] toInpatientStays(String response/*, String patientIen*/) {
        PatientAdmission[] rtc;
        if ("".equals(response)) {
            rtc = new PatientAdmission[0];
        } else {
            String[] rex = StringUtils.split(response, StringUtils.CRLF);
            rex = StringUtils.trimArray(rex);
            PatientAdmission[] stays = new PatientAdmission[rex.length];
    
            for (int i = 0; i < rex.length; i++) {
                String[] flds = StringUtils.split(rex[i], StringUtils.CARET);
                stays[i] = new PatientAdmission();
                fillCommonDataBeanValues(stays[i]);
                stays[i].setAdmissionDate(Utils.getCal(Timestamp.toUtcString(flds[0])));
                stays[i].setWard(flds[2]);
    
                String dischargeSummaryIen = flds[5];
                if (!"0".equals(dischargeSummaryIen)) {
                    stays[i].setDischargeSummaryId(dischargeSummaryIen);
                }
                /*
                stays[i].Patient = new Patient();
                stays[i].Patient.LocalPid = patientIen;
                stays[i].AdmitTimestamp = VistaTimestamp.toUtcString(flds[0]);
                stays[i].Location = new HospitalLocation(flds[1], flds[2]);
                stays[i].Type = flds[3];
                 */
            }
            rtc = stays;
        }
        return rtc;
    }

    public AdmissionDetail getAdmissionDetailsReport(Calendar startDate, Calendar endDate) throws RPCException {
        return getAdmissionDetailsReport(cxn.getPatientIen(), startDate, endDate);
    }

    public AdmissionDetail getAdmissionDetailsReport(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildAdmissionDetailReportRequest(patientIen, startDate, endDate);
        String response = (String) cxn.query(request);

        AdmissionDetail admDetail = new AdmissionDetail();
        fillCommonDataBeanValues(admDetail);
        admDetail.setDetailText(response);

        return admDetail;
    }

    private String buildAdmissionDetailReportRequest(String patientIen, Calendar startDate, Calendar endDate) {

        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, "OR_ADC:ADM./DISCHARGE~;;10;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        String startDateString = Timestamp.fromDateTime(startDate);
        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
        startDateString = flds[0];
        String endDateString = Timestamp.fromDateTime(endDate);
        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
        endDateString = flds[0];

        vquery.addParameter(VistaQuery.LITERAL, startDateString);
        vquery.addParameter(VistaQuery.LITERAL, endDateString);
        return vquery.buildMessage();

    }

    public AdmissionDetail getExpandedAdmissionDetailsReport(Calendar startDate, Calendar endDate) throws RPCException {
        return getExpandedAdmissionDetailsReport(cxn.getPatientIen(), startDate, endDate);
    }

    public AdmissionDetail getExpandedAdmissionDetailsReport(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildExpandedAdmissionDetailReportRequest(patientIen, startDate, endDate);
        String response = (String) cxn.query(request);

        AdmissionDetail admDetail = new AdmissionDetail();
        fillCommonDataBeanValues(admDetail);
        admDetail.setDetailText(response);

        return admDetail;
    }

    private String buildExpandedAdmissionDetailReportRequest(String patientIen, Calendar startDate, Calendar endDate) {

        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
//        vquery.addParameter(VistaQuery.LITERAL, "OR_EADT:EXPANDED ADT~;;64;999");
        vquery.addParameter(VistaQuery.LITERAL, "OR_EADT:EXPANDED ADT~;;67;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        String startDateString = Timestamp.fromDateTime(startDate);
        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
        startDateString = flds[0];
        String endDateString = Timestamp.fromDateTime(endDate);
        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
        endDateString = flds[0];

        vquery.addParameter(VistaQuery.LITERAL, startDateString);
        vquery.addParameter(VistaQuery.LITERAL, endDateString);
        return vquery.buildMessage();

    }

//    /**
//     * This method calculates the number of days until the given date from the current date.
//     * If the day given is prior to today, the result will be negative. If the number of days differe
//     *
//     * @param date the date to count days until
//     * @return the number of days until date. (negative if the date is prior to the current date)
//     */
//    private long daysFromNow(Calendar date) {
//        GregorianCalendar now = new GregorianCalendar();
//        long nowMs = now.getTimeInMillis();
//        long dateMs = date.getTimeInMillis();
//
//        long millisecondDifference = dateMs - nowMs;
//
//        // 1000 ms in a sec, 60 sec in a min, 60 min in an hour, 24 hr in a day
//        double days = millisecondDifference / 1000.0d / 60.0d / 60.0d / 24.0d;
//
//        return (long) (days + 0.5);
//
//    }

    /**
     * Builds the VPR RPC request string using "visits" as the type. The problems contain the list of
     * visits. The RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when
     * setting up the connection.
     *
     * @param patientIen Patient IEN
     * @param startDate  Start date of range
     * @param endDate    End date of range
     * @return The RPC response message string
     */
//    private String buildGetVisitsVPRRequest(String patientIen, Calendar startDate, Calendar endDate) {
//        return VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.VISITS);
//    }


    /**
     * Parses the report from
     *
     * @param admissions
     * @param admissionDetailReport
     */
    private void parseAdmissionsDetail(PatientAdmission[] admissions, String admissionDetailReport) {
        String nextLine = null;
        SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yyyy", Locale.getDefault());

        Pattern dateAtStartOfInputRegexPattern = Pattern.compile(DATE_AT_START_OF_INPUT_REGEX);
        Pattern dateAnywhereRegexPattern = Pattern.compile(DATE_REGEX);


        for (int i = 0; i < admissions.length; i++) {

            // Find the admission date that we are interested in

            if (admissions[i].getAdmissionDate() != null) {
                String admissionDateString = sdf.format(admissions[i].getAdmissionDate().getTime());
                int admStartIndex = admissionDetailReport.indexOf("\n" + admissionDateString + " - ") + 1;
                if (admStartIndex <= 0) {
                    continue;
                }

                int specialtyIndex = admissionDetailReport.indexOf(SPECIALTY_MARKER, admStartIndex);
                int providerIndex = admissionDetailReport.indexOf(PROVIDER_MARKER, specialtyIndex);
                int specialtyProviderEOLIndex = admissionDetailReport.indexOf("\n", providerIndex);
                int nextEOLIndex;

                String specialty = admissionDetailReport.substring(specialtyIndex + SPECIALTY_MARKER.length(), providerIndex).trim();
                String provider = admissionDetailReport.substring(providerIndex + PROVIDER_MARKER.length(), specialtyProviderEOLIndex).trim();
                admissions[i].setProvider(provider);
                admissions[i].setSpecialty(specialty);

                // Find the discharge date if there is one, this should be in the second date match of the first line
                nextEOLIndex = admissionDetailReport.indexOf("\n", admStartIndex);
                nextLine = admissionDetailReport.substring(admStartIndex, nextEOLIndex);
                Matcher datesInFirstLineOfAdm = dateAnywhereRegexPattern.matcher(nextLine);
                // Try to find a date after the admit date "MM/dd/yyyy".length = 10;
                if (datesInFirstLineOfAdm.find(10)) {
                    String dischargeDateString = datesInFirstLineOfAdm.group();
                    Date dischargeDate = null;
                    try {
                        dischargeDate = sdf.parse(dischargeDateString);
                    } catch (ParseException pe) {
                        // bad date, do nothing we'll leave the discharge date as null.
                        LOGGER.debug(pe.getMessage());
                    }
                    if (dischargeDate != null) {
                        Calendar dischargeCalendar = Calendar.getInstance();
                        dischargeCalendar.setTime(dischargeDate);
                        admissions[i].setDispositionDate(dischargeCalendar);
                    }
                }

                // Get the next line until it reaches the next start date. Find relevant info in the line if it
                // is still part of the same admission
                int startOfNextLineIndex = specialtyProviderEOLIndex + 1;
                while (!nextLine.contains(END_MARKER)) {
                    // use String.substring() to get next line
                    nextEOLIndex = admissionDetailReport.indexOf("\n", startOfNextLineIndex);
                    if (nextEOLIndex < 0) {
                        nextLine = "";
                    } else {
                        nextLine = admissionDetailReport.substring(startOfNextLineIndex, nextEOLIndex);
                    }
                    startOfNextLineIndex = nextEOLIndex + 1;

                    // use regex to determine if the line starts with a date, if so, that is the start of the next
                    // admission, so we need to break out of this loop
                    Matcher dateAtStartOfInputMatcher = dateAtStartOfInputRegexPattern.matcher(nextLine);
                    if (dateAtStartOfInputMatcher.matches()) {
                        break;
                    }

                    // Line is still part of the same admission, so we can look for other details of the admission
                    if (nextLine.contains(ADMITTING_DIAGNOSIS_MARKER)) {
                        // Find the "Admitting Diagnosis"
                        admissions[i].setDrgDescription(nextLine.substring(nextLine.indexOf(ADMITTING_DIAGNOSIS_MARKER) + ADMITTING_DIAGNOSIS_MARKER.length()).trim());
                    }
                }
            }
        }
    }

    public static void addDataFromEncountersToAdmissions(PatientAdmission[] admissionList, ArrayList<Encounter> encounterList) {
        for (PatientAdmission patientAdmission: admissionList) {

            Iterator<Encounter> encounterIterator = encounterList.iterator();
            if (!encounterIterator.hasNext()) {
                break;
            }
            while (encounterIterator.hasNext()) {
                Encounter nextEncounter = encounterIterator.next();

                int calendarDifference = nextEncounter.getArrivalDateTime().compareTo(patientAdmission.getAdmissionDate());
                if (calendarDifference == 0) {
                    if (patientAdmission.getDispositionDate() == null) {
                        patientAdmission.setDispositionDate(nextEncounter.getDepartureDateTime());
                    }
                    patientAdmission.setProvider(nextEncounter.getProvider());
                    patientAdmission.setAdmittingProvider(nextEncounter.getAdmittingProvider());
                    patientAdmission.setRoomBed(nextEncounter.getRoomBed());
                    patientAdmission.setSpecialty(nextEncounter.getSpecialty());
                    patientAdmission.setCodes(nextEncounter.getCodes());
                    patientAdmission.setAdmissionId(nextEncounter.getApptId());
                    patientAdmission.setWard(nextEncounter.getClinic());
                    patientAdmission.setCodes(nextEncounter.getCodes());
                    patientAdmission.setRegisterNumber(nextEncounter.getAdmissionRegistrationNum());
                    
                    Code reason = new Code();
                    String nextEncounterReasonCode = nextEncounter.getReasonCode();
                    reason.setSystem("ICD9");
                    reason.setDisplay(nextEncounter.getReason());
                    reason.setCode(nextEncounterReasonCode);
                    reason.setNarrative(nextEncounter.getReasonNarrative());
                    patientAdmission.setReason(reason);
                    patientAdmission.setDrgDescription(nextEncounter.getReasonNarrative());
                    if (nextEncounterReasonCode != null && nextEncounterReasonCode.length() > 0) {
                        ArrayList<Code> admCodes = null;
                        if (patientAdmission.getCodes() != null && patientAdmission.getCodes().length > 0) {
                            admCodes = new ArrayList<Code>(Arrays.asList(patientAdmission.getCodes()));
                        } else {
                            admCodes = new ArrayList<Code>();
                        }
                        admCodes.add(reason);
                        patientAdmission.setCodes(admCodes.toArray(new Code[0]));
                    }

                    EncounterDocument[] encounterDocuments = nextEncounter.getDocuments();
                    if (encounterDocuments != null) {
                        patientAdmission.setEncounterDocuments(Arrays.asList(encounterDocuments));
                    }

                    encounterIterator.remove();
                    break;
                }
            }
        }
    }

}
