/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Allergy;
import gov.va.med.vds.vistabeans.AllergyDetail;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.Comment;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import gov.va.med.DNS   .common.Utils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Allergy RPC data access object.
 */
public class AllergyRPC extends BaseRPC {
    private static final Logger LOGGER = LoggerFactory.getLogger(AllergyRPC.class);

    private static final String NO_KNOWN_ALLERGIES = "NO KNOWN ALLERGIES";
    private static final String NO_ALLERGY_ASSESSMENT = "NO ALLERGY ASSESSMENT";

    public AllergyRPC(RPCConnection cxn) {
        super(cxn);
    }

    /**
     * Retrieves a list of patient allergies.
     *
     * @return List of patient allergies.
     * @throws RPCException if there was an error..
     */
    public Allergy[] getAllergies() throws RPCException {
        return getAllergies(cxn.getPatientIen());
    }

    /**
     * Retrieves a list of patient allergies.
     *
     * @param patientIen Patient data file number (or IEN).
     * @return List of patient allergies.
     * @throws RPCException if there was an error..
     */
    public Allergy[] getAllergies(String patientIen) throws RPCException {
        String request = buildGetAllergiesRequest(patientIen);
        String response = (String) cxn.query(request);
        return toAllergies(response/*, patientIen*/);
    }

    /**
     * Retrieves details of an allergy.
     *
     * @param allergyIen Internal entry number of the patient allergy file.
     * @return AllergyDetail containing allergy report text.
     * @throws RPCException if there was an error..
     */
    public AllergyDetail getAllergyDetail(String allergyIen) throws RPCException {
        return getAllergyDetail(cxn.getPatientIen(), allergyIen);
    }

    /**
     * Retrieves patient allergy record details.
     *
     * @param patientIen Patient data file number (or IEN).
     * @param allergyIen Internal entry number of the patient allergy file.
     * @return AllergyDetail containing allergy report text.
     * @throws RPCException if there was an error..
     */
    public AllergyDetail getAllergyDetail(String patientIen, String allergyIen) throws RPCException {

        //build request
        VistaQuery vistaQuery = vistaQueryFactory.getVistaQuery("ORQQAL DETAIL");
        vistaQuery.addParameter(VistaQuery.LITERAL, patientIen);
        vistaQuery.addParameter(VistaQuery.LITERAL, allergyIen);
        String request = vistaQuery.buildMessage();

        String response = (String) cxn.query(request);

        AllergyDetail allergyDetail = new AllergyDetail();
        allergyDetail.setReportText(response);

        fillCommonDataBeanValues(allergyDetail);

        return allergyDetail;
    }

    private String buildGetAllergiesRequest(String patientIen) {
        VistaQuery vistaQuery = vistaQueryFactory.getVistaQuery("ORQQAL LIST");
        vistaQuery.addParameter(VistaQuery.LITERAL, patientIen);
        return vistaQuery.buildMessage();
    }

    private Allergy[] toAllergies(String response/*, String patientIen*/) {
        Allergy[] rtc;
        
        final int ALLERGY_IEN = 0;
        final int ALLERGY_NAME = 1;
        final int ALLERGY_SEVERITY = 2;
        final int ALLERGY_REACTION = 3;

        if ("".equals(response)) {
            rtc = new Allergy[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            List<Allergy> allergies = new ArrayList<Allergy>();
    
            for (String line : lines) {
                String[] pieces = StringUtils.split(line, "^");
    
                if (pieces.length < 1) {
                    continue;
                }
    
                Allergy allergy = new Allergy();
    
                //single array literal possible values:
                //  No Known Allergy Assessment or No Known Allergies
                if (pieces.length == 1) {
                    allergy.setAllergyName(pieces[0]);
                } else {
                    for (int i = 0; i < pieces.length; i++) {
                        String item = pieces[i];
    
                        //skip if piece item is null or empty
                        if (StringUtils.isEmpty(item)) {
                            continue;
                        }
    
                        switch (i) {
                            case ALLERGY_IEN:
                                allergy.setAllergyId(item);
                                break;
                            case ALLERGY_NAME:
                                allergy.setAllergyName(item);
                                break;
                            case ALLERGY_SEVERITY:
                                allergy.setSeverity(item);
                                break;
                            case ALLERGY_REACTION:
                                allergy.setComment(item);
                                break;
                            default:
                                LOGGER.error("Invalid length");
                                break;
                        }
                    }
    
                    fillCommonDataBeanValues(allergy);
                }
    
                allergies.add(allergy);
            }
    
            rtc = allergies.toArray(new Allergy[0]);
        }
        return rtc;
    }


    /**
     * Uses the VPR RPC with "allergies" as the type.
     *
     * @param startDate The start of the date range for the list of allergies
     * @param endDate   The end of the date range for the list of allergies
     * @return The list of allergies for the specified date range
     * @throws RPCException
     */
    public Allergy[] getAllergiesVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getAllergiesVPR(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "allergies" as the type.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of allergies
     * @param endDate    The end of the date range for the list of allergies
     * @return The list of allergies for the specified date range
     * @throws RPCException
     */
    public Allergy[] getAllergiesVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.REACTIONS);
        String response = (String) cxn.query(request);

        return parseAllergiesFromVPR(response);
    }

    private Allergy[] parseAllergiesFromVPR(String response) throws RPCException {
        Allergy[] rtc;
        if ("".equals(response)) {
            rtc = new Allergy[0];
        } else {

            StringBuilder sbld = new StringBuilder(Utils.stripInvalidXMLCharacters(response));
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");

            List<Allergy> allergyList = new ArrayList<Allergy>();
    
            final String XPATH_ALLERGY = "//allergy";

            Document doc = null;
    
            try {
                doc = parseXMLDocument(sbld.toString());
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
    
            XPathExpression exprAllergy;
    
            try {
                exprAllergy = xpath.compile(XPATH_ALLERGY);
    
            } catch (Exception e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            }
    
            NodeList allergiesNodeList = null;
            try {
                allergiesNodeList = (NodeList) exprAllergy.evaluate(doc, XPathConstants.NODESET);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            List<VPRAllergyParser> allergyParserList = new ArrayList<VPRAllergyParser>();
            int listLength = allergiesNodeList.getLength();
            for (int allergyNum = 0; allergyNum < listLength; allergyNum++) {
                Node allergyNode = allergiesNodeList.item(allergyNum);
                allergyNode.getParentNode().removeChild(allergyNode);
    
                allergyParserList.add(new VPRAllergyParser(allergyNode));
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(10);
            List<Future<Allergy>> futures = null;
            try {
                futures = executorService.invokeAll(allergyParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse medications results : accession node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<Allergy>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<Allergy> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Allergy Data");
                    } else {
                        allergyList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    LOGGER.error(e.getMessage(), e);
                } catch (ExecutionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
    
            rtc = allergyList.toArray(new Allergy[0]);
        }
        return rtc;
    }

    private class VPRAllergyParser implements Callable<Allergy> {
        private Node allergyNode;

        public VPRAllergyParser(Node allergyNode) {
            this.allergyNode = allergyNode;
        }

        public Allergy call() throws RPCException {
            final String XPATH_NAME = "./name/@value";
            final String XPATH_ID = "./id/@value";
            final String XPATH_SEVERITY = "./severity/@value";
            final String XPATH_VUID = "./vuid/@value";
            final String XPATH_ASSESSMENT = "./assessment/@value";
            final String XPATH_ENTERED = "./entered/@value";

            final String XPATH_REACTIONS = "./reactions";
            final String XPATH_REACTION = "./reaction";
            final String XPATH_REACTION_NAME = "./@name";
            final String XPATH_REACTION_VUID = "./@vuid";

            final String XPATH_DRUGCLASSES = "./drugClasses";
            final String XPATH_DRUGCLASS = "./drugClass";
            final String XPATH_DRUGCLASS_NAME = "./@name";
            final String XPATH_DRUGCLASS_VUID = "./@vuid";

            final String XPATH_DRUGINGREDIENTS = "./drugIngredients";
            final String XPATH_DRUGINGREDIENT = "./drugIngredient";
            final String XPATH_DRUGINGREDIENT_NAME = "./@name";
            final String XPATH_DRUGINGREDIENT_VUID = "./@vuid";

            final String XPATH_COMMENTS = "./comments";
            final String XPATH_COMMENT = "./comment";
            final String XPATH_COMMENT_TEXT_ATTR = "./@commentText";
            final String XPATH_ENTERED_BY_ATTR = "./@enteredBy";
            final String XPATH_ENTERED_ATTR = "./@entered";
            final String XPATH_COMMENT_TYPE = "./@commentType";

            Allergy allergy = new Allergy();

            XPathExpression exprName = null;
            XPathExpression exprAssessment = null;
            XPathExpression exprId = null;
            XPathExpression exprSeverity = null;
            XPathExpression exprReactions = null;
            XPathExpression exprReaction = null;
            XPathExpression exprReactionName = null;
            XPathExpression exprReactionVuid = null;
            XPathExpression exprEntered = null;
            XPathExpression exprDrugClasses = null;
            XPathExpression exprDrugClass = null;
            XPathExpression exprDrugClassName = null;
            XPathExpression exprDrugClassVuid = null;
            XPathExpression exprDrugIngredients = null;
            XPathExpression exprDrugIngredient = null;
            XPathExpression exprDrugIngredientName = null;
            XPathExpression exprDrugIngredientVuid = null;
            XPathExpression exprVuid = null;
                    
            XPathExpression exprComments = null;
            XPathExpression exprComment = null;
            XPathExpression exprCommentText = null;
            XPathExpression exprEnteredByAttr = null;
            XPathExpression exprEnteredAttr = null;
            XPathExpression exprCommentType = null;

            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();

            try {
                exprName = xpath.compile(XPATH_NAME);
                exprId = xpath.compile(XPATH_ID);
                exprSeverity = xpath.compile(XPATH_SEVERITY);
                exprVuid = xpath.compile(XPATH_VUID);
                exprAssessment = xpath.compile(XPATH_ASSESSMENT);
                exprEntered = xpath.compile(XPATH_ENTERED);

                exprReactions = xpath.compile(XPATH_REACTIONS);
                exprReaction = xpath.compile(XPATH_REACTION);
                exprReactionName = xpath.compile(XPATH_REACTION_NAME);
                exprReactionVuid = xpath.compile(XPATH_REACTION_VUID);

                exprDrugClasses = xpath.compile(XPATH_DRUGCLASSES);
                exprDrugClass = xpath.compile(XPATH_DRUGCLASS);
                exprDrugClassName = xpath.compile(XPATH_DRUGCLASS_NAME);
                exprDrugClassVuid = xpath.compile(XPATH_DRUGCLASS_VUID);

                exprDrugIngredients = xpath.compile(XPATH_DRUGINGREDIENTS);
                exprDrugIngredient = xpath.compile(XPATH_DRUGINGREDIENT);
                exprDrugIngredientName = xpath.compile(XPATH_DRUGINGREDIENT_NAME);
                exprDrugIngredientVuid = xpath.compile(XPATH_DRUGINGREDIENT_VUID);

                exprComments = xpath.compile(XPATH_COMMENTS);
                exprComment = xpath.compile(XPATH_COMMENT);
                exprCommentText = xpath.compile(XPATH_COMMENT_TEXT_ATTR);
                exprEnteredByAttr = xpath.compile(XPATH_ENTERED_BY_ATTR);
                exprEnteredAttr = xpath.compile(XPATH_ENTERED_ATTR);
                exprCommentType = xpath.compile(XPATH_COMMENT_TYPE);


            } catch (Exception e) {
                throw new RPCException(e);
            }

            String allergyName;
            String allergyIen;
            String allergySeverity;
            String allergyVuid;
            String allergyAssessment;

            try {
                allergyName = (String) exprName.evaluate(allergyNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                allergyName = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                allergyAssessment = (String) exprAssessment.evaluate(allergyNode, XPathConstants.STRING);

                if (StringUtils.isEmpty(allergyName) && !StringUtils.isEmpty(allergyAssessment))
                {
                    if ("not done".equalsIgnoreCase(allergyAssessment)) {
                        allergyName = NO_ALLERGY_ASSESSMENT;
                    } else if ("nka".equalsIgnoreCase(allergyAssessment)) {
                        allergyName = NO_KNOWN_ALLERGIES;
                    } else {
                        allergyName = allergyAssessment;
                    }
                }

            } catch (XPathExpressionException e) {
                allergyName = null;
                LOGGER.error(e.getMessage(), e);
            }

            try {
                allergyIen = (String) exprId.evaluate(allergyNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                allergyIen = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                allergySeverity = (String) exprSeverity.evaluate(allergyNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                allergySeverity = null;
                LOGGER.error(e.getMessage(), e);
            }
            try {
                allergyVuid = (String) exprVuid.evaluate(allergyNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                allergyVuid = null;
                LOGGER.error(e.getMessage(), e);
            }

            Node drugClassesNode = null;
            try {
                drugClassesNode = (Node) exprDrugClasses.evaluate(allergyNode, XPathConstants.NODE);
            } catch (XPathExpressionException e) {
                LOGGER.error(e.getMessage(), e);
            }
            if (drugClassesNode != null) {
                NodeList drugClassNodeList = null;
                try {
                    drugClassNodeList = (NodeList) exprDrugClass.evaluate(drugClassesNode, XPathConstants.NODESET);
                } catch (XPathExpressionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
                if (drugClassNodeList != null) {
                    String drugClassName = null;
                    String drugClassVuid = null;
                    List<Code> drugClassCodeList = new ArrayList<Code>();
                    int listLength = drugClassNodeList.getLength();
                    for (int drugClassNum = 0; drugClassNum < listLength; drugClassNum++) {
                        Node drugClassNode = drugClassNodeList.item(drugClassNum);
                        drugClassNode.getParentNode().removeChild(drugClassNode);
                        try {
                            drugClassName = (String) exprDrugClassName.evaluate(drugClassNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }
                        try {
                            drugClassVuid = (String) exprDrugClassVuid.evaluate(drugClassNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }

                        if (!StringUtils.isEmpty(drugClassName) && !StringUtils.isEmpty(drugClassVuid)) {
                            Code code = new Code();
                            code.setSystem("VUID");
                            code.setDisplay(drugClassName);
                            code.setCode(drugClassVuid);
                            drugClassCodeList.add(code);
                            drugClassName = null;
                            drugClassVuid = null;
                        }
                    }
                    allergy.setDrugClasses(drugClassCodeList.toArray(new Code[0]));
                }
            }

            Node drugIngredientsNode = null;
            try {
                drugIngredientsNode = (Node) exprDrugIngredients.evaluate(allergyNode, XPathConstants.NODE);
            } catch (XPathExpressionException e) {
                LOGGER.error(e.getMessage(), e);
            }
            if (drugIngredientsNode != null) {
                NodeList drugIngredientNodeList = null;
                try {
                    drugIngredientNodeList = (NodeList) exprDrugIngredient.evaluate(drugIngredientsNode, XPathConstants.NODESET);
                } catch (XPathExpressionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
                if (drugIngredientNodeList != null) {
                    String drugIngredientName = null;
                    String drugIngredientVuid = null;
                    List<Code> drugIngredientCodeList = new ArrayList<Code>();
                    int drugIngredientNodeListLength = drugIngredientNodeList.getLength();
                    for (int drugIngredientNum = 0; drugIngredientNum < drugIngredientNodeListLength; drugIngredientNum++) {
                        Node drugIngredientNode = drugIngredientNodeList.item(drugIngredientNum);
                        drugIngredientNode.getParentNode().removeChild(drugIngredientNode);
                        try {
                            drugIngredientName = (String) exprDrugIngredientName.evaluate(drugIngredientNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }
                        try {
                            drugIngredientVuid = (String) exprDrugIngredientVuid.evaluate(drugIngredientNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }

                        if (!StringUtils.isEmpty(drugIngredientName) && !StringUtils.isEmpty(drugIngredientVuid)) {
                            Code code = new Code();
                            code.setSystem("VUID");
                            code.setDisplay(drugIngredientName);
                            code.setCode(drugIngredientVuid);
                            drugIngredientCodeList.add(code);
                            drugIngredientName = null;
                            drugIngredientVuid = null;
                        }
                    }
                    allergy.setDrugIngredients(drugIngredientCodeList.toArray(new Code[0]));
                }
            }

            Node reactionsNode = null;
            try {
                reactionsNode = (Node) exprReactions.evaluate(allergyNode, XPathConstants.NODE);
            } catch (XPathExpressionException e) {
                LOGGER.error(e.getMessage(), e);
            }
            if (reactionsNode != null) {
                NodeList reactionNodeList = null;
                try {
                    reactionNodeList = (NodeList) exprReaction.evaluate(reactionsNode, XPathConstants.NODESET);
                } catch (XPathExpressionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
                if (reactionNodeList != null) {
                    String reactionName = null;
                    String reactionVuid = null;
                    List<Code> reactionCodeList = new ArrayList<Code>();
                    StringBuilder sbReactions = new StringBuilder();
                    int reactionNodeListLength = reactionNodeList.getLength();
                    for (int reactionNum = 0; reactionNum < reactionNodeListLength; reactionNum++) {
                        Node reactionNode = reactionNodeList.item(reactionNum);
                        reactionNode.getParentNode().removeChild(reactionNode);
                        try {
                            reactionName = (String) exprReactionName.evaluate(reactionNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }
                        try {
                            reactionVuid = (String) exprReactionVuid.evaluate(reactionNode, XPathConstants.STRING);
                        } catch (XPathExpressionException e) {
                            LOGGER.error(e.getMessage(), e);
                        }

                        if (!StringUtils.isEmpty(reactionName) && !StringUtils.isEmpty(reactionVuid)) {
                            if (sbReactions.length() > 0) {
                                sbReactions.append("; ");
                            }
                            sbReactions.append(reactionName);
                            Code code = new Code();
                            code.setSystem("VUID");
                            code.setDisplay(reactionName);
                            code.setCode(reactionVuid);
                            reactionCodeList.add(code);
                            reactionName = null;
                            reactionVuid = null;
                        }
                    }
//                    allergy.setComment(sbReactions.toString());
                    allergy.setReactions(reactionCodeList.toArray(new Code[0]));
                }
            }

            String enteredDateString = RPCUtil.evaluateNodeToString(allergyNode, exprEntered, LOGGER);
            if (!StringUtils.isEmpty(enteredDateString)) {
                Calendar enteredCalendar = Timestamp.toDateTime(enteredDateString);
                Calendar onsetCalendar = Timestamp.toDateTime(enteredDateString);
                allergy.setEnteredDate(enteredCalendar);
                allergy.setOnsetDate(onsetCalendar);
            }

            Node commentsNode = RPCUtil.evaluateNodeToNode(allergyNode, exprComments, LOGGER);
            if (commentsNode != null) {
                NodeList commentNodeList = RPCUtil.evaluateNodeToNodeList(commentsNode, exprComment, LOGGER);
                if (commentNodeList != null) {
                    SimpleDateFormat format1 = new SimpleDateFormat("yyyy-MM-dd", Locale.getDefault());
                    ArrayList<Comment> commentArrayList = new ArrayList<Comment>();
                    int commentNodeListLength = commentNodeList.getLength();
                    StringBuilder commentSB = new StringBuilder();

                    for (int commentNum = 0; commentNum < commentNodeListLength; commentNum++) {
                        Node commentNode = commentNodeList.item(commentNum);
                        commentNode.getParentNode().removeChild(commentNode);

                        String enteredString = RPCUtil.evaluateNodeToString(commentNode, exprEnteredAttr, LOGGER);
                        String enteredBy = RPCUtil.evaluateNodeToString(commentNode, exprEnteredByAttr, LOGGER);
                        String commentType = RPCUtil.evaluateNodeToString(commentNode, exprCommentType, LOGGER);
                        String commentText = RPCUtil.evaluateNodeToString(commentNode, exprCommentText, LOGGER);

                        Comment comment = new Comment();
                        comment.setCommentText(commentText);
                        comment.setCommentType(commentType);
                        comment.setEnteredBy(enteredBy);

                        String enteredFormattedString = null;
                        if (!StringUtils.isEmpty(enteredString)) {
                            Calendar entered = Timestamp.toDateTime(enteredString);
                            comment.setEnteredDate(entered);
                            enteredFormattedString = format1.format(entered.getTime());
                        }

                        commentSB.append("Entered By: ");
                        commentSB.append(enteredBy);
                        commentSB.append(" on ");
                        commentSB.append(enteredFormattedString);
                        commentSB.append("\nComment:\n");
                        commentSB.append(commentText);
                        commentSB.append("\n\n");

                        commentArrayList.add(comment);

                    }
                    allergy.setComments(commentArrayList.toArray(new Comment[0]));
                    allergy.setComment(commentSB.toString());
                }
            }

            allergy.setAllergyName(allergyName);
            allergy.setSeverity(allergySeverity);
            allergy.setAllergyId(allergyIen);
            if (!StringUtils.isEmpty(allergyVuid)) {
                Code allergyCode = new Code();
                allergyCode.setCode(allergyVuid);
                allergyCode.setSystem("VUID");
                Code[] codeArray = new Code[]{allergyCode};
                allergy.setCodes(codeArray);
            }

            fillCommonDataBeanValues(allergy);

            return allergy;
        }
    }

}
