/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.FreeTextReport;
import gov.va.med.vds.vistabeans.LabOrder;
import gov.va.med.vds.vistabeans.LabResult;
import gov.va.med.vds.vistabeans.Provider;
import gov.va.med.DNS   .common.Utils;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 *
 */
public class LabsRPC extends BaseRPC {

    private static final Logger LOGGER = LoggerFactory.getLogger(LabsRPC.class);

    private static final String XPATH_PANEL = "//panel";
    private static final String XPATH_ACCESSION = "//accession";
//    private static final String XPATH_ID_VALUE = "./id/@value";
    private static final String XPATH_ORDER_NAME = "./order/@name";
    private static final String XPATH_ORDER_CODE = "./order/@code";
    private static final String XPATH_COLLECTED = "./collected/@value";
    private static final String XPATH_RESULTED = "./resulted/@value";
    private static final String XPATH_COMMENT = "./comment";
    private static final String XPATH_GROUPNAME = "./groupName/@value";
    private static final String XPATH_SPECIMEN = "./specimen/@name";
    private static final String XPATH_STATUS = "./status/@value";
    private static final String XPATH_TYPE = "./type/@value";
    private static final String XPATH_NAME = "./name/@value";
    private static final String XPATH_FACILITY_NAME = "./facility/@name";
    private static final String XPATH_FACILITY_CODE = "./facility/@code";
    private static final String XPATH_VALUES = "./values";
    private static final String XPATH_TEST_VALUE = "./value";
    private static final String XPATH_TEST_ID = "@id";
    private static final String XPATH_TEST_TEST = "@test";
    private static final String XPATH_TEST_RESULT = "@result";
    private static final String XPATH_TEST_UNITS = "@units";
    private static final String XPATH_TEST_INTERPRETATION = "@interpretation";
    private static final String XPATH_TEST_HIGH = "@high";
    private static final String XPATH_TEST_LOW = "@low";
    private static final String XPATH_TEST_LOINC = "@loinc";
    private static final String XPATH_TEST_VUID = "@vuid";
    private static final String XPATH_TEST_PERFORMING_LAB = "@performingLab";
    private static final String XPATH_DOCUMENTS = "./documents";
    private static final String XPATH_DOCUMENT_LOCAL_TITLE = "./@localTitle";
//    private static final String XPATH_DOCUMENT_LOCAL_TITLE = "./document/@localTitle";
    private static final String XPATH_DOCUMENT = "./document";
    private static final String XPATH_CONTENT = "./content";

    private static final String DATETIMEFORMAT = "MM/dd/yyy HH:mm";
    private static final String DATEFORMAT = "MM/dd/yyy";
    private static final String FOOTER = "===============================================================================";

    private static final String XPATH_PROVIDER = "./provider";
    private static final String XPATH_PROVIDER_CODE = "./@code";
    private static final String XPATH_PROVIDER_NAME = "./@name";
    private static final String XPATH_PROVIDER_TAXONOMY_CODE = "./@taxonomyCode";
    private static final String XPATH_PROVIDER_PROVIDER_TYPE = "./@providerType";
    private static final String XPATH_PROVIDER_CLASSIFICATION = "./@classification";
    private static final String XPATH_PROVIDER_SPECIALIZATION = "./@specialization";
    private static final String XPATH_SERVICE_ATTR = "./@service";
    private static final String XPATH_OFFICE_PHONE = "./@officePhone";
    private static final String XPATH_FAX = "./@fax";
    private static final String XPATH_ANALOG_PAGER = "./@analogPager";
    private static final String XPATH_EMAIL = "./@email";

    public LabsRPC(RPCConnection cxn) {
        super(cxn);
    }

    public LabOrder[] getPatientLabs(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientLabs(cxn.getPatientIen(), startDate, endDate);
    }

    public LabOrder[] getPatientLabs(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetPatientLabsRequest(patientIen/*, startDate, endDate*/);
        String response = (String) cxn.query(request);

        LabOrder[] labs = toPatientLabs(response);

        ArrayList<LabOrder> labsFiltered = new ArrayList<LabOrder>();
        //filter by date range
        for (LabOrder labOrder : labs) {
            if (labOrder.getCollectionDate() == null) {
                labsFiltered.add(labOrder);
            }  else {
                Calendar labOrderDate = labOrder.getCollectionDate();
                if (labOrderDate.compareTo(startDate) >= 0 && labOrderDate.compareTo(endDate) <= 0) {
                    labsFiltered.add(labOrder);
                }
            }
        }


        return labsFiltered.toArray(new LabOrder[0]);
    }

    public LabOrder[] getPatientMicroLabOrders(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientMicroLabOrders(cxn.getPatientIen(), startDate, endDate);
    }

    public LabOrder[] getPatientMicroLabOrders(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetPatientMicroLabsRequest(patientIen/*, startDate, endDate*/);
        String microResponse = (String) cxn.query(request);

        LabOrder[] microLabs = toPatientMicroLabs(microResponse);

        ArrayList<LabOrder> microLabsFiltered = new ArrayList<LabOrder>();
        //filter by date range
        for (LabOrder labOrder : microLabs) {
            Calendar labOrderDate = labOrder.getCollectionDate();
            if (labOrderDate.compareTo(startDate) >= 0 && labOrderDate.compareTo(endDate) <= 0) {
                microLabsFiltered.add(labOrder);
            }
        }

        return microLabsFiltered.toArray(new LabOrder[0]);
    }

    public LabResult[] getPatientMicroLabResults(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientMicroLabResults(cxn.getPatientIen(), startDate, endDate);
    }

    public LabResult[] getPatientMicroLabResults(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        LabOrder[] microLabs = getPatientMicroLabOrders(patientIen, startDate, endDate);

        // Get the first lab result from each order and add them to a list
        ArrayList<LabResult> labResultsList = new ArrayList<LabResult>();
        for (LabOrder labOrder : microLabs) {
            LabResult[] labOrderResultsArray = labOrder.getLabResults();
            if (labOrderResultsArray.length > 0) {
                LabResult labResult = labOrderResultsArray[0];
                // add the subscript to the accession like the ones from the VPR
                String oldAccession = labResult.getAccession();
                labResult.setAccession(oldAccession + "^MI");
                labResultsList.add(labResult);
            }
        }

        return labResultsList.toArray(new LabResult[0]);
    }


    /**
     * Get the list of Anatomic Pathology Labs. This is done in the same way as CPRS using the ORWRP REPORT TEXT RPC
     * with L:OR_APL:ANATOMIC PATHOLOGY - ALL REPORTS~SP;ORDV02A;0;999 parameter.
     *
     * @return
     * @throws RPCException
     */
    public LabOrder[] getPatientAnaPathLabOrders(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientAnaPathLabOrders(cxn.getPatientIen(), startDate, endDate);
    }

    public LabOrder[] getPatientAnaPathLabOrders(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetPatientAnaPathLabsRequest(patientIen);
        String anaPathResponse = (String) cxn.query(request);

        LabOrder[] anaPathLabs = toPatientAnaPathLabs(anaPathResponse);

        ArrayList<LabOrder> apLabsFiltered = new ArrayList<LabOrder>();
        //filter by date range
        for (LabOrder labOrder : anaPathLabs) {
            Calendar labOrderDate = labOrder.getCollectionDate();
            if (labOrderDate.compareTo(startDate) >= 0 && labOrderDate.compareTo(endDate) <= 0) {
                apLabsFiltered.add(labOrder);
            }
        }

        return apLabsFiltered.toArray(new LabOrder[0]);
    }

    public LabResult[] getPatientAnaPathLabResults(Calendar startDate, Calendar endDate) throws RPCException {
        return getPatientAnaPathLabResults(cxn.getPatientIen(), startDate, endDate);
    }

    public LabResult[] getPatientAnaPathLabResults(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        LabOrder[] microLabs = getPatientAnaPathLabOrders(patientIen, startDate, endDate);


        // Get the first lab result from each order and add them to a list
        ArrayList<LabResult> labResultsList = new ArrayList<LabResult>();
        for (LabOrder labOrder : microLabs) {
            LabResult[] labOrderResultsArray = labOrder.getLabResults();
            if (labOrderResultsArray.length > 0) {
                LabResult labResult = labOrderResultsArray[0];
                // add the subscript to the accession like the ones from the VPR
                String oldAccession = labResult.getAccession();
                labResult.setAccession(oldAccession + "^AP");
                labResultsList.add(labResult);
            }
        }

        return labResultsList.toArray(new LabResult[0]);
    }


    private String buildGetPatientLabsRequest(String patientIen/*, Calendar startDate, Calendar endDate*/) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        /* ORWRP REPORT TEXT (Lab Order Status)
         *
         * This RPC is used to retrieve information on lab orders and their status from the Lab Order Entry file (#69).
         */
        vquery.addParameter(VistaQuery.LITERAL, "OR_LO:LAB ORDERS~LO;ORDV02;32;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "550000");
        vquery.addParameter(VistaQuery.LITERAL, "");

//        String startDateString = Timestamp.fromDateTime(startDate);
//        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
//        startDateString = flds[0];
//        String endDateString = Timestamp.fromDateTime(endDate);
//        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
//        endDateString = flds[0];
//        vq.addParameter(vq.LITERAL, startDateString);
//        vq.addParameter(vq.LITERAL, endDateString);
        vquery.addParameter(VistaQuery.LITERAL, "0");
        vquery.addParameter(VistaQuery.LITERAL, "0");


        return vquery.buildMessage();
    }

    private String buildGetPatientMicroLabsRequest(String patientIen/*, Calendar startDate, Calendar endDate*/) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);

        vquery.addParameter(VistaQuery.LITERAL, "L:OR_MICL:MICROBIOLOGY~MI;ORDV05;38;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "550000");
        vquery.addParameter(VistaQuery.LITERAL, "");

//        String startDateString = Timestamp.fromDateTime(startDate);
//        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
//        startDateString = flds[0];
//        String endDateString = Timestamp.fromDateTime(endDate);
//        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
//        endDateString = flds[0];
//        vq.addParameter(vq.LITERAL, startDateString);
//        vq.addParameter(vq.LITERAL, endDateString);
        vquery.addParameter(VistaQuery.LITERAL, "0");
        vquery.addParameter(VistaQuery.LITERAL, "0");

        return vquery.buildMessage();
    }

    private String buildGetPatientAnaPathLabsRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);

        vquery.addParameter(VistaQuery.LITERAL, "L:OR_APL:ANATOMIC PATHOLOGY - ALL REPORTS~SP;ORDV02A;0;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        return vquery.buildMessage();
    }

    private LabOrder[] toPatientLabs(String response) {
        /*
         * ORWRP REPORT TEXT (Lab Order Status); OR_LO:LAB ORDERS~LO;ORDV02;32;999
         * The results of this RPC are returned as a delimited 13-node global array.  Each node has different delimited
         * data elements and contains the collection date/time, the lab test name, the lab test IEN, the topography
         * field name (this is also known as the specimen), the topography field IEN, the urgency (i.e., "ROUTINE",
         * "STAT", etc.), the provider name (which is truncated to 10 characters), the provider IEN, the ordered
         * date/time, the accession number, the available date/time and the status (i.e., COMPLETED, CANCELLED, etc.).
         */

        final int ORDER_FACILITY = 1;
        final int ORDER_COLLECTION_DATETIME = 2;
        final int ORDER_LAB_TEST_NAME = 3;
        final int ORDER_LAB_TEST_IEN = 4;
        final int ORDER_TOPO_FIELD_NAME = 5;
        final int ORDER_TOPO_FIELD_IEN = 6;
        final int ORDER_URGENCY = 7;
        final int ORDER_PROVIDER_NAME = 8;
        final int ORDER_PROVIDER_IEN = 9;
        final int ORDER_ORDERED_DATETIME = 10;
        final int ORDER_ACCESSION = 11;
        final int ORDER_AVAILABLE_DATETIME = 12;
        final int ORDER_STATUS = 13;

        LabOrder[] rtc;
        if ("".equals(response)) {
            rtc = new LabOrder[0];
        } else {
    
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<LabOrder> labsList = new ArrayList<LabOrder>();
    
            LabOrder labOrder = null;
    
            for (String line : lines) {
                if ("".equals(line)) {
                    continue;
                }
    
                // Each line is the line number and the data delimited by a CARET
                String[] fields = StringUtils.split(line, StringUtils.CARET);
                int lineNumber = 0;
                String fieldData = "";
                for (int i = 0; i < fields.length; i++) {
                    if (i == 0) {
                        lineNumber = Integer.parseInt(fields[i]);
                    } else if (i == 1) {
                        fieldData = fields[i];
                    }
                }
    
                switch (lineNumber) {
                    case ORDER_FACILITY:
                        // Each Lab Test starts with the facility name (line 1) so add the old one to the list
                        if (labOrder != null) {
                            addLabResultToOrder(labOrder);
                            labsList.add(labOrder);
                        }
                        labOrder = new LabOrder();
                        fillCommonDataBeanValues(labOrder);
    
                        String[] subfields = StringUtils.split(fieldData, StringUtils.SEMICOLON);
                        for (int j = 0; j < subfields.length; j++) {
                            switch (j) {
                                case 0:
                                    labOrder.setFacilityName(subfields[0]);
                                    break;
                                case 1:
                                    labOrder.setStationNumber(subfields[1]);
                                    break;
                                default:
                                    break;
                            }
                        }
                        break;
                    case ORDER_COLLECTION_DATETIME:
                        Calendar collectionDateTime = parseDateTime(fieldData);
                        if (collectionDateTime != null && labOrder != null) {
                            labOrder.setCollectionDate(collectionDateTime);
                        }
                        break;
                    case ORDER_LAB_TEST_NAME:
                        if (labOrder != null) {
                            labOrder.setTestName(fieldData);
                        }
                        break;
                    case ORDER_LAB_TEST_IEN:
                        if (labOrder != null) {
                            labOrder.setLabTestIEN(fieldData);
                        }
                        break;
                    case ORDER_TOPO_FIELD_NAME:
                        if (labOrder != null) {
                            labOrder.setCollectionSample(fieldData);
                        }
                        break;
                    case ORDER_TOPO_FIELD_IEN:
                        if (labOrder != null) {
                            labOrder.setCollectionSampleIEN(fieldData);
                        }
                        break;
                    case ORDER_URGENCY:
                        if (labOrder != null) {
                            labOrder.setUrgency(fieldData);
                        }
                        break;
                    case ORDER_PROVIDER_NAME:
                        if (labOrder != null) {
                            Provider provider = new Provider();
                            provider.setName(fieldData);
                            labOrder.setRequestingProvider(provider);
                        }
                        break;
                    case ORDER_PROVIDER_IEN:
                        if (labOrder != null) {
                            labOrder.setRequestingProviderIEN(fieldData);
                        }
                        break;
                    case ORDER_ORDERED_DATETIME:
                        Calendar orderDateTime = parseDateTime(fieldData);
                        if (orderDateTime != null && labOrder != null) {
                            labOrder.setOrderDate(orderDateTime);
                        }
                        break;
                    case ORDER_ACCESSION:
                        if (labOrder != null) {
                            labOrder.setAccessionNumber(fieldData);
                        }
                        break;
                    case ORDER_AVAILABLE_DATETIME:
                        Calendar availableDateTime = parseDateTime(fieldData);
                        if (availableDateTime != null && labOrder != null) {
                            labOrder.setAvailableDate(availableDateTime);
                        }
                        break;
                    case ORDER_STATUS:
                        if (labOrder != null) {
                            labOrder.setStatus(fieldData);
                        }
                        break;
                    default:
                        break;
                }
            }
    
            if (labOrder != null) {
                addLabResultToOrder(labOrder);
                labsList.add(labOrder);
            }
    
            rtc = labsList.toArray(new LabOrder[0]);
        }
        return rtc;
    }

    private void addLabResultToOrder(LabOrder labOrder) {
        if (labOrder != null) {
            LabResult labResult = new LabResult();
            fillCommonDataBeanValues(labResult);
            // Micro labs have only a single result so copy the result with things from the order
            labResult.setAccession(labOrder.getAccessionNumber());
            labResult.setTestName(labOrder.getTestName());
            labResult.setResultDate(labOrder.getCollectionDate());
            labResult.setSpecimen(labOrder.getCollectionSample());
            labResult.setFacilityName(labOrder.getFacilityName());
            labResult.setStationNumber(labOrder.getStationNumber());
            labResult.setOrderComment(labOrder.getComment());
            // Add the result to the order
            labOrder.setLabResults(new LabResult[]{labResult});
        }
    }

    private LabOrder[] toPatientMicroLabs(String response) {
        /*
         * ORWRP REPORT TEXT (Lab Order Status); OR_LO:LAB ORDERS~LO;ORDV02;32;999
         * The results of this RPC are returned as a delimited 13-node global array.  Each node has different delimited
         * data elements and contains the collection date/time, the lab test name, the lab test IEN, the topography
         * field name (this is also known as the specimen), the topography field IEN, the urgency (i.e., "ROUTINE",
         * "STAT", etc.), the provider name (which is truncated to 10 characters), the provider IEN, the ordered
         * date/time, the accession number, the available date/time and the status (i.e., COMPLETED, CANCELLED, etc.).
         */

        final int ORDER_FACILITY = 1;
        final int ORDER_COLLECTION_DATETIME = 2;
        final int ORDER_LAB_TEST_NAME = 3;
        final int ORDER_TOPO_FIELD_NAME = 4;
        final int ORDER_ACCESSION = 6;
        final int ORDER_REPORT    ;

        LabOrder[] rtc;
        if ("".equals(response)) {
            rtc = new LabOrder[0];
        } else {
    
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<LabOrder> labsList = new ArrayList<LabOrder>();
    
            LabOrder labOrder = null;
            StringBuilder reportTextSB = new StringBuilder();
    
            for (String line : lines) {
                if ("".equals(line)) {
                    continue;
                }
    
                // Each line is the line number and the data delimited by a CARET
                String[] fields = StringUtils.split(line, StringUtils.CARET);
                int lineNumber = 0;
                String fieldData = "";
                for (int i = 0; i < fields.length; i++) {
                    if (i == 0) {
                        lineNumber = Integer.parseInt(fields[i]);
                    } else if (i == 1) {
                        fieldData = fields[i];
                    }
                }
    
                switch (lineNumber) {
                    case ORDER_FACILITY:
                        // Each Lab Test starts with the facility name (line 1) so add the old one to the list
                        if (labOrder != null && reportTextSB.length() > 0) {
                            addMicroResultToOrder(labOrder, reportTextSB.toString());
                            labsList.add(labOrder);
                        }
                        labOrder = new LabOrder();
                        fillCommonDataBeanValues(labOrder);
                        labOrder.setSubscript("MI");
                        reportTextSB.setLength(0);
    
                        String[] subfields = StringUtils.split(fieldData, StringUtils.SEMICOLON);
                        for (int j = 0; j < subfields.length; j++) {
                            switch (j) {
                                case 0:
                                    labOrder.setFacilityName(subfields[0]);
                                    break;
                                case 1:
                                    labOrder.setStationNumber(subfields[1]);
                                    break;
                                default:
                                    break;
                            }
                        }
                        break;
                    case ORDER_COLLECTION_DATETIME:
                        Calendar collectionDateTime = parseDateTime(fieldData);
                        if (collectionDateTime != null && labOrder != null) {
                            labOrder.setCollectionDate(collectionDateTime);
                        }
                        break;
                    case ORDER_LAB_TEST_NAME:
                        if (labOrder != null) {
                            labOrder.setTestName(fieldData);
                        }
                        break;
                    case ORDER_TOPO_FIELD_NAME:
                        if (labOrder != null) {
                            labOrder.setCollectionSample(fieldData);
                        }
                        break;
                    case ORDER_ACCESSION:
                        if (labOrder != null) {
                            labOrder.setAccessionNumber(fieldData);
                        }
                        break;
                    case ORDER_REPORT:
                        reportTextSB.append(fieldData);
                        reportTextSB.append("\n");
                        break;
                    default:
                        break;
                }
            }
    
            if (labOrder != null && reportTextSB != null) {
                addMicroResultToOrder(labOrder, reportTextSB.toString());
                labsList.add(labOrder);
            }
    
            rtc = labsList.toArray(new LabOrder[0]);
        }
        return rtc;
    }

    private LabOrder[] toPatientAnaPathLabs(String response) {
        /*
         * ORWRP REPORT TEXT (Lab Order Status); OR_LO:LAB ORDERS~LO;ORDV02;32;999
         * The results of this RPC are returned as a delimited 13-node global array.  Each node has different delimited
         * data elements and contains the collection date/time, the lab test name, the lab test IEN, the topography
         * field name (this is also known as the specimen), the topography field IEN, the urgency (i.e., "ROUTINE",
         * "STAT", etc.), the provider name (which is truncated to 10 characters), the provider IEN, the ordered
         * date/time, the accession number, the available date/time and the status (i.e., COMPLETED, CANCELLED, etc.).
         */

        final int ORDER_FACILITY = 1;
        final int ORDER_COLLECTION_DATETIME = 2;
        final int ORDER_TOPO_FIELD_NAME = 3;
        final int ORDER_ACCESSION = 4;
        final int ORDER_REPORT    ;

        LabOrder[] rtc;
        if ("".equals(response)) {
            rtc = new LabOrder[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<LabOrder> labsList = new ArrayList<LabOrder>();
    
            LabOrder labOrder = null;
            StringBuilder reportTextSB = new StringBuilder();
    
            for (String line : lines) {
                if ("".equals(line)) {
                    continue;
                }
    
                // Each line is the line number and the data delimited by a CARET
                String[] fields = StringUtils.split(line, StringUtils.CARET);
                int lineNumber = 0;
                String fieldData = "";
                for (int i = 0; i < fields.length; i++) {
                    if (i == 0) {
                        lineNumber = Integer.parseInt(fields[i]);
                    } else if (i == 1) {
                        fieldData = fields[i];
                    }
                }
    
                switch (lineNumber) {
                    case ORDER_FACILITY:
                        // Each Lab Test starts with the facility name (line 1) so add the old one to the list
                        if (labOrder != null && reportTextSB.length() > 0) {
                            addAnaPathResultToOrder(labOrder, reportTextSB.toString());
                            labsList.add(labOrder);
                        }
                        labOrder = new LabOrder();
                        fillCommonDataBeanValues(labOrder);
                        labOrder.setSubscript("AP");
    
                        reportTextSB.setLength(0);
    
                        String[] subfields = StringUtils.split(fieldData, StringUtils.SEMICOLON);
                        for (int j = 0; j < subfields.length; j++) {
                            switch (j) {
                                case 0:
                                    labOrder.setFacilityName(subfields[0]);
                                    break;
                                case 1:
                                    labOrder.setStationNumber(subfields[1]);
                                    break;
                                default:
                                    break;
                            }
                        }
                        break;
                    case ORDER_COLLECTION_DATETIME:
                        Calendar collectionDateTime = parseDateTime(fieldData);
                        if (collectionDateTime != null && labOrder != null) {
                            labOrder.setCollectionDate(collectionDateTime);
                        }
                        break;
                    case ORDER_TOPO_FIELD_NAME:
                        if (labOrder != null) {
                            labOrder.setCollectionSample(fieldData);
                        }
                        break;
                    case ORDER_ACCESSION:
                        if (labOrder != null) {
                            labOrder.setAccessionNumber(fieldData);
                        }
                        break;
                    case ORDER_REPORT:
                        reportTextSB.append(fieldData);
                        reportTextSB.append("\n");
                        break;
                    default:
                        break;
                }
            }
    
            if (labOrder != null && reportTextSB != null) {
                addAnaPathResultToOrder(labOrder, reportTextSB.toString());
                labsList.add(labOrder);
            }
    
            rtc = labsList.toArray(new LabOrder[0]);
        }
        return rtc;
    }

    private void addMicroResultToOrder(LabOrder labOrder, String labReport) {
        if (labOrder != null) {
            LabResult labResult = new LabResult();
            fillCommonDataBeanValues(labResult);
            labResult.setResult(labReport);
            // Micro labs have only a single result so copy the result with things from the order
            labResult.setAccession(labOrder.getAccessionNumber());
            labResult.setTestName(labOrder.getTestName());
            labResult.setResultDate(labOrder.getCollectionDate());
            labResult.setSpecimen(labOrder.getCollectionSample());
            labResult.setFacilityName(labOrder.getFacilityName());
            labResult.setStationNumber(labOrder.getStationNumber());
            // Add the result to the order
            labOrder.setLabResults(new LabResult[]{labResult});
        }
    }

//    private static final String titleMarker = "----";

    private void addAnaPathResultToOrder(LabOrder labOrder, String labReport) {
        if (labOrder != null) {
            // extract test name from report text
            // it is in the first non-white space line surrounded by "----" on each side
            StringBuilder titleSB = new StringBuilder();
            if (labOrder.getAccessionNumber() != null) {
                String accessionNumber = labOrder.getAccessionNumber().trim();
                int firstSpaceIndex = labOrder.getAccessionNumber().indexOf(" ");
                if (firstSpaceIndex > 0) {
                    titleSB.append(accessionNumber.substring(0, firstSpaceIndex));
                } else {
                    titleSB.append(accessionNumber);
                }
            }
//            int titleStartIndex = labReport.indexOf(titleMarker) + titleMarker.length();
//            int titleEndIndex = labReport.indexOf(titleMarker, titleStartIndex);
//            titleSB.append(labReport.substring(titleStartIndex, titleEndIndex).trim());
            if (labOrder.getCollectionSample() != null) {
                titleSB.append(" - ");
                titleSB.append(labOrder.getCollectionSample());
            }
            labOrder.setTestName(titleSB.toString());
            // add the fields to thee first result in the order
            addMicroResultToOrder(labOrder, labReport);
        }
    }

    public LabResult[] getLabOrderResult(Calendar startDate, Calendar endDate) throws RPCException {
        return getLabOrderResult(cxn.getPatientIen(), startDate, endDate);
    }

    public LabResult[] getLabOrderResult(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetLabOrderResultRequest(patientIen, startDate, endDate);
        String response = (String) cxn.query(request);

        LabResult[] labs = toLabOrderResult(/*patientIen, */response);

        return labs;
    }

    public LabResult[] getLabOrderResult(Calendar collectionDate, String accessionNumber) throws RPCException {
        return getLabOrderResult(cxn.getPatientIen(), collectionDate, accessionNumber);
    }

    public LabResult[] getLabOrderResult(String patientIen, Calendar collectionDate, String accessionNumber) throws RPCException {

        // Need to search around the collection date in case the date is off
        Calendar endDate = (Calendar) collectionDate.clone();
        Calendar startDate = (Calendar) collectionDate.clone();

        // Pad by two days before and after 48 hr * 60 min * 60 sec * 1000 ms
        endDate.setTimeInMillis(collectionDate.getTimeInMillis() + 172800000);
        startDate.setTimeInMillis(collectionDate.getTimeInMillis() - 172800000);

        // Get the Lab Orders using the VPR call
        LabOrder[] labOrders = getVPRLabOrders(patientIen, startDate, endDate);

        // find the order(s) that have the accession number
        ArrayList<LabResult> resultsList = new ArrayList<LabResult>();

        for (LabOrder labOrder : labOrders) {
            if (labOrder.getAccessionNumber() != null
                    && accessionNumber.equals(labOrder.getAccessionNumber())
                    && labOrder.getLabResults() != null) {
                resultsList.addAll(Arrays.asList(labOrder.getLabResults()));
            }
        }

        return resultsList.toArray(new LabResult[0]);
    }

    private String buildGetLabOrderResultRequest(String patientIen, Calendar startDate, Calendar endDate) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWRP REPORT TEXT");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        /* ORWRP REPORT TEXT (Lab Order Status)
         *
         * This RPC is used to retrieve information on lab results from the Lab Data file (#63).
         */
        vquery.addParameter(VistaQuery.LITERAL, "OR_CH:CHEM & HEMATOLOGY~CH;ORDV02;3;999");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");
        vquery.addParameter(VistaQuery.LITERAL, "");

        String startDateString = Timestamp.fromDateTime(startDate);
        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
        startDateString = flds[0];
        String endDateString = Timestamp.fromDateTime(endDate);
        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
        endDateString = flds[0];
        vquery.addParameter(VistaQuery.LITERAL, startDateString);
        vquery.addParameter(VistaQuery.LITERAL, endDateString);

        return vquery.buildMessage();
    }

    private LabResult[] toLabOrderResult(/*String patientIen, */String response) {
        /*
         * ORWRP REPORT TEXT (Lab Results); OR_CH:CHEM & HEMATOLOGY~CH;ORDV02;3;999
         * The results of this RPC are returned as a delimited 9 or 11-node global array.  Nodes 10 and 11 only exist
         * if there are comments with the results.  Each node has different delimited data elements and contains the
         * facility name and station number (separated by a ";"), the collection date/time, the lab test name, the
         * specimen (also known as topography field) name, the lab test result, the test result flag (which can be "H"
         * for "High", "L" for "Low" or NULL for "Normal"), the units, the numeric reference low, the numeric reference
         * high, the comments and a "[+]" symbol.  The "[+]" symbol is used in the CPRS GUI to signify that there is more
         * text or lines for comments that may not be seen in the comments field in the CPRS GUI.  Thus, the symbol that
         * is given in Node 11 is not necessarily needed.
         */

        final int RESULT_FACILITY_NAME = 1;
        final int RESULT_COLLECTION_DATETIME = 2;
        final int RESULT_LAB_TEST_NAME = 3;
        final int RESULT_SPECIMEN_NAME = 4;
        final int RESULT_RESULT = 5;
        final int RESULT_FLAG = 6;
        final int RESULT_UNITS = 7;
        final int RESULT_REFERENCE_LOW = 8;
        final int RESULT_REFERENCE_HIGH = 9;
        final int RESULT_COMMENTS = 10;

        LabResult[] rtc;
        if ("".equals(response)) {
            rtc = new LabResult[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<LabResult> resultsList = new ArrayList<LabResult>();
    
            LabResult labResult = new LabResult();
            StringBuilder referenceRangeSB = new StringBuilder();
            StringBuilder commentStringBuilder = new StringBuilder();
    
            for (String line : lines) {
                if ("".equals(line)) {
                    continue;
                }
    
                // Each line is the line number and the data delimited by a CARET
                String[] fields = StringUtils.split(line, StringUtils.CARET);
                int lineNumber = 0;
                String fieldData = "";
                for (int i = 0; i < fields.length; i++) {
                    if (i == 0) {
                        lineNumber = Integer.parseInt(fields[i]);
                    } else if (i == 1) {
                        fieldData = fields[i];
                    }
                }
    
                switch (lineNumber) {
                    case RESULT_FACILITY_NAME:
                        // Each Lab Test starts with the facility name (line 1) so add the old one to the list
                        resultsList.add(labResult);
                        fillCommonDataBeanValues(labResult);
                        referenceRangeSB.setLength(0);
                        commentStringBuilder.setLength(0);
    
                        String[] subfields = StringUtils.split(fieldData, StringUtils.SEMICOLON);
                        for (int j = 0; j < subfields.length; j++) {
                            switch (j) {
                                case 0:
                                    labResult.setFacilityName(subfields[0]);
                                    break;
                                case 1:
                                    labResult.setStationNumber(subfields[1]);
                                    break;
                                default:
                                    break;
                            }
                        }
                        break;
                    case RESULT_COLLECTION_DATETIME:
                        Calendar collectionDateTime = parseDateTime(fieldData);
                        if (collectionDateTime != null) {
                            labResult.setResultDate(collectionDateTime);
                        }
                        break;
                    case RESULT_LAB_TEST_NAME:
                        labResult.setTestName(fieldData);
                        break;
                    case RESULT_SPECIMEN_NAME:
                        labResult.setSpecimen(fieldData);
                        break;
                    case RESULT_RESULT:
                        labResult.setResult(fieldData);
                        break;
                    case RESULT_FLAG:
                        labResult.setInterpretation(fieldData);
                        break;
                    case RESULT_UNITS:
                        labResult.setUnits(fieldData);
                        break;
                    case RESULT_REFERENCE_LOW:
                        referenceRangeSB.append(fieldData);
                        referenceRangeSB.append(" - ");
                        break;
                    case RESULT_REFERENCE_HIGH:
                        referenceRangeSB.append(fieldData);
                        labResult.setReferenceRange(referenceRangeSB.toString());
                        break;
                    case RESULT_COMMENTS:
                        if (labResult.getComment() != null) {
                            commentStringBuilder.delete(0, commentStringBuilder.length());
                            commentStringBuilder.append(labResult.getComment());
                        }
                        commentStringBuilder.append("\r\n");
                        commentStringBuilder.append(fieldData);
                        labResult.setComment(commentStringBuilder.toString());
                        break;
                    default:
                        break;
                }
            }
    
            fillCommonDataBeanValues(labResult);
            resultsList.add(labResult);
    
            rtc = resultsList.toArray(new LabResult[0]);
        }
        return rtc;
    }


    private static Calendar parseDateTime(String date) {
        SimpleDateFormat sdf = new SimpleDateFormat(DATETIMEFORMAT, Locale.getDefault());

        Calendar returnCalendar = new GregorianCalendar();

        try {
            returnCalendar.setTime(sdf.parse(date));
        } catch (ParseException pe) {
            returnCalendar = parseDate(date);
        }

        return returnCalendar;
    }

    private static Calendar parseDate(String date) {
        SimpleDateFormat sdf = new SimpleDateFormat(DATEFORMAT, Locale.getDefault());

        Calendar returnCalendar = new GregorianCalendar();

        try {
            returnCalendar.setTime(sdf.parse(date));
        } catch (ParseException pe) {
            returnCalendar = null;
        }

        return returnCalendar;
    }

    /**
     * For the date range, makes a call to ORWRP REPORT TEXT "OR_CH:CHEM & HEMATOLOGY~CH;ORDV02;3;999" and scans all
     * returned results for those that match the test name.
     *
     * @param startDate Start timestamp for range to search
     * @param endDate   End timestamp for range to search
     * @param testName  Name of the Lab Test to search
     * @return Array of LabResults that match testName in the given date range
     * @throws RPCException
     */
    public LabResult[] getPatientLabTestResults(Calendar startDate, Calendar endDate, String testName) throws RPCException {

        return getPatientLabTestResults(cxn.getPatientIen(), startDate, endDate, testName);
    }

    /**
     * For the date range, makes a call to ORWRP REPORT TEXT "OR_CH:CHEM & HEMATOLOGY~CH;ORDV02;3;999" and scans all
     * returned results for those that match the test name.
     *
     * @param patientIen the patient IEN
     * @param startDate  Start timestamp for range to search
     * @param endDate    End timestamp for range to search
     * @param testName   Name of the Lab Test to search
     * @return Array of LabResults that match testName in the given date range
     * @throws RPCException
     */
    public LabResult[] getPatientLabTestResults(String patientIen, Calendar startDate, Calendar endDate, String testName) throws RPCException {
        LabResult[] results = getVPRLabResults(patientIen, startDate, endDate);

        ArrayList<LabResult> resultArrayList = new ArrayList<LabResult>();

        for (LabResult result : results) {
            if (testName.equals(result.getTestName())) {
                resultArrayList.add(result);
            }
        }

        return resultArrayList.toArray(new LabResult[0]);
    }

    /**
     * Gets a list of Lab Orders using the VPR RPC with "panels" as the type. This will return the results for each
     * order embedded in the LabOrder.
     *
     * @param startDate The start of the date range for the list of orders.
     * @param endDate   The end of the date range for the list of orders.
     * @return The list of lab orders for the specified date range.
     * @throws RPCException
     */
    public LabOrder[] getVPRLabOrders(Calendar startDate, Calendar endDate) throws RPCException {
        return getVPRLabOrders(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Gets a list of Lab Orders using the VPR RPC with "panels" as the type. This will return the results for each
     * order embedded in the LabOrder.
     *
     * @param patientIen The patient data file number
     * @param startDate  The start of the date range for the list of orders.
     * @param endDate    The end of the date range for the list of orders.
     * @return The list of lab orders for the specified date range.
     * @throws RPCException
     */
    public LabOrder[] getVPRLabOrders(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
//        String request = buildGetPatientLabsVPRRequest(patientIen, startDate, endDate);
        String request = VPRUtil.buildVPRRequestWithTypeFilter(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.PANELS, "CH");
        String response = (String) cxn.query(request);

        LabOrder[] labs = parseLabsFromVPR(/*patientIen, */response);
        ArrayList<LabOrder> labsArrayList = new ArrayList<LabOrder>(Arrays.asList(labs));

        request = VPRUtil.buildVPRRequestWithTypeFilter(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.ACCESSIONS, "MI;CY;EM;SP;AU", true);
        response = (String) cxn.query(request);

        LabOrder[] labsAP = parseAccessionsFromVPR(/*patientIen, */response);
        labsArrayList.addAll(Arrays.asList(labsAP));

        return labsArrayList.toArray(new LabOrder[0]);
    }

    /**
     * Builds the VPR RPC request string using "accessions" as the type. The accessions contain the test results. The
     * RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when setting up the connection.
     *
     * @param patientIen Patient IEN
     * @param startDate  Start date of range
     * @param endDate    End date of range
     * @return The RPC message string.
     */
//    private String buildGetPatientLabsVPRRequest(String patientIen, Calendar startDate, Calendar endDate) {
//        VistaQuery vq = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA");
//        vq.addParameter(vq.LITERAL, patientIen);
//        /* ORWRP REPORT TEXT (Lab Order Status)
//         *
//         * This RPC is used to retrieve information on lab orders and their status from the Lab Order Entry file (#69).
//         */
////        vq.addParameter(vq.LITERAL, "accessions;orders");
//        vq.addParameter(vq.LITERAL, "panels");
//
//        String startDateString = Timestamp.fromDateTime(startDate);
//        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
//        startDateString = flds[0];
//        String endDateString = Timestamp.fromDateTime(endDate);
//        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
//        endDateString = flds[0];
//        vq.addParameter(vq.LITERAL, startDateString);
//        vq.addParameter(vq.LITERAL, endDateString);
//
//        return vq.buildMessage();
//    }

    /**
     * Parses the response from the "panels" VPR. This is the list of lab orders and the test results associated
     * with those orders.
     *
     * @param response   The response string from the RPC.
     * @return Array of LabOrders, the results are embedded in each LabOrder as an array of LabResults.
     */
//    private LabOrder[] parseLabsFromVPR(String patientIen, String response) {
    private LabOrder[] parseLabsFromVPR(String response) throws RPCException {
        LabOrder[] rtc;
        if ("".equals(response)) {
            rtc = new LabOrder[0];
        } else {
    
            StringBuilder sb = new StringBuilder(Utils.stripInvalidXMLCharacters(response));
            sb.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<LabOrder> labsList = new ArrayList<LabOrder>();
    
            XPathFactory xPathfactory;
            XPath xpath;
    
            XPathExpression exprPanel = null;
    
            xPathfactory = XPathFactory.newInstance();
            xpath = xPathfactory.newXPath();
    
            try {
                exprPanel = xpath.compile(XPATH_PANEL);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            Document doc = null;
            try {
                doc = parseXMLDocument(sb.toString());
            } catch (Exception e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            NodeList panelNodeList = null;
            try {
                panelNodeList = (NodeList) exprPanel.evaluate(doc, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            List<VPRLabOrderParser> nodeParserList = new ArrayList<VPRLabOrderParser>();
            int panelNodeListLength = panelNodeList.getLength();
            for (int i = 0; i < panelNodeListLength; i++) {
    
                Node panelNode = panelNodeList.item(i);
                panelNode.getParentNode().removeChild(panelNode);
    
                VPRLabOrderParser nodeParser = new VPRLabOrderParser(panelNode);
                nodeParserList.add(nodeParser);
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(10);
            List<Future<LabOrder>> futures = null;
            try {
                futures = executorService.invokeAll(nodeParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse labs results : panel node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<LabOrder>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<LabOrder> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Lab Order Data");
                    } else {
                        labsList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    logErrorMessage(e);
                } catch (ExecutionException e) {
                    logErrorMessage(e);
                }
            }
    
            rtc = labsList.toArray(new LabOrder[0]);
        }
        return rtc;
    }

    /**
     * Parses the response from the "accessions" VPR. This is the list of lab orders and the test results associated
     * with those orders.
     *
     * @param response   The response string from the RPC.
     * @return Array of LabOrders, the results are embedded in each LabOrder as an array of LabResults.
     */
    private LabOrder[] parseAccessionsFromVPR(String response) throws RPCException {
        LabOrder[] rtc;
        if ("".equals(response)) {
            rtc = new LabOrder[0];
        } else {
    
            StringBuilder sbld = new StringBuilder(Utils.stripInvalidXMLCharacters(response));
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<LabOrder> labsList = new ArrayList<LabOrder>();
    
            XPathFactory xPathfactory;
            XPath xpath;
    
            XPathExpression exprAccession = null;
    
            xPathfactory = XPathFactory.newInstance();
            xpath = xPathfactory.newXPath();
    
            try {
                exprAccession = xpath.compile(XPATH_ACCESSION);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            Document doc = null;
            try {
                doc = parseXMLDocument(sbld.toString());
            } catch (Exception e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            NodeList accessionNodeList = null;
            try {
                accessionNodeList = (NodeList) exprAccession.evaluate(doc, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }
    
            List<VPRAccessionParser> nodeParserList = new ArrayList<VPRAccessionParser>();
            int accessionNodeListLength = accessionNodeList.getLength();
            for (int i = 0; i < accessionNodeListLength; i++) {
    
                Node accessionNode = accessionNodeList.item(i);
                accessionNode.getParentNode().removeChild(accessionNode);
    
                VPRAccessionParser nodeParser = new VPRAccessionParser(accessionNode);
                nodeParserList.add(nodeParser);
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(10);
            List<Future<LabOrder>> futures = null;
            try {
                futures = executorService.invokeAll(nodeParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse labs results : accession node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<LabOrder>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<LabOrder> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Lab Accession Data");
                    } else {
                        labsList.add(future.get());
                    }
    
                } catch (InterruptedException e) {
                    logErrorMessage(e);
                } catch (ExecutionException e) {
                    logErrorMessage(e);
                }
            }
    
            rtc = labsList.toArray(new LabOrder[0]);
        }
        return rtc;
    }


    public FreeTextReport getLabReport(Calendar collectedDate, String accessionNumber) throws RPCException {
        return getLabReport(cxn.getPatientIen(), collectedDate, accessionNumber);
    }

    /**
     * Query vista for lab report.
     */
    public FreeTextReport getLabReport(String patientIen, Calendar collectedDate, String accessionNumber) throws RPCException {

        // There are documents so we'll make an RPC call to get the Report text
        Calendar endDate = (Calendar) collectedDate.clone();
        Calendar startDate = (Calendar) collectedDate.clone();

        // Pad by two days before and after 48 hr * 60 min * 60 sec * 1000 ms
        endDate.setTimeInMillis(collectedDate.getTimeInMillis() + 172800000);
        startDate.setTimeInMillis(collectedDate.getTimeInMillis() - 172800000);

        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORWLRR INTERIM");

        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        vquery.addParameter(VistaQuery.LITERAL, Timestamp.fromDateTime(endDate));
        vquery.addParameter(VistaQuery.LITERAL, Timestamp.fromDateTime(startDate));
        String reportText = (String) cxn.query(vquery.buildMessage());

        String extractedReport = extractReport(reportText, accessionNumber, collectedDate);

        FreeTextReport report = new FreeTextReport();
        report.setReportText(extractedReport);
        fillCommonDataBeanValues(report);

        return report;
    }

//    private static final String footer = "===============================================================================\r\n \r\n\r\n";

    private String extractReport(String fullReport, String accessionNumber, Calendar collectedDate) {
        int startIndex = 0;
        int endIndex = 0;

        StringBuilder compositeReportsSB = new StringBuilder();
        String report = null;
        String accession = null;

        if (accessionNumber.contains("^")) {
            accession = accessionNumber.substring(0, accessionNumber.indexOf('^'));
        } else {
            accession = accessionNumber;
        }

        int fullReportLength = fullReport.length();
        while (startIndex < fullReportLength) {

            startIndex = endIndex;
            endIndex = fullReport.indexOf(FOOTER, startIndex);
            if (endIndex < 0) {
                // this may be the last report
                endIndex = fullReportLength;
            } else {
                // add in the footer for the report
                endIndex = endIndex + FOOTER.length();
            }

            report = fullReport.substring(startIndex, endIndex);

            if (report.contains(accession)) {
                compositeReportsSB.append(report);
                compositeReportsSB.append("\n\n\n");
            }
        }

        // couldn't find the microbiology report
        if (compositeReportsSB.length() == 0) {
            final String PARSE_DATE_FORMAT = "MMM dd, yyyy HH:mm";
            SimpleDateFormat sdf = new SimpleDateFormat(PARSE_DATE_FORMAT, Locale.getDefault());

            compositeReportsSB.append("No Lab Report on ");
            compositeReportsSB.append(sdf.format(collectedDate.getTime()));
            compositeReportsSB.append(" for ");
            compositeReportsSB.append(accessionNumber);

        }

        return compositeReportsSB.toString();
    }

    private void logErrorMessage(Exception except) {
        LOGGER.error(except.getMessage(), except);
    }

    private class VPRLabOrderParser implements Callable<LabOrder> {
        private Node panelNode;

        public VPRLabOrderParser(Node panelNode) {
            this.panelNode = panelNode;
        }

        public LabOrder call() throws RPCException {
            XPathFactory xPathfactory;
            XPath xpath;

            XPathExpression exprOrderName = null;
            XPathExpression exprCollected = null;
            XPathExpression exprComment = null;
            XPathExpression exprGroupName = null;
            XPathExpression exprSpecimen = null;
            XPathExpression exprStatus = null;
            XPathExpression exprType = null;
//            XPathExpression exprName = null;
            XPathExpression exprId = null;
            XPathExpression exprFacilityName = null;
            XPathExpression exprFacilityCode = null;
            XPathExpression exprValues = null;
            XPathExpression exprTestValue = null;
            XPathExpression exprTestId = null;
            XPathExpression exprTestTest = null;
            XPathExpression exprTestResult = null;
            XPathExpression exprTestUnits = null;
            XPathExpression exprTestInterpretation = null;
            XPathExpression exprTestHigh = null;
            XPathExpression exprTestLow = null;
            XPathExpression exprTestLoinc = null;
            XPathExpression exprTestVuid = null;
            XPathExpression exprTestPerformingLab = null;
//            XPathExpression exprDocuments = null;
//            XPathExpression exprDocumentLocalTitle = null;
            XPathExpression exprResulted = null;
//            XPathExpression exprDocument = null;
//            XPathExpression exprDocumentContent = null;
            XPathExpression exprProvider = null;
            XPathExpression exprProviderCode = null;
            XPathExpression exprProviderName = null;
            XPathExpression exprProviderTaxonomyCode = null;
            XPathExpression exprProviderProviderType = null;
            XPathExpression exprProviderClassification = null;
            XPathExpression exprProviderSpecialization = null;
            XPathExpression exprProviderService = null;
            XPathExpression exprProviderOfficePhone = null;
            XPathExpression exprProviderAnalogPager = null;
            XPathExpression exprProviderFax = null;
            XPathExpression exprProviderEmail = null;

            xPathfactory = XPathFactory.newInstance();
            xpath = xPathfactory.newXPath();


            try {
                exprOrderName = xpath.compile(XPATH_ORDER_NAME);
                exprCollected = xpath.compile(XPATH_COLLECTED);
                exprComment = xpath.compile(XPATH_COMMENT);
                exprGroupName = xpath.compile(XPATH_GROUPNAME);
                exprSpecimen = xpath.compile(XPATH_SPECIMEN);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprType = xpath.compile(XPATH_TYPE);
                exprId = xpath.compile(XPATH_ORDER_CODE);
                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
                exprFacilityCode = xpath.compile(XPATH_FACILITY_CODE);
                exprValues = xpath.compile(XPATH_VALUES);
                exprTestValue = xpath.compile(XPATH_TEST_VALUE);
                exprTestId = xpath.compile(XPATH_TEST_ID);
                exprTestTest = xpath.compile(XPATH_TEST_TEST);
                exprTestResult = xpath.compile(XPATH_TEST_RESULT);
                exprTestUnits = xpath.compile(XPATH_TEST_UNITS);
                exprTestInterpretation = xpath.compile(XPATH_TEST_INTERPRETATION);
                exprTestHigh = xpath.compile(XPATH_TEST_HIGH);
                exprTestLow = xpath.compile(XPATH_TEST_LOW);
                exprTestLoinc = xpath.compile(XPATH_TEST_LOINC);
                exprTestVuid = xpath.compile(XPATH_TEST_VUID);
                exprTestPerformingLab = xpath.compile(XPATH_TEST_PERFORMING_LAB);
//                exprDocuments = xpath.compile(XPATH_DOCUMENTS);
//                exprDocument = xpath.compile(XPATH_DOCUMENT);
//                exprDocumentLocalTitle = xpath.compile(XPATH_DOCUMENT_LOCAL_TITLE);
//                exprDocumentContent = xpath.compile(XPATH_CONTENT);
                exprResulted = xpath.compile(XPATH_RESULTED);
                exprProvider = xpath.compile(XPATH_PROVIDER);
                exprProviderCode = xpath.compile(XPATH_PROVIDER_CODE);
                exprProviderName = xpath.compile(XPATH_PROVIDER_NAME);
                exprProviderTaxonomyCode = xpath.compile(XPATH_PROVIDER_TAXONOMY_CODE);
                exprProviderProviderType = xpath.compile(XPATH_PROVIDER_PROVIDER_TYPE);
                exprProviderClassification = xpath.compile(XPATH_PROVIDER_CLASSIFICATION);
                exprProviderSpecialization = xpath.compile(XPATH_PROVIDER_SPECIALIZATION);
                exprProviderService = xpath.compile(XPATH_SERVICE_ATTR);
                exprProviderOfficePhone = xpath.compile(XPATH_OFFICE_PHONE);
                exprProviderAnalogPager = xpath.compile(XPATH_ANALOG_PAGER);
                exprProviderFax = xpath.compile(XPATH_FAX);
                exprProviderEmail = xpath.compile(XPATH_EMAIL);

            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }

            LabOrder labOrder = new LabOrder();
            fillCommonDataBeanValues(labOrder);

            String collected = null;
            String comment = null;
            String groupName = null;
            String specimen = null;
            String status = null;
            String name = null;
            String facilityName = null;
            String facilityCode = null;
            String type = null;

            try {
                collected = (String) exprCollected.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            String resulted = RPCUtil.evaluateNodeToString(panelNode, exprResulted, LOGGER);
            String labOrderId = RPCUtil.evaluateNodeToString(panelNode, exprId, LOGGER);
	    labOrder.setOrderId(labOrderId);

            try {
                comment = (String) exprComment.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                groupName = (String) exprGroupName.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                specimen = (String) exprSpecimen.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                status = (String) exprStatus.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                type = (String) exprType.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                name = (String) exprOrderName.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                facilityName = (String) exprFacilityName.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                facilityCode = (String) exprFacilityCode.evaluate(panelNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            Calendar collectedDate = null;

            if (!StringUtils.isEmpty(collected)) {
                collectedDate = Timestamp.toDateTime(collected);
            }

            Calendar resultedDate = null;
            if (!StringUtils.isEmpty(resulted)) {
                resultedDate = Timestamp.toDateTime(resulted);
            }

            if (StringUtils.isEmpty(name)) {
                //if name is still empty, use group name
                name = groupName;
            }
            labOrder.setTestName(name);

            Node providerNode = RPCUtil.evaluateNodeToNode(panelNode, exprProvider, LOGGER);
            if (providerNode != null) {
                String providerCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderCode, LOGGER);
                String providerName = RPCUtil.evaluateNodeToString(providerNode, exprProviderName, LOGGER);
                String providerTaxonomyCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderTaxonomyCode, LOGGER);
                String providerType = RPCUtil.evaluateNodeToString(providerNode, exprProviderProviderType, LOGGER);
                String providerClassification = RPCUtil.evaluateNodeToString(providerNode, exprProviderClassification, LOGGER);
                String providerSpecialization = RPCUtil.evaluateNodeToString(providerNode, exprProviderSpecialization, LOGGER);
                String providerService = RPCUtil.evaluateNodeToString(providerNode, exprProviderService, LOGGER);
                String providerOfficePhone = RPCUtil.evaluateNodeToString(providerNode, exprProviderOfficePhone, LOGGER);
                String providerAnalogPager = RPCUtil.evaluateNodeToString(providerNode, exprProviderAnalogPager, LOGGER);
                String providerFax = RPCUtil.evaluateNodeToString(providerNode, exprProviderFax, LOGGER);
                String providerEmail = RPCUtil.evaluateNodeToString(providerNode, exprProviderEmail, LOGGER);

                Provider provider = new Provider();
                provider.setCode(providerCode);
                provider.setName(providerName);
                provider.setTaxonomyCode(providerTaxonomyCode);
                provider.setProviderType(providerType);
                provider.setClassification(providerClassification);
                provider.setSpecialization(providerSpecialization);
                provider.setService(providerService);
                provider.setPhone(providerOfficePhone);
                provider.setAnalogPager(providerAnalogPager);
                provider.setFax(providerFax);
                provider.setEmail(providerEmail);

                labOrder.setRequestingProvider(provider);
            }

            Node resultNode = null;
            NodeList resultsNodeList = null;

            try {
                resultNode = (Node) exprValues.evaluate(panelNode, XPathConstants.NODE);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            if (resultNode != null) {
                try {
                    resultsNodeList = (NodeList) exprTestValue.evaluate(resultNode, XPathConstants.NODESET);
                } catch (XPathExpressionException e) {
                    logErrorMessage(e);
                }
            }

            ArrayList<LabResult> resultsList = new ArrayList<LabResult>();


            if (resultsNodeList != null && "CH".equalsIgnoreCase(type)) {
                int listLength = resultsNodeList.getLength();
                for (int j = 0; j < listLength; j++) {
                    Node testNode = resultsNodeList.item(j);
                    testNode.getParentNode().removeChild(testNode);

                    LabResult labResult = new LabResult();
                    fillCommonDataBeanValues(labResult);

                    labResult.setProvider(labOrder.getRequestingProvider());

                    String testId = null;
                    String testName = null;
                    String testResult = null;
                    String testUnits = null;
                    String testInterpretation = null;
                    String testHigh = null;
                    String testLow = null;
                    String testLoinc = null;
                    String testVuid = null;

                    try {
                        testId = (String) exprTestId.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testName = (String) exprTestTest.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testResult = (String) exprTestResult.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testUnits = (String) exprTestUnits.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testInterpretation = (String) exprTestInterpretation.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testHigh = (String) exprTestHigh.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testLow = (String) exprTestLow.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testLoinc = (String) exprTestLoinc.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testVuid = (String) exprTestVuid.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }

                    String testPerformingLab = RPCUtil.evaluateNodeToString(testNode, exprTestPerformingLab, LOGGER);

                    labResult.setTestId(testId);
                    labResult.setTestName(testName);
                    labResult.setResult(testResult);
                    labResult.setUnits(testUnits);
                    labResult.setInterpretation(testInterpretation);
                    labResult.setReferenceRange(testLow + " - " + testHigh);
                    List<Code> codeList = new ArrayList<Code>();
                    if (!StringUtils.isEmpty(testLoinc)) {
                        Code loincCode = new Code();
                        loincCode.setSystem("LOINC");
                        loincCode.setCode(testLoinc);
                        codeList.add(loincCode);
                    }
                    if (!StringUtils.isEmpty(testVuid)) {
                        Code vuidCode = new Code();
                        vuidCode.setSystem("VUID");
                        vuidCode.setCode(testVuid);
                        codeList.add(vuidCode);
                    }
                    labResult.setCodes(codeList.toArray(new Code[0]));
                    labResult.setPerformingLabName(testPerformingLab);

                    // Set the fields that match the order
                    labResult.setCollectedDate(collectedDate);
                    labResult.setResultDate(resultedDate);
                    labResult.setComment(comment);
                    labResult.setAccession(groupName);
                    labResult.setSpecimen(specimen);
                    labResult.setResultStatus(status);
                    labResult.setFacilityName(facilityName);
                    labResult.setStationNumber(facilityCode);
                    labResult.setOrderId(labOrder.getOrderId());

                    resultsList.add(labResult);
                }
            } else {
                // These are not CH labs so we just need one result
                LabResult labResult = new LabResult();
                fillCommonDataBeanValues(labResult);

                labResult.setProvider(labOrder.getRequestingProvider());

                // Set the fields that match the order
                labResult.setResultDate(resultedDate);
                labResult.setCollectedDate(collectedDate);
                labResult.setComment(comment);
                labResult.setAccession(groupName);
                labResult.setSpecimen(specimen);
                labResult.setResultStatus(status);
                labResult.setFacilityName(facilityName);
                labResult.setStationNumber(facilityCode);
                labResult.setTestName(labOrder.getTestName());

                resultsList.add(labResult);
            }

            if ("ACCESSION".equalsIgnoreCase(name)) {
                // if the order name is "ACCESSION", change it to the first test name from the resultsList
                StringBuilder labOrderNameSB = new StringBuilder();
                for (LabResult labResult : resultsList) {
                    labOrderNameSB.append("; ");
                    labOrderNameSB.append(labResult.getTestName());
                }
                labOrder.setTestName(labOrderNameSB.substring(2));
            }

            labOrder.setCollectionDate(collectedDate);
            labOrder.setAvailableDate(resultedDate);
            labOrder.setComment(comment);
            labOrder.setAccessionNumber(groupName);
            labOrder.setCollectionSample(specimen);
            labOrder.setStatus(status);
            labOrder.setSubscript(type);
            labOrder.setFacilityName(facilityName);
            labOrder.setStationNumber(facilityCode);

            labOrder.setLabResults(resultsList.toArray(new LabResult[0]));

            return labOrder;
        }
    }

    private class VPRAccessionParser implements Callable<LabOrder> {
        private Node accessionNode;

        public VPRAccessionParser(Node accessionNode) {
            this.accessionNode = accessionNode;
        }

        public LabOrder call() throws RPCException {
            XPathFactory xPathfactory;
            XPath xpath;

            XPathExpression exprCollected = null;
            XPathExpression exprComment = null;
            XPathExpression exprGroupName = null;
            XPathExpression exprSpecimen = null;
            XPathExpression exprStatus = null;
            XPathExpression exprType = null;
            XPathExpression exprName = null;
            XPathExpression exprId = null;
            XPathExpression exprFacilityName = null;
            XPathExpression exprFacilityCode = null;
            XPathExpression exprValues = null;
            XPathExpression exprTestValue = null;
            XPathExpression exprTestId = null;
            XPathExpression exprTestTest = null;
            XPathExpression exprTestResult = null;
            XPathExpression exprTestUnits = null;
            XPathExpression exprTestInterpretation = null;
            XPathExpression exprTestHigh = null;
            XPathExpression exprTestLow = null;
            XPathExpression exprTestLoinc = null;
            XPathExpression exprTestVuid = null;
            XPathExpression exprTestPerformingLab = null;
            XPathExpression exprDocuments = null;
            XPathExpression exprDocumentLocalTitle = null;
            XPathExpression exprResulted = null;
            XPathExpression exprDocument = null;
            XPathExpression exprDocumentContent = null;
            XPathExpression exprProvider = null;
            XPathExpression exprProviderCode = null;
            XPathExpression exprProviderName = null;
            XPathExpression exprProviderTaxonomyCode = null;
            XPathExpression exprProviderProviderType = null;
            XPathExpression exprProviderClassification = null;
            XPathExpression exprProviderSpecialization = null;
            XPathExpression exprProviderService = null;
            XPathExpression exprProviderOfficePhone = null;
            XPathExpression exprProviderAnalogPager = null;
            XPathExpression exprProviderFax = null;
            XPathExpression exprProviderEmail = null;

            xPathfactory = XPathFactory.newInstance();
            xpath = xPathfactory.newXPath();


            try {
                exprCollected = xpath.compile(XPATH_COLLECTED);
                exprComment = xpath.compile(XPATH_COMMENT);
                exprGroupName = xpath.compile(XPATH_GROUPNAME);
                exprSpecimen = xpath.compile(XPATH_SPECIMEN);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprType = xpath.compile(XPATH_TYPE);
                exprId = xpath.compile(XPATH_ORDER_CODE);
                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
                exprFacilityCode = xpath.compile(XPATH_FACILITY_CODE);
                exprValues = xpath.compile(XPATH_VALUES);
                exprTestValue = xpath.compile(XPATH_TEST_VALUE);
                exprTestId = xpath.compile(XPATH_TEST_ID);
                exprTestTest = xpath.compile(XPATH_TEST_TEST);
                exprTestResult = xpath.compile(XPATH_TEST_RESULT);
                exprTestUnits = xpath.compile(XPATH_TEST_UNITS);
                exprTestInterpretation = xpath.compile(XPATH_TEST_INTERPRETATION);
                exprTestHigh = xpath.compile(XPATH_TEST_HIGH);
                exprTestLow = xpath.compile(XPATH_TEST_LOW);
                exprTestLoinc = xpath.compile(XPATH_TEST_LOINC);
                exprTestVuid = xpath.compile(XPATH_TEST_VUID);
                exprTestPerformingLab = xpath.compile(XPATH_TEST_PERFORMING_LAB);
                exprDocuments = xpath.compile(XPATH_DOCUMENTS);
                exprDocument = xpath.compile(XPATH_DOCUMENT);
                exprDocumentLocalTitle = xpath.compile(XPATH_DOCUMENT_LOCAL_TITLE);
                exprDocumentContent = xpath.compile(XPATH_CONTENT);
                exprName = xpath.compile(XPATH_NAME);
                exprResulted = xpath.compile(XPATH_RESULTED);
                exprProvider = xpath.compile(XPATH_PROVIDER);
                exprProviderCode = xpath.compile(XPATH_PROVIDER_CODE);
                exprProviderName = xpath.compile(XPATH_PROVIDER_NAME);
                exprProviderTaxonomyCode = xpath.compile(XPATH_PROVIDER_TAXONOMY_CODE);
                exprProviderProviderType = xpath.compile(XPATH_PROVIDER_PROVIDER_TYPE);
                exprProviderClassification = xpath.compile(XPATH_PROVIDER_CLASSIFICATION);
                exprProviderSpecialization = xpath.compile(XPATH_PROVIDER_SPECIALIZATION);
                exprProviderService = xpath.compile(XPATH_SERVICE_ATTR);
                exprProviderOfficePhone = xpath.compile(XPATH_OFFICE_PHONE);
                exprProviderAnalogPager = xpath.compile(XPATH_ANALOG_PAGER);
                exprProviderFax = xpath.compile(XPATH_FAX);
                exprProviderEmail = xpath.compile(XPATH_EMAIL);

            } catch (XPathExpressionException e) {
                logErrorMessage(e);
                throw new RPCException(e);
            }

            LabOrder labOrder = new LabOrder();
            fillCommonDataBeanValues(labOrder);

            String collected = null;
            String comment = null;
            String groupName = null;
            String specimen = null;
            String status = null;
            String facilityName = null;
            String facilityCode = null;
            String type = null;

            try {
                collected = (String) exprCollected.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            String resulted = RPCUtil.evaluateNodeToString(accessionNode, exprResulted, LOGGER);
            String labOrderId = RPCUtil.evaluateNodeToString(accessionNode, exprId, LOGGER);

            try {
                comment = (String) exprComment.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                groupName = (String) exprGroupName.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                specimen = (String) exprSpecimen.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                status = (String) exprStatus.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                type = (String) exprType.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            String name = RPCUtil.evaluateNodeToString(accessionNode, exprName, LOGGER);
            Node documentsNode = RPCUtil.evaluateNodeToNode(accessionNode, exprDocuments, LOGGER);
            String documentContent = null;
            if (documentsNode != null) {
                NodeList documentNodeList = RPCUtil.evaluateNodeToNodeList(documentsNode, exprDocument, LOGGER);
                if (documentNodeList != null) {
                    int documentNodeListLength = documentNodeList.getLength();
                    StringBuilder documentContentSB = new StringBuilder();
                    for (int documentNum = 0; documentNum < documentNodeListLength; documentNum++) {
                        Node documentNode = documentNodeList.item(documentNum);
                        if (StringUtils.isEmpty(name)) {
                            //if name not present, fall back on document local title
                            name = RPCUtil.evaluateNodeToString(documentNode, exprDocumentLocalTitle, LOGGER);
                        }
                        // concat all document content
                        String content = RPCUtil.evaluateNodeToString(documentNode, exprDocumentContent, LOGGER);
                        documentContentSB.append(content);
                        documentContentSB.append("\n\n");
                    }
                    if (documentContentSB.length() > 0) {
                        documentContent = documentContentSB.toString();
                    }
                }
            }

            try {
                facilityName = (String) exprFacilityName.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            try {
                facilityCode = (String) exprFacilityCode.evaluate(accessionNode, XPathConstants.STRING);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            Calendar collectedDate = null;

            if (!StringUtils.isEmpty(collected)) {
                collectedDate = Timestamp.toDateTime(collected);
            }

            Calendar resultedDate = null;
            if (!StringUtils.isEmpty(resulted)) {
                resultedDate = Timestamp.toDateTime(resulted);
            }

            labOrder.setTestName(name);

            Node providerNode = RPCUtil.evaluateNodeToNode(accessionNode, exprProvider, LOGGER);
            if (providerNode != null) {
                String providerCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderCode, LOGGER);
                String providerName = RPCUtil.evaluateNodeToString(providerNode, exprProviderName, LOGGER);
                String providerTaxonomyCode = RPCUtil.evaluateNodeToString(providerNode, exprProviderTaxonomyCode, LOGGER);
                String providerType = RPCUtil.evaluateNodeToString(providerNode, exprProviderProviderType, LOGGER);
                String providerClassification = RPCUtil.evaluateNodeToString(providerNode, exprProviderClassification, LOGGER);
                String providerSpecialization = RPCUtil.evaluateNodeToString(providerNode, exprProviderSpecialization, LOGGER);
                String providerService = RPCUtil.evaluateNodeToString(providerNode, exprProviderService, LOGGER);
                String providerOfficePhone = RPCUtil.evaluateNodeToString(providerNode, exprProviderOfficePhone, LOGGER);
                String providerAnalogPager = RPCUtil.evaluateNodeToString(providerNode, exprProviderAnalogPager, LOGGER);
                String providerFax = RPCUtil.evaluateNodeToString(providerNode, exprProviderFax, LOGGER);
                String providerEmail = RPCUtil.evaluateNodeToString(providerNode, exprProviderEmail, LOGGER);

                Provider provider = new Provider();
                provider.setCode(providerCode);
                provider.setName(providerName);
                provider.setTaxonomyCode(providerTaxonomyCode);
                provider.setProviderType(providerType);
                provider.setClassification(providerClassification);
                provider.setSpecialization(providerSpecialization);
                provider.setService(providerService);
                provider.setPhone(providerOfficePhone);
                provider.setAnalogPager(providerAnalogPager);
                provider.setFax(providerFax);
                provider.setEmail(providerEmail);

                labOrder.setRequestingProvider(provider);
            }

            Node resultNode = null;
            NodeList resultsNodeList = null;

            try {
                resultNode = (Node) exprValues.evaluate(accessionNode, XPathConstants.NODE);
            } catch (XPathExpressionException e) {
                logErrorMessage(e);
            }

            if (resultNode != null) {
                try {
                    resultsNodeList = (NodeList) exprTestValue.evaluate(resultNode, XPathConstants.NODESET);
                } catch (XPathExpressionException e) {
                    logErrorMessage(e);
                }
            }

            ArrayList<LabResult> resultsList = new ArrayList<LabResult>();


            if (resultsNodeList != null && "CH".equalsIgnoreCase(type)) {
                int listLength = resultsNodeList.getLength();
                for (int j = 0; j < listLength; j++) {
                    Node testNode = resultsNodeList.item(j);
                    testNode.getParentNode().removeChild(testNode);

                    LabResult labResult = new LabResult();
                    fillCommonDataBeanValues(labResult);

                    labResult.setProvider(labOrder.getRequestingProvider());

                    String testId = null;
                    String testName = null;
                    String testResult = null;
                    String testUnits = null;
                    String testInterpretation = null;
                    String testHigh = null;
                    String testLow = null;
                    String testLoinc = null;
                    String testVuid = null;

                    try {
                        testId = (String) exprTestId.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testName = (String) exprTestTest.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testResult = (String) exprTestResult.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testUnits = (String) exprTestUnits.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testInterpretation = (String) exprTestInterpretation.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testHigh = (String) exprTestHigh.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testLow = (String) exprTestLow.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testLoinc = (String) exprTestLoinc.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    try {
                        testVuid = (String) exprTestVuid.evaluate(testNode, XPathConstants.STRING);
                    } catch (XPathExpressionException e) {
                        logErrorMessage(e);
                    }
                    String testPerformingLab = RPCUtil.evaluateNodeToString(testNode, exprTestPerformingLab, LOGGER);

                    labResult.setTestId(testId);
                    labResult.setTestName(testName);
                    labResult.setResult(testResult);
                    labResult.setUnits(testUnits);
                    labResult.setInterpretation(testInterpretation);
                    labResult.setReferenceRange(testLow + " - " + testHigh);
                    List<Code> codeList = new ArrayList<Code>();
                    if (!StringUtils.isEmpty(testLoinc)) {
                        Code loincCode = new Code();
                        loincCode.setSystem("LOINC");
                        loincCode.setCode(testLoinc);
                        codeList.add(loincCode);
                    }
                    if (!StringUtils.isEmpty(testVuid)) {
                        Code vuidCode = new Code();
                        vuidCode.setSystem("VUID");
                        vuidCode.setCode(testVuid);
                        codeList.add(vuidCode);
                    }
                    labResult.setCodes(codeList.toArray(new Code[0]));
                    labResult.setPerformingLabName(testPerformingLab);

                    // Set the fields that match the order
                    labResult.setCollectedDate(collectedDate);
                    labResult.setResultDate(resultedDate);
                    labResult.setComment(comment);
                    labResult.setAccession(groupName);
                    labResult.setSpecimen(specimen);
                    labResult.setResultStatus(status);
                    labResult.setFacilityName(facilityName);
                    labResult.setStationNumber(facilityCode);

                    resultsList.add(labResult);
                }
            } else {
                // These are not CH labs so we just need one result
                LabResult labResult = new LabResult();
                fillCommonDataBeanValues(labResult);

                labResult.setResult(documentContent);
                labResult.setProvider(labOrder.getRequestingProvider());

                // Set the fields that match the order
                labResult.setResultDate(resultedDate);
                labResult.setCollectedDate(collectedDate);
                labResult.setComment(comment);
                labResult.setAccession(groupName);
                labResult.setSpecimen(specimen);
                labResult.setResultStatus(status);
                labResult.setFacilityName(facilityName);
                labResult.setStationNumber(facilityCode);
                labResult.setTestName(labOrder.getTestName());

                resultsList.add(labResult);
            }

            labOrder.setCollectionDate(collectedDate);
            labOrder.setAvailableDate(resultedDate);
            labOrder.setComment(comment);
            labOrder.setAccessionNumber(groupName);
            labOrder.setCollectionSample(specimen);
            labOrder.setStatus(status);
            labOrder.setSubscript(type);
            labOrder.setFacilityName(facilityName);
            labOrder.setStationNumber(facilityCode);
            labOrder.setOrderId(labOrderId);

            labOrder.setLabResults(resultsList.toArray(new LabResult[0]));

            return labOrder;
        }
    }

    /**
     * Gets a list of Lab Orders using the VPR RPC. This will return the results for each
     * order embedded in the LabOrder.
     *
     * @param startDate The start of the date range for the list of orders.
     * @param endDate   The end of the date range for the list of orders.
     * @return The list of lab orders for the specified date range.
     * @throws RPCException
     */
    public LabResult[] getVPRLabResults(Calendar startDate, Calendar endDate) throws RPCException {
        return getVPRLabResults(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Gets a list of Lab Orders using the VPR RPC. This will return the results for each
     * order embedded in the LabOrder.
     *
     * @param patientIen The patient data file number
     * @param startDate  The start of the date range for the list of orders.
     * @param endDate    The end of the date range for the list of orders.
     * @return The list of lab orders for the specified date range.
     * @throws RPCException
     */
    public LabResult[] getVPRLabResults(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        LabOrder[] labOrders = getVPRLabOrders(patientIen, startDate, endDate);

        ArrayList<LabResult> labResultArrayList = new ArrayList<LabResult>();

        for (LabOrder labOrder: labOrders) {
            labResultArrayList.addAll(Arrays.asList(labOrder.getLabResults()));
        }

        return labResultArrayList.toArray(new LabResult[0]);
    }

//    /**
//     * Builds the VPR RPC request string using "panels" as the type. The panels contain the test results. The
//     * RPC will fail if the "VPR APPLICATION PROXY" context is not set. That is set when setting up the connection.
//     *
//     * @param patientIen Patient IEN
//     * @param startDate  Start date of range
//     * @param endDate    End date of range
//     * @return The RPC message string.
//     */
//    private String buildGetPatientLabResultsVPRRequest(String patientIen, Calendar startDate, Calendar endDate) {
//        VistaQuery vq = vistaQueryFactory.getVistaQuery("VPR GET PATIENT DATA");
//        vq.addParameter(vq.LITERAL, patientIen);
//        /* ORWRP REPORT TEXT (Lab Order Status)
//         *
//         * This RPC is used to retrieve information on lab orders and their status from the Lab Order Entry file (#69).
//         */
//        vq.addParameter(vq.LITERAL, "panels");
//
//        String startDateString = Timestamp.fromDateTime(startDate);
//        String[] flds = StringUtils.split(startDateString, StringUtils.PERIOD);
//        startDateString = flds[0];
//        String endDateString = Timestamp.fromDateTime(endDate);
//        flds = StringUtils.split(endDateString, StringUtils.PERIOD);
//        endDateString = flds[0];
//        vq.addParameter(vq.LITERAL, startDateString);
//        vq.addParameter(vq.LITERAL, endDateString);
//
//        return vq.buildMessage();
//    }

//    /**
//     * Parses the response from the VPR. This is the list of lab orders and the test results associated
//     * with those orders.
//     *
//     * @param patientIen The patient IEN.
//     * @param response   The response string from the RPC.
//     * @return Array of LabOrders, the results are embedded in each LabOrder as an array of LabResults.
//     */
//    private LabResult[] parseLabResultsFromVPR(String patientIen, String response) throws RPCException {
//        if ("".equals(response)) {
//            return new LabResult[0];
//        }
//
//        StringBuilder sb = new StringBuilder(Utils.stripInvalidXMLCharacters(response));
//        sb.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
//
//        ArrayList<LabResult> labsList = new ArrayList<LabResult>();
//
//        XPathFactory xPathfactory;
//        XPath xpath;
//
//        XPathExpression exprPanel = null;
//
//        xPathfactory = XPathFactory.newInstance();
//        xpath = xPathfactory.newXPath();
//
//        try {
//            exprPanel = xpath.compile(XPATH_PANEL);
//        } catch (XPathExpressionException e) {
//            logErrorMessage(e);
//            throw new RPCException(e);
//        }
//
//        Document doc = null;
//        try {
//
//            doc = parseXMLDocument(sb.toString());
//
//        } catch (Exception e) {
//            logErrorMessage(e);
//            throw new RPCException(e);
//        }
//
//        NodeList panelNodeList = null;
//        try {
//            panelNodeList = (NodeList) exprPanel.evaluate(doc, XPathConstants.NODESET);
//        } catch (XPathExpressionException e) {
//            logErrorMessage(e);
//            throw new RPCException(e);
//        }
//
//        List<VPRLabResultParser> nodeParserList = new ArrayList<VPRLabResultParser>();
//        int listLength = panelNodeList.getLength();
//        for (int i = 0; i < listLength; i++) {
//
//            Node panelNode = panelNodeList.item(i);
//            panelNode.getParentNode().removeChild(panelNode);
//
//
//            nodeParserList.add(new VPRLabResultParser(patientIen, panelNode));
//        }
//
//        ExecutorService executorService = Executors.newFixedThreadPool(10);
//        List<Future<LabResult[]>> futures = null;
//        try {
//            futures = executorService.invokeAll(nodeParserList, 2, TimeUnit.MINUTES);
//
//            if (futures == null)
//                throw new RPCException("Failed to parse labs results : lab node parser futures list is null.");
//
//        } catch (InterruptedException e) {
//            logErrorMessage(e);
//            throw new RPCException(e);
//        } finally {
//            executorService.shutdown();
//        }
//
//        Iterator<Future<LabResult[]>> futuresItr = futures.iterator();
//
//        while (futuresItr.hasNext()) {
//            try {
//                Future<LabResult[]> future = futuresItr.next();
//                if (future.isCancelled()) {
//                    throw new RPCException("Failed to parse VPR Lab Order Data");
//                } else labsList.addAll(Arrays.asList(future.get()));
//
//            } catch (InterruptedException e) {
//                logErrorMessage(e);
//            } catch (ExecutionException e) {
//                logErrorMessage(e);
//            }
//        }
//
//        return labsList.toArray(new LabResult[0]);
//    }

//    private class VPRLabResultParser implements Callable<LabResult[]> {
//        //private String patientIen;
//        private Node panelNode;
//
//        public VPRLabResultParser(String patientIen, Node panelNode) {
//            //this.patientIen = patientIen;
//            this.panelNode = panelNode;
//        }
//
//        public LabResult[] call() throws RPCException {
//            XPathFactory xPathfactory;
//            XPath xpath;
//
//            XPathExpression exprCollected = null, exprComment = null, exprGroupName = null,
//                    exprSpecimen = null, exprStatus = null, exprType = null, exprName = null,
//                    exprFacilityName = null, exprFacilityCode = null, exprValues = null, exprTestValue = null,
//                    exprTestId = null, exprTestTest = null, exprTestResult = null, exprTestUnits = null,
//                    exprTestInterpretation = null, exprTestHigh = null, exprTestLow = null,
//                    exprTestLoinc = null, exprTestVuid = null, exprTestDocuments = null,
//                    exprTestDocumentLocalTitle = null;
//
//            xPathfactory = XPathFactory.newInstance();
//            xpath = xPathfactory.newXPath();
//
//
//            try {
//                exprCollected = xpath.compile(XPATH_COLLECTED);
//                exprComment = xpath.compile(XPATH_COMMENT);
//                exprGroupName = xpath.compile(XPATH_GROUPNAME);
//                exprSpecimen = xpath.compile(XPATH_SPECIMEN);
//                exprStatus = xpath.compile(XPATH_STATUS);
//                exprType = xpath.compile(XPATH_TYPE);
//                exprName = xpath.compile(XPATH_NAME);
//                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
//                exprFacilityCode = xpath.compile(XPATH_FACILITY_CODE);
//                exprValues = xpath.compile(XPATH_VALUES);
//                exprTestValue = xpath.compile(XPATH_TEST_VALUE);
//                exprTestId = xpath.compile(XPATH_TEST_ID);
//                exprTestTest = xpath.compile(XPATH_TEST_TEST);
//                exprTestResult = xpath.compile(XPATH_TEST_RESULT);
//                exprTestUnits = xpath.compile(XPATH_TEST_UNITS);
//                exprTestInterpretation = xpath.compile(XPATH_TEST_INTERPRETATION);
//                exprTestHigh = xpath.compile(XPATH_TEST_HIGH);
//                exprTestLow = xpath.compile(XPATH_TEST_LOW);
//                exprTestLoinc = xpath.compile(XPATH_TEST_LOINC);
//                exprTestVuid = xpath.compile(XPATH_TEST_VUID);
//                exprTestDocuments = xpath.compile(XPATH_DOCUMENTS);
//                exprTestDocumentLocalTitle = xpath.compile(XPATH_DOCUMENT_LOCAL_TITLE);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//                throw new RPCException(e);
//            }
//
//            String collected = null, comment = null, groupName = null,
//                    specimen = null, status = null, name = null,
//                    facilityName = null, facilityCode = null, type = null;
//
//            try {
//                collected = (String) exprCollected.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                comment = (String) exprComment.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                groupName = (String) exprGroupName.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                specimen = (String) exprSpecimen.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                status = (String) exprStatus.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                type = (String) exprType.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                name = (String) exprName.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            //if name not present, fall back on document local title
//            if (StringUtils.isEmpty(name)) {
//                try {
//                    Node documentsNode = (Node) exprTestDocuments.evaluate(panelNode, XPathConstants.NODE);
//                    if(documentsNode != null)
//                        name = (String) exprTestDocumentLocalTitle.evaluate(documentsNode, XPathConstants.STRING);
//                } catch (XPathExpressionException e) {
//                    logErrorMessage(e);
//                }
//            }
//
//            try {
//                facilityName = (String) exprFacilityName.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            try {
//                facilityCode = (String) exprFacilityCode.evaluate(panelNode, XPathConstants.STRING);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            Calendar collectedDate = null;
//
//            if (!StringUtils.isEmpty(collected)) collectedDate = Timestamp.toDateTime(collected);
//
//            Node resultNode = null;
//            NodeList resultsNodeList = null;
//
//            try {
//                resultNode = (Node) exprValues.evaluate(panelNode, XPathConstants.NODE);
//            } catch (XPathExpressionException e) {
//                logErrorMessage(e);
//            }
//
//            if (resultNode != null) {
//                try {
//                    resultsNodeList = (NodeList) exprTestValue.evaluate(resultNode, XPathConstants.NODESET);
//                } catch (XPathExpressionException e) {
//                    logErrorMessage(e);
//                }
//            }
//
//            ArrayList<LabResult> resultsList = new ArrayList<LabResult>();
//
//            if (resultsNodeList != null && "CH".equalsIgnoreCase(type)) {
//                int listLength = resultsNodeList.getLength();
//                for (int j = 0; j < listLength; j++) {
//                    Node testNode = resultsNodeList.item(j);
//                    testNode.getParentNode().removeChild(testNode);
//
//                    LabResult labResult = new LabResult();
//                    fillCommonDataBeanValues(labResult);
//
//                    String testId = null, testName = null, testResult = null,
//                            testUnits = null, testInterpretation = null, testHigh = null, testLow = null,
//                            testLoinc = null, testVuid = null;
//
//                    try {
//                        testId = (String) exprTestId.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testName = (String) exprTestTest.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testResult = (String) exprTestResult.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testUnits = (String) exprTestUnits.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testInterpretation = (String) exprTestInterpretation.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testHigh = (String) exprTestHigh.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testLow = (String) exprTestLow.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testLoinc = (String) exprTestLoinc.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//                    try {
//                        testVuid = (String) exprTestVuid.evaluate(testNode, XPathConstants.STRING);
//                    } catch (XPathExpressionException e) {
//                        logErrorMessage(e);
//                    }
//
//                    labResult.setTestId(testId);
//                    labResult.setTestName(testName);
//                    labResult.setResult(testResult);
//                    labResult.setUnits(testUnits);
//                    labResult.setInterpretation(testInterpretation);
//                    labResult.setReferenceRange(testLow + " - " + testHigh);
//                    List<Code> codeList = new ArrayList<Code>();
//                    if (!StringUtils.isEmpty(testLoinc)) {
//                        Code loincCode = new Code();
//                        loincCode.setSystem("LOINC");
//                        loincCode.setCode(testLoinc);
//                        codeList.add(loincCode);
//                    }
//                    if (!StringUtils.isEmpty(testVuid)) {
//                        Code vuidCode = new Code();
//                        vuidCode.setSystem("VUID");
//                        vuidCode.setCode(testVuid);
//                        codeList.add(vuidCode);
//                    }
//                    labResult.setCodes(codeList.toArray(new Code[0]));
//
//                    // Set the fields that match the order
//                    labResult.setResultDate(collectedDate);
//                    labResult.setComment(comment);
//                    labResult.setAccession(groupName+"^"+type);
//                    labResult.setSpecimen(specimen);
//                    labResult.setResultStatus(status);
//                    labResult.setFacilityName(facilityName);
//                    labResult.setStationNumber(facilityCode);
//
//                    resultsList.add(labResult);
//                }
//            } else {
//                // These are not CH labs so we just need one result
//                LabResult labResult = new LabResult();
//                fillCommonDataBeanValues(labResult);
//
//                // Set the fields that match the order
//                labResult.setResultDate(collectedDate);
//                labResult.setComment(comment);
//                labResult.setAccession(groupName+"^"+type);
//                labResult.setSpecimen(specimen);
//                labResult.setResultStatus(status);
//                labResult.setFacilityName(facilityName);
//                labResult.setStationNumber(facilityCode);
//                labResult.setTestName(name);
//
//                resultsList.add(labResult);
//            }
//
//            return resultsList.toArray(new LabResult[0]);
//        }
//    }

}

