/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.patient.dao.rpc;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.RPCUtils;
import gov.va.med.rpc.StringUtils;
import gov.va.med.rpc.Timestamp;
import gov.va.med.rpc.VistaQuery;
import gov.va.med.vds.base.dao.rpc.BaseRPC;
import gov.va.med.vds.patient.dao.rpc.utils.RPCUtil;
import gov.va.med.vds.patient.dao.rpc.utils.VPRUtil;
import gov.va.med.vds.vistabeans.Code;
import gov.va.med.vds.vistabeans.Consult;
import gov.va.med.vds.vistabeans.Encounter;
import gov.va.med.vds.vistabeans.EncounterDocument;
import gov.va.med.vds.vistabeans.FreeTextReport;
import gov.va.med.vds.vistabeans.ProgressNote;
import gov.va.med.vds.vistabeans.Provider;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 *
 */
public class NotesRPC extends BaseRPC {
    private static final Logger LOGGER = LoggerFactory.getLogger(NotesRPC.class);

//    private static final String PROGRESS_NOTES_CATEGORY_FILTER = "PN";
    private static final String DISCHARGE_SUMMARIES_CATEGORY_FILTER = "DS";
    private static final String CONSULT_RESULTS_CATEGORY_FILTER = "CR";
//    private static final String CWAD_CATEGORY_FILTER = "CWAD";
    private static final String SURGERIES_CATEGORY_FILTER = "SR";
    private static final String CLINICAL_PROCEDURES_CATEGORY_FILTER = "CP";
    private static final String ACTIVE_STATUS = "ACTIVE";
    private static final String SCHEDULED_STATUS = "SCHEDULED";

    private static final Map<String, String> NOTETYPEMAP;
    static {
        Map<String, String> typeMap = new HashMap<String, String>();
        typeMap.put("A", "CWAD Note/Allergy");
        typeMap.put("D", "CWAD Note/Advance Directive");
        typeMap.put("C", "CWAD Note/Crisis Note");
        typeMap.put("W", "CWAD Note/Clinical Warning");
        typeMap.put("PN", "Progress Note");
        typeMap.put("CP", "Clinical Procedure");
        typeMap.put("SR", "Surgery");
        typeMap.put("LR", "Laboratory");
        typeMap.put("DS", "Discharge Summary");
        typeMap.put("CR", "Consult Result");
        typeMap.put("CWAD", "CWAD Notes/All");

        NOTETYPEMAP = Collections.unmodifiableMap(typeMap);
    }

    //VA consult status abbreviation to name mapping
    private static final Map<String, String> VA_STATUSMAP = new LinkedHashMap<String, String>() { 
            {
                put("dc", "DISCONTINUED");
                put("c", "COMPLETE");
                put("h", "HOLD");
                put("\"?\"", "FLAGGED");
                put("p", "PENDING");
                put("a", "ACTIVE");
                put("e", "EXPIRED");
                put("s", "SCHEDULED");
                put("pr", "PARTIAL RESULTS");
                put("dly", "DELAYED");
                put("u", "UNRELEASED");
                put("dce", "DISCONTINUED/EDIT");
                put("x", "CANCELLED");
                put("l", "LAPSED");
                put("rn", "RENEWED");
                put("'", "NO STATUS");
            }
        };

    public NotesRPC(RPCConnection cxn) {
        super(cxn);
    }

    //SB: missing location, vistDate, status
    public ProgressNote[] getOutpatientNotes(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        //passing zero returns all notes
        String request = buildGetNotesRequest(patientIen, Timestamp.fromDateTimeToTextTS(startDate), Timestamp.fromDateTimeToTextTS(endDate), 0);
        String response = (String) cxn.query(request);
        return toNotesFromRdv(response/*, patientIen*/);
    }

    public ProgressNote[] getOutpatientNotes(Calendar startDate, Calendar endDate) throws RPCException {
        return getOutpatientNotes(cxn.getPatientIen(), startDate, endDate);
    }

    private String buildGetNotesRequest(String patientIen, String fromDate, String toDate, int nNotes) {
        return RPCUtils.buildReportTextRequest(vistaQueryFactory, patientIen, fromDate, toDate, nNotes, "OR_PN:PROGRESS NOTES~TIUPRG;ORDV04;15;");
    }


    private ProgressNote[] toNotesFromRdv(String response/*, String patientIen*/) {
        ProgressNote[] rtc;
        if ("".equals(response)) {
            rtc = new ProgressNote[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<ProgressNote> lst = new ArrayList<ProgressNote>();
            ProgressNote note = null;
            StringBuilder noteText = new StringBuilder();
    
            for (int i = 0; i < lines.length; i++) {
                if ("".equals(lines[i])) {
                    continue;
                }
    
                String[] flds = StringUtils.split(lines[i], StringUtils.CARET);
    
                if (flds.length < 2) {
                    continue;
                }
    
                if ("[+]".equals(flds[1])) {
                    //note.Text = VistaDaoUtils.removeCtlChars(note.Text);
                    String text = "";
                    if (note != null) {
                        text = note.getNoteText();
                    }
    
                    text = RPCUtils.removeCtlChars(text);
                    if (note != null) {
                        note.setNoteText(text);
                    }
    
                    lst.add(note);
                } else if ("1".equals(flds[0])) {
                    note = new ProgressNote();
    //                String[] subflds = StringUtils.split(flds[1], StringUtils.SEMICOLON);
    //                KeyValuePair<String, String> site;
    //                if (subflds.length == 2)
    //                {
    //                    site = new KeyValuePair<String, String>(subflds[1], subflds[0]);
    //                }
    //                else if (!flds[1].equals(""))
    //                {
    //                    site = new KeyValuePair<String, String>((String) cxn.getSiteCode().getKey(), flds[1]);
    //                }
    //                else
    //                {
    //                    site = cxn.getSiteCode();
    //                }
    
                    fillCommonDataBeanValues(note);
                } else if ("2".equals(flds[0])) {
                    if (note != null) {
                        note.setNoteId(flds[1]);
                    }
                } else if ("3".equals(flds[0])) {
                    if (note != null) {
                        note.setNoteDate(Utils.getCal(Timestamp.toUtcFromRdv(flds[1])));
                    }
                } else if ("4".equals(flds[0])) {
                    //note.LocalTitle = flds[1];
                    if (note != null) {
                        note.setNoteType(flds[1]);
                    }
                } else if ("5".equals(flds[0])) {
                    if (note != null) {
                        Provider provider = new Provider();
                        provider.setName(flds[1]);
                        note.setProvider(provider);
                    }
                    //note.Author = new Author("", flds[1], "");
                } else if ("6".equals(flds[0]) && note != null) {
                    noteText.setLength(0);
                    noteText.append(note.getNoteText());
                    noteText.append(flds[1]);
                    noteText.append("\r\n");
                    note.setNoteText(noteText.toString());
                    /*
                    if (flds[1].startsWith("STANDARD TITLE"))
                    {
                        String[] parts = StringUtils.split(flds[1], StringUtils.COLON);
                        note.StandardTitle = parts[1].Trim();
                    }
                    */
                }
            }
    
            rtc = lst.toArray(new ProgressNote[0]);
        }
        return rtc;
    }

    //SB: Missing location
    public ProgressNote[] getInpatientNotes(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetDischargeSummariesRequest(patientIen, Timestamp.fromDateTimeToTextTS(startDate), Timestamp.fromDateTimeToTextTS(endDate), 0);
        String response = (String) cxn.query(request);
        return toNotesFromDischargeSummaries(response/*, patientIen*/);
    }

    private String buildGetDischargeSummariesRequest(String patientIen, String fromDate, String toDate, int nNotes) {
        //return VistaUtils.buildReportTextRequest(patientIen, fromDate, toDate, nNotes, "OR_DS:DISCHARGE SUMMARY REPORTS~TIUDCS;ORDV04;57;");
        return RPCUtils.buildReportTextRequest(vistaQueryFactory, patientIen, fromDate, toDate, nNotes, "OR_DS:DISCHARGE SUMMARY~TIUDCS;ORDV04;57;");
    }

    private ProgressNote[] toNotesFromDischargeSummaries(String response/*, String patientIen*/) {
        ProgressNote[] rtc;
        if ("".equals(response)) {
            rtc = new ProgressNote[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            List<ProgressNote> lst = new ArrayList<ProgressNote>();
            ProgressNote note = null;
            StringBuilder noteText = new StringBuilder();
            for (int i = 0; i < lines.length; i++) {
                if ("".equals(lines[i])) {
                    continue;
                }
    
                String[] flds = StringUtils.split(lines[i], StringUtils.CARET);
    
                if (flds.length < 2) {
                    continue;
                }
    
                if ("[+]".equals(flds[1])) {
                    if (note != null) {
                        lst.add(note);
                    }
                } else if ("1".equals(flds[0])) {
                    note = new ProgressNote();
                    //note.StandardTitle = "DISCHARGE SUMMARY";
                    note.setNoteType("Discharge Summary");
                    //note.Text = VistaDaoUtils.removeCtlChars(note.Text);
    //                String[] subflds = StringUtils.split(flds[1], StringUtils.SEMICOLON);
    //
    //                KeyValuePair site;
    //
    //                if (subflds.length == 2)
    //                {
    //                    site = new KeyValuePair<String, String>(subflds[1], subflds[0]);
    //                }
    //                else if (!flds[1].equals(""))
    //                {
    //                    site = new KeyValuePair<String, String>((String) cxn.getSiteCode().getKey(), flds[1]);
    //                }
    //                else
    //                {
    //                    site = cxn.getSiteCode();
    //                }
    
                    fillCommonDataBeanValues(note);
                } else if ("2".equals(flds[0])) {
                    //note.AdmitTimestamp = VistaTimestamp.toUtcFromRdv(flds[1]);
                    if (note != null) {
                        note.setVisitDate(Utils.formatDisplayDateTime(Utils.getCal(Timestamp.toUtcFromRdv(flds[1])).getTime()));
                    }
                } else if ("3".equals(flds[0])) {
                    //note.DischargeTimestamp = VistaTimestamp.toUtcFromRdv(flds[1]);
                    if (note != null) {
                        note.setNoteDate(Utils.getCal(Timestamp.toUtcFromRdv(flds[1])));
                    }
                } else if ("4".equals(flds[0])) {
                    //note.Author = new Author("", flds[1], "");
                    if (note != null) {
                        Provider provider = new Provider();
                        provider.setName(flds[1]);
                        note.setProvider(provider);
                    }
//                } else if ("5".equals(flds[0])) {
                    //note.ApprovedBy = new Author("", flds[1], "");
                } else if ("6".equals(flds[0])) {
                    // note.Status = flds[1];
                    if (note != null) {
                        note.setStatus(flds[1]);
                    }
                } else if ("7".equals(flds[0]) && note != null) {
                    noteText.setLength(0);
                    noteText.append(note.getNoteText());
                    noteText.append(flds[1]);
                    noteText.append("\r\n");
                    note.setNoteText(noteText.toString());
                }
            }
            rtc = lst.toArray(new ProgressNote[lst.size()]);
        }
        return rtc;
    }

    public ProgressNote[] getInpatientNotes(Calendar startDate, Calendar endDate) throws RPCException {
        return getInpatientNotes(cxn.getPatientIen(), startDate, endDate);
    }

    //SB: Status does not exactly map
    public Consult[] getOutpatientConsults(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = buildGetConsultsForPatientRequest(patientIen);
        String response = (String) cxn.query(request);
        Consult[] consults = toConsults(response/*, patientIen*/);
        ArrayList<Consult> consultFiltered = new ArrayList<Consult>();

        //filter by date range
        for (Consult consult : consults) {
            Calendar apptDate = consult.getRequestDate();
            if (apptDate.compareTo(startDate) >= 0 && apptDate.compareTo(endDate) <= 0) {
                consultFiltered.add(consult);
            }
        }

        // make the secondary call for each consult to get the Provider
        for (int i = 0; i < consultFiltered.size(); i++) {
            request = buildGetConsultRequest(consultFiltered.get(i).getId());
            response = (String) cxn.query(request);
            consultFiltered.get(i).setProvider(extractConsultProviderName(response));
        }

        return consultFiltered.toArray(new Consult[0]);
    }

    public Consult[] getOutpatientConsults(Calendar startDate, Calendar endDate) throws RPCException {
        return getOutpatientConsults(cxn.getPatientIen(), startDate, endDate);
    }

    private String buildGetConsultsForPatientRequest(String patientIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQCN LIST");
        vquery.addParameter(VistaQuery.LITERAL, patientIen);
        return vquery.buildMessage();
    }

    private String buildGetConsultRequest(String consultIen) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQCN GET CONSULT");
        vquery.addParameter(VistaQuery.LITERAL, consultIen);
        vquery.addParameter(VistaQuery.LITERAL, "0");
        return vquery.buildMessage();
    }

    private String extractConsultProviderName(String consult) {
        String rtc;
        if ("".equals(consult)
                || consult.startsWith("-1")) {
            rtc = "";
        } else {
            String[] lines = StringUtils.split(consult, StringUtils.CRLF);
            lines = StringUtils.trimArray(lines);
    
            String[] flds = StringUtils.split(lines[0], StringUtils.CARET);
            String[] subFlds = StringUtils.split(flds[13], StringUtils.SEMICOLON);
    
            rtc = subFlds[1];
        }
        return rtc;
    }

    public FreeTextReport getOutpatientConsultReport(String consultId) throws RPCException {
        String request = buildGetConsultReportRequest(consultId);
        String response = (String) cxn.query(request);

        FreeTextReport report = new FreeTextReport();
        report.setReportText(response);
        fillCommonDataBeanValues(report);

        return report;
    }

    private String buildGetConsultReportRequest(String consultId) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("ORQQCN DETAIL");
        vquery.addParameter(VistaQuery.LITERAL, consultId);
        return vquery.buildMessage();
    }

    private Consult[] toConsults(String response/*, String patientIen*/) {
        Consult[] rtc;
        if ("".equals(response)
                || response.startsWith("< PATIENT DOES NOT HAVE ANY CONSULTS/REQUESTS")) {
            rtc = new Consult[0];
        } else {
            String[] rex = StringUtils.split(response, StringUtils.CRLF);
            rex = StringUtils.trimArray(rex);
            Consult[] result = new Consult[rex.length];
            for (int i = 0; i < rex.length; i++) {
                String[] flds = StringUtils.split(rex[i], StringUtils.CARET);
                Consult cons = new Consult();
                cons.setId(flds[0]);
                cons.setRequestDate(Timestamp.toDateTime(flds[1]));
    
                //map VA consult status abbreviation to status name
                String status = flds[2];
    
                String mappedStatus = VA_STATUSMAP.get(status);
    
                if (!StringUtils.isEmpty(mappedStatus)) {
                    status = mappedStatus;
                }
    
                cons.setStatus(status);
                cons.setService(flds[3]);
                cons.setProcedureConsult(flds[4]);
                //c.Title = flds[6];
                //c.Service = new KeyValuePair<String, String>("", flds[2]);
                fillCommonDataBeanValues(cons);
                result[i] = cons;
            }
            rtc = result;
        }
        return rtc;
    }

    public FreeTextReport getVisitNotes(QueryBean queryBean) throws RPCException {

        // Find the TIU Document file IEN from the Visit IEN
        String request = buildVisitTIUIENRequest(queryBean.getItemId());
        String response = (String) cxn.query(request);
        String[] tiuDocFileIENs = toTIUDocFileIENs(response);

        // Get the Record Text
        StringBuilder recordTextSB = new StringBuilder("");
        for (String tiuIEN : tiuDocFileIENs) {
            request = buildTIUGetRecordTextRequest(tiuIEN);
            response = (String) cxn.query(request);

            recordTextSB.append(response);
            recordTextSB.append("\n");
        }

        FreeTextReport record = new FreeTextReport();
        record.setReportText(recordTextSB.toString());
        fillCommonDataBeanValues(record);

        return record;
    }

    private String buildVisitTIUIENRequest(String visitIEN) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("TIU NOTES BY VISIT");
        vquery.addParameter(VistaQuery.LITERAL, visitIEN);
        return vquery.buildMessage();
    }

    private String[] toTIUDocFileIENs(String response) {
        String[] rtc;
        if ("".equals(response)) {
            rtc = new String[0];
        } else {
            String[] lines = StringUtils.split(response, StringUtils.CRLF);
            ArrayList<String> lst = new ArrayList<String>();
//            String ien = null;
    
            for (String line : lines) {
                if ("".equals(line)) {
                    continue;
                }
    
                String[] flds = StringUtils.split(line, StringUtils.CARET);
    
                // the first field is the IEN
                if (!("".equals(flds[0]))) {
                    lst.add(flds[0]);
                }
            }
    
            rtc  = (String[]) lst.toArray(new String[0]);
        }
        return rtc;
    }

    private String buildTIUGetRecordTextRequest(String tiuFileDocIEN) {
        VistaQuery vquery = vistaQueryFactory.getVistaQuery("TIU GET RECORD TEXT");
        vquery.addParameter(VistaQuery.LITERAL, tiuFileDocIEN);
        return vquery.buildMessage();
    }

    /**
     * This method gets the detailed text of a TIU document from the TIU Document file (#8925).
     * It makes a call to the "TIU GET RECORD TEXT" RPC.<p/>
     * <p/>
     * This method can be used in conjunction with the getProgressNotesVPR() method. It can be
     * the secondary call to get the details of a progress note when the noteId is passed as the
     * parameter.
     *
     * @param docId TIU document ID
     * @return the text report for the document
     * @throws RPCException
     */
    public FreeTextReport getNotesById(String docId) throws RPCException {

        // Find the TIU Document file IEN from the Visit IEN
        String request = buildTIUGetRecordTextRequest(docId);
        String response = (String) cxn.query(request);

        FreeTextReport record = new FreeTextReport();
        record.setReportText(response);
        fillCommonDataBeanValues(record);

        return record;
    }

    public String getProgressNote(QueryBean queryBean) throws RPCException {
        String rtc = "";
        String noteID = queryBean.getItemId();

        // The text of the note is parsed when the list is retrieved
        // so we get the list again
        ProgressNote[] progressNotes = getOutpatientNotes(queryBean.getStartDate(), queryBean.getEndDate());

        // find the progress note we want and return the text of that note
        for (ProgressNote note : progressNotes) {
            if (note.getNoteId().equalsIgnoreCase(noteID)) {
                rtc = note.getNoteText();
                break;
            }
        }

        // if it didn't return from the for loop that means we couldn't match the note with the id
        return rtc;
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of discharge summaries. Will
     * return only documents that don't have have <code>DS</code> (Discharge Summaries) or <code>CR</code>
     * (Consult Results) as the <code>type</code>in the VPR response. The types in the VPR response for progress
     * notes are <code>A, D, C, W, PN, CP, SR, LR</code> for CWAD note/Allergy, CWAD note/Advance Directive,
     * CWAD note/Crisis Note, CWAD note/Clinical Warning, progress note, clinical procedure, surgery, and laboratory
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of documents
     * @param endDate    The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getProgressNotesVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        // Get PN, CR,, H&P(LOINC 34117), DS, CWAD (see SETUP in VPRDTIU CLASS = "3^244")
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, "", false);
//        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, "PN", false);
        String response = (String) cxn.query(request);

        ProgressNote[] progressNoteArray = parseDocumentsFromVPR(response);
        ArrayList<ProgressNote> progressNotesArrayList = new ArrayList<ProgressNote>(Arrays.asList(progressNoteArray));

        // Add Surgeries
        request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, SURGERIES_CATEGORY_FILTER, false);
        response = (String) cxn.query(request);

        ProgressNote[] surgeriesArray = parseDocumentsFromVPR(response);
        progressNotesArrayList.addAll(Arrays.asList(surgeriesArray));

        // Add Clinical Procedures
        request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, CLINICAL_PROCEDURES_CATEGORY_FILTER, false);
        response = (String) cxn.query(request);

        ProgressNote[] clinProcsArray = parseDocumentsFromVPR(response);
        progressNotesArrayList.addAll(Arrays.asList(clinProcsArray));

        return progressNotesArrayList.toArray(new ProgressNote[0]);
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of discharge summaries. Will
     * return only documents that don't have have <code>DS</code> (Discharge Summaries) or <code>CR</code>
     * (Consult Results) as the <code>type</code>in the VPR response. The types in the VPR response for progress
     * notes are <code>A, D, C, W, PN, CP, SR, LR</code> for CWAD note/Allergy, CWAD note/Advance Directive,
     * CWAD note/Crisis Note, CWAD note/Clinical Warning, progress note, clinical procedure, surgery, and laboratory
     *
     * @param startDate The start of the date range for the list of documents
     * @param endDate   The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getProgressNotesVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getProgressNotesVPR(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of discharge summaries. Will
     * return only documents that have <code>DS</code> as the <code>type</code>in the VPR response.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of documents
     * @param endDate    The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getDischargeSummariesVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, DISCHARGE_SUMMARIES_CATEGORY_FILTER, true);
        String response = (String) cxn.query(request);

        ProgressNote[] summaries = parseDocumentsFromVPR(response);

        EncountersRPC encountersRPC = new EncountersRPC(cxn);
        for (ProgressNote summary : summaries) {
            Encounter admission = encountersRPC.getPatientAdmission(patientIen, summary.getEncounter());

            if (admission != null && admission.getArrivalDateTime() != null) {
                String admissionDateString = Utils.formatDisplayDateTime(admission.getArrivalDateTime().getTime());
                summary.setVisitDate(admissionDateString);
    
            }
        }

        return summaries;
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of discharge summaries. Will
     * return only documents that have <code>DS</code> as the <code>type</code>in the VPR response.
     *
     * @param startDate The start of the date range for the list of documents
     * @param endDate   The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getDischargeSummariesVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getDischargeSummariesVPR(cxn.getPatientIen(), startDate, endDate);
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of consult results. Will
     * return only documents that have <code>CR</code> as the <code>type</code>in the VPR response. This is
     * not to be confused with the list of all the consults. It is only a list of consult result documents. If
     * there is a consult that does not have a resulted document, that consult will not be represented in this
     * list. The documents in this list must be cross-referenced to the consult by the document id. This list
     * does not contain the contents of the consult result.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of documents
     * @param endDate    The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getConsultResultsVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.DOCUMENTS, CONSULT_RESULTS_CATEGORY_FILTER, false);
        String response = (String) cxn.query(request);

        return parseDocumentsFromVPR(response);
    }

    /**
     * Uses the VPR RPC with "documents" as the type to get a list of consult results. Will
     * return only documents that have <code>CR</code> as the <code>type</code>in the VPR response. This is
     * not to be confused with the list of all the consults. It is only a list of consult result documents. If
     * there is a consult that does not have a resulted document, that consult will not be represented in this
     * list. The documents in this list must be cross-referenced to the consult by the document id. This list
     * does not contain the contents of the consult result.
     *
     * @param startDate The start of the date range for the list of documents
     * @param endDate   The end of the date range for the list of documents
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public ProgressNote[] getConsultResultsVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getConsultResultsVPR(cxn.getPatientIen(), startDate, endDate);
    }


    /**
     * Uses the VPR RPC with "consults" as the type to get a list of consult requests.
     *
     * @param patientIen The patient IEN
     * @param startDate  The start of the date range for the list of consults
     * @param endDate    The end of the date range for the list of consults
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public Consult[] getConsultRequestsVPR(String patientIen, Calendar startDate, Calendar endDate) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, startDate, endDate, VPRUtil.VPRType.CONSULTS);
        String response = (String) cxn.query(request);
        Consult[] consultRequests =  parseConsultsFromVPR(response);

        return consultRequests;
    }

    /**
     * Uses the VPR RPC with "consults" as the type to get a list of consult requests.
     *
     * @param startDate  The start of the date range for the list of consults
     * @param endDate    The end of the date range for the list of consults
     * @return The list of progress notes for the specified date range
     * @throws RPCException
     */
    public Consult[] getConsultRequestsVPR(Calendar startDate, Calendar endDate) throws RPCException {
        return getConsultRequestsVPR(cxn.getPatientIen(), startDate, endDate);
    }


    public Consult getOutpatientConsultRequestByIdVPR(String patientIen, String consultId) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, VPRUtil.VPRType.CONSULTS, consultId);
        String response = (String) cxn.query(request);

        Consult[] consultRequests =  parseConsultsFromVPR(response);

        Consult consult = null;
        if (consultRequests.length > 0) {
            consult = consultRequests[0];
        } 
//        else {
//            consult = new Consult();
//            fillCommonDataBeanValues(consult);
//        }
        return consult;
    }

    public Consult getOutpatientConsultRequestByIdVPR(String consultId) throws RPCException {
        return getOutpatientConsultRequestByIdVPR(cxn.getPatientIen(), consultId);
    }

    public Consult[] getActiveConsultRequestsVPR(String patientIen) throws RPCException {
        String request = VPRUtil.buildVPRRequest(vistaQueryFactory, patientIen, VPRUtil.VPRType.CONSULTS);
        String response = (String) cxn.query(request);
        Consult[] consultRequests =  parseConsultsFromVPR(response);

        ArrayList<Consult> activeConsults = new ArrayList<Consult>();
        for (Consult consult : consultRequests) {
            if (ACTIVE_STATUS.equalsIgnoreCase(consult.getStatus()) || SCHEDULED_STATUS.equalsIgnoreCase(consult.getStatus())) {
                activeConsults.add(consult);
            }
        }

        return activeConsults.toArray(new Consult[0]);
    }

    public Consult[] getActiveConsultRequestsVPR() throws  RPCException {
        return getActiveConsultRequestsVPR(cxn.getPatientIen());
    }

    private ProgressNote[] parseDocumentsFromVPR(String response) throws RPCException {
        ProgressNote[] rtc;
        if ("".equals(response)) {
            rtc = new ProgressNote[0];
        } else {
    
            List<ProgressNote> documentList = new ArrayList<ProgressNote>();
    
            final String XPATH_DOCUMENT = "//document";
    
            Document doc = null;
    
            try {
                doc = parseXMLDocument(response);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();
    
            XPathExpression exprDocument;
    
            try {
                exprDocument = xpath.compile(XPATH_DOCUMENT);
    
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList documentsNodeList = null;
            try {
                documentsNodeList = (NodeList) exprDocument.evaluate(doc, XPathConstants.NODESET);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            List<VPRDocumentParser> documentParserList = new ArrayList<VPRDocumentParser>();
            int listLength = documentsNodeList.getLength();
            for (int documentNum = 0; documentNum < listLength; documentNum++) {
                Node documentNode = documentsNodeList.item(documentNum);
                documentNode.getParentNode().removeChild(documentNode);
    
                documentParserList.add(new VPRDocumentParser(documentNode));
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(10);
            List<Future<ProgressNote>> futures = null;
            try {
                futures = executorService.invokeAll(documentParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse documents results : documents node parser futures list is null.");
                }
    
            } catch (InterruptedException e) {
                LOGGER.error(e.getMessage(), e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<ProgressNote>> futuresItr = futures.iterator();
    
            while (futuresItr.hasNext()) {
                try {
                    Future<ProgressNote> future = futuresItr.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR ProgressNote Data");
                    } else {
                        ProgressNote pnote = future.get();
                        if (pnote != null && !StringUtils.isEmpty(pnote.getNoteTitle())) {
                            documentList.add(pnote);
                        }
                    }
    
                } catch (InterruptedException e) {
                    LOGGER.error(e.getMessage(), e);
                } catch (ExecutionException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
    
            rtc = documentList.toArray(new ProgressNote[0]);
        }
        return rtc;
    }

    private class VPRDocumentParser implements Callable<ProgressNote> {
        private Node documentNode;

        public VPRDocumentParser(Node documentNode) {
            this.documentNode = documentNode;
        }

        public ProgressNote call() throws RPCException {
//            final String XPATH_CATEGORY = "./category/@value";
//            final String XPATH_DOCUMENTCLASS = "./documentClass/@value";
            final String XPATH_ENCOUNTER = "./encounter/@value";
//            final String XPATH_FACILITYCODE = "./facility/@code";
            final String XPATH_FACILITYNAME = "./facility/@name";
            final String XPATH_ID = "./id/@value";
            final String XPATH_IMAGES = "./images/@value";
            final String XPATH_LOCALTITLE = "./localTitle/@value";
            final String XPATH_REFERENCEDATETIME = "./referenceDateTime/@value";
            final String XPATH_STATUS = "./status/@value";
            final String XPATH_TYPE = "./type/@value";
            final String XPATH_CONTENT = "./content";

            final String XPATH_NATLTITLE_CODE = "./nationalTitle/@code";
            final String XPATH_NATLTITLE_NAME = "./nationalTitle/@name";
            final String XPATH_NATLTITLEROLE_CODE = "./nationalTitleRole/@code";
            final String XPATH_NATLTITLEROLE_NAME = "./nationalTitleRole/@name";
            final String XPATH_NATLTITLESERVICE_CODE = "./nationalTitleService/@code";
            final String XPATH_NATLTITLESERVICE_NAME = "./nationalTitleService/@name";
            final String XPATH_NATLTITLESETTING_CODE = "./nationalTitleSetting/@code";
            final String XPATH_NATLTITLESETTING_NAME = "./nationalTitleSetting/@name";
            final String XPATH_NATLTITLESUBJECT_CODE = "./nationalTitleSubject/@code";
            final String XPATH_NATLTITLESUBJECT_NAME = "./nationalTitleSubject/@name";
            final String XPATH_NATLTITLETYPE_CODE = "./nationalTitleType/@code";
            final String XPATH_NATLTITLETYPE_NAME = "./nationalTitleType/@name";

            final String XPATH_CLINICIANS = "./clinicians";
            final String XPATH_CLINICIAN = "./clinician";
            final String XPATH_CLINICIAN_CODE = "./@code";
            final String XPATH_CLINICIAN_NAME = "./@name";
            final String XPATH_CLINICIAN_ROLE = "./@role";
            final String XPATH_CLINICIAN_DATETIME = "./@dateTime";
            final String XPATH_CLINICIAN_SIGNATURE = "./@signature";
            final String XPATH_CLINICIAN_TAXONOMY_CODE = "./@taxonomyCode";
            final String XPATH_CLINICIAN_PROVIDER_TYPE = "./@providerType";
            final String XPATH_CLINICIAN_CLASSIFICATION = "./@classification";
            final String XPATH_CLINICIAN_SPECIALIZATION = "./@specialization";

            ProgressNote document = new ProgressNote();

//            XPathExpression exprCategory = null;
//            XPathExpression exprDocumentClass = null;
            XPathExpression exprEncounter = null;
//            XPathExpression exprFacilityCode = null;
            XPathExpression exprFacilityName = null;
            XPathExpression exprId = null;
            XPathExpression exprImages = null;
            XPathExpression exprLocalTitle = null;
            XPathExpression exprReferenceDate = null;
            XPathExpression exprStatus = null;
            XPathExpression exprType = null;
            XPathExpression exprClinicians = null;
            XPathExpression exprClinician = null;
            XPathExpression exprClinicianCode = null;
            XPathExpression exprClinicianName = null;
            XPathExpression exprClinicianRole = null;
            XPathExpression exprClinicianDateTime = null;
            XPathExpression exprClinicianSignature = null;
            XPathExpression exprClinicianTaxonomyCode = null;
            XPathExpression exprClinicianProviderType = null;
            XPathExpression exprClinicianClassification = null;
            XPathExpression exprClinicianSpecialization = null;
            XPathExpression exprNationalTitleCode = null;
            XPathExpression exprNationalTitleName = null;
            XPathExpression exprNationalTitleRoleCode = null;
            XPathExpression exprNationalTitleRoleName = null;
            XPathExpression exprNationalTitleServiceCode = null;
            XPathExpression exprNationalTitleServiceName = null;
            XPathExpression exprNationalTitleSettingCode = null;
            XPathExpression exprNationalTitleSettingName = null;
            XPathExpression exprNationalTitleSubjectCode = null;
            XPathExpression exprNationalTitleSubjectName = null;
            XPathExpression exprNationalTitleTypeCode = null;
            XPathExpression exprNationalTitleTypeName = null;
            XPathExpression exprContent = null;
            XPathFactory xPathfactory = XPathFactory.newInstance();
            XPath xpath = xPathfactory.newXPath();

            try {
//                exprCategory = xpath.compile(XPATH_CATEGORY);
//                exprDocumentClass = xpath.compile(XPATH_DOCUMENTCLASS);
                exprEncounter = xpath.compile(XPATH_ENCOUNTER);
//                exprFacilityCode = xpath.compile(XPATH_FACILITYCODE);
                exprFacilityName = xpath.compile(XPATH_FACILITYNAME);
                exprId = xpath.compile(XPATH_ID);
                exprImages = xpath.compile(XPATH_IMAGES);
                exprLocalTitle = xpath.compile(XPATH_LOCALTITLE);
                exprReferenceDate = xpath.compile(XPATH_REFERENCEDATETIME);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprType = xpath.compile(XPATH_TYPE);
                exprContent = xpath.compile(XPATH_CONTENT);

                exprNationalTitleCode = xpath.compile(XPATH_NATLTITLE_CODE);
                exprNationalTitleName = xpath.compile(XPATH_NATLTITLE_NAME);
                exprNationalTitleRoleCode = xpath.compile(XPATH_NATLTITLEROLE_CODE);
                exprNationalTitleRoleName = xpath.compile(XPATH_NATLTITLEROLE_NAME);
                exprNationalTitleServiceCode = xpath.compile(XPATH_NATLTITLESERVICE_CODE);
                exprNationalTitleServiceName = xpath.compile(XPATH_NATLTITLESERVICE_NAME);
                exprNationalTitleSettingCode = xpath.compile(XPATH_NATLTITLESETTING_CODE);
                exprNationalTitleSettingName = xpath.compile(XPATH_NATLTITLESETTING_NAME);
                exprNationalTitleSubjectCode = xpath.compile(XPATH_NATLTITLESUBJECT_CODE);
                exprNationalTitleSubjectName = xpath.compile(XPATH_NATLTITLESUBJECT_NAME);
                exprNationalTitleTypeCode = xpath.compile(XPATH_NATLTITLETYPE_CODE);
                exprNationalTitleTypeName = xpath.compile(XPATH_NATLTITLETYPE_NAME);

                exprClinicians = xpath.compile(XPATH_CLINICIANS);
                exprClinician = xpath.compile(XPATH_CLINICIAN);
                exprClinicianCode = xpath.compile(XPATH_CLINICIAN_CODE);
                exprClinicianName = xpath.compile(XPATH_CLINICIAN_NAME);
                exprClinicianRole = xpath.compile(XPATH_CLINICIAN_ROLE);
                exprClinicianDateTime = xpath.compile(XPATH_CLINICIAN_DATETIME);
                exprClinicianSignature = xpath.compile(XPATH_CLINICIAN_SIGNATURE);
                exprClinicianTaxonomyCode = xpath.compile(XPATH_CLINICIAN_TAXONOMY_CODE);
                exprClinicianProviderType = xpath.compile(XPATH_CLINICIAN_PROVIDER_TYPE);
                exprClinicianClassification = xpath.compile(XPATH_CLINICIAN_CLASSIFICATION);
                exprClinicianSpecialization = xpath.compile(XPATH_CLINICIAN_SPECIALIZATION);

            } catch (Exception e) {
                throw new RPCException(e);
            }

            String noteType = RPCUtil.evaluateNodeToString(documentNode, exprType, LOGGER);

            // If we are filtering for Progress Notes, then we should have all the documents from the VPR call, so we
            // need to exclude "DS" and "CR" types.
            // For DS and CR, we should only have those types of notes
//            if (PROGRESS_NOTES_CATEGORY_FILTER.equals(typeFilter)) {
//                if (DISCHARGE_SUMMARIES_CATEGORY_FILTER.equals(noteType) || CONSULT_RESULTS_CATEGORY_FILTER.equals(noteType)) {
//                    return null;
//                }
//            }

            String documentName = RPCUtil.evaluateNodeToString(documentNode, exprLocalTitle, LOGGER);
            String documentIen = RPCUtil.evaluateNodeToString(documentNode, exprId, LOGGER);
            String encounter = RPCUtil.evaluateNodeToString(documentNode, exprEncounter, LOGGER);
            String referenced = RPCUtil.evaluateNodeToString(documentNode, exprReferenceDate, LOGGER);
            String facilityName = RPCUtil.evaluateNodeToString(documentNode, exprFacilityName, LOGGER);
            String images = RPCUtil.evaluateNodeToString(documentNode, exprImages, LOGGER);
            String status = RPCUtil.evaluateNodeToString(documentNode, exprStatus, LOGGER);
            String content = RPCUtil.evaluateNodeToString(documentNode, exprContent, LOGGER);

            document.setNoteTitle(documentName);
            document.setNoteId(documentIen);
            if (noteType != null) {
                String noteTypeDescription = NOTETYPEMAP.get(noteType);
                if (noteTypeDescription != null) {
                    document.setNoteType(noteTypeDescription);
                } else {
                    document.setNoteType(noteType);
                }
            }
            document.setEncounter(encounter);
            document.setLocation(facilityName);
            document.setImages(images);
            document.setStatus(status);
            document.setNoteText(content);

            if (!StringUtils.isEmpty(referenced)) {
                Calendar referenceDate;
                referenceDate = Timestamp.toDateTime(referenced);
                document.setNoteDate(referenceDate);
            }

            // Get the national codes
            String nationalTitleCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleCode, LOGGER);
            String nationalTitleName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleName, LOGGER);
            String nationalTitleRoleCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleRoleCode, LOGGER);
            String nationalTitleRoleName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleRoleName, LOGGER);
            String nationalTitleServiceCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleServiceCode, LOGGER);
            String nationalTitleServiceName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleServiceName, LOGGER);
            String nationalTitleSettingCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleSettingCode, LOGGER);
            String nationalTitleSettingName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleSettingName, LOGGER);
            String nationalTitleSubjectCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleSubjectCode, LOGGER);
            String nationalTitleSubjectName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleSubjectName, LOGGER);
            String nationalTitleTypeCode = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleTypeCode, LOGGER);
            String nationalTitleTypeName = RPCUtil.evaluateNodeToString(documentNode, exprNationalTitleTypeName, LOGGER);

            List<Code> nationalCodeList = new ArrayList<Code>();
            if (!StringUtils.isEmpty(nationalTitleCode)) {
                Code natlTitle = new Code();
                natlTitle.setCode(nationalTitleCode);
                natlTitle.setSystem("nationalTitle");
                natlTitle.setDisplay(nationalTitleName);
                nationalCodeList.add(natlTitle);

                Code vuid = new Code();
                vuid.setCode(nationalTitleCode);
                vuid.setSystem("VUID");
                vuid.setDisplay(nationalTitleName);
                nationalCodeList.add(vuid);
            }
            if (!StringUtils.isEmpty(nationalTitleRoleCode)) {
                Code natlTitleRole = new Code();
                natlTitleRole.setCode(nationalTitleRoleCode);
                natlTitleRole.setSystem("nationalTitleRole");
                natlTitleRole.setDisplay(nationalTitleRoleName);
                nationalCodeList.add(natlTitleRole);
            }
            if (!StringUtils.isEmpty(nationalTitleServiceCode)) {
                Code natlTitleService = new Code();
                natlTitleService.setCode(nationalTitleServiceCode);
                natlTitleService.setSystem("nationalTitleService");
                natlTitleService.setDisplay(nationalTitleServiceName);
                nationalCodeList.add(natlTitleService);
            }
            if (!StringUtils.isEmpty(nationalTitleSettingCode)) {
                Code natlTitleSetting = new Code();
                natlTitleSetting.setCode(nationalTitleSettingCode);
                natlTitleSetting.setSystem("nationalTitleSetting");
                natlTitleSetting.setDisplay(nationalTitleSettingName);
                nationalCodeList.add(natlTitleSetting);
            }
            if (!StringUtils.isEmpty(nationalTitleSubjectCode)) {
                Code natlTitleSubject = new Code();
                natlTitleSubject.setCode(nationalTitleSubjectCode);
                natlTitleSubject.setSystem("nationalTitleSubject");
                natlTitleSubject.setDisplay(nationalTitleSubjectName);
                nationalCodeList.add(natlTitleSubject);
            }
            if (!StringUtils.isEmpty(nationalTitleTypeCode)) {
                Code natlTitleType = new Code();
                natlTitleType.setCode(nationalTitleTypeCode);
                natlTitleType.setSystem("nationalTitleType");
                natlTitleType.setDisplay(nationalTitleTypeName);
                nationalCodeList.add(natlTitleType);
            }
            document.setCodes(nationalCodeList.toArray(new Code[0]));

            // Get provider information (clinicians) Look for the "S" role.
            Node cliniciansNode = RPCUtil.evaluateNodeToNode(documentNode, exprClinicians, LOGGER);
            if (cliniciansNode != null) {
                NodeList clinicianNodeList = RPCUtil.evaluateNodeToNodeList(cliniciansNode, exprClinician, LOGGER);
                if (clinicianNodeList != null) {
                    int listLength = clinicianNodeList.getLength();
                    for (int clinicianNum = 0; clinicianNum < listLength; clinicianNum++) {
                        Node clinicianNode = clinicianNodeList.item(clinicianNum);
                        clinicianNode.getParentNode().removeChild(clinicianNode);

                        String role = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianRole, LOGGER);
                        if (!StringUtils.isEmpty(role) && "S".equalsIgnoreCase(role)) {
                            Provider provider = new Provider();
                            provider.setRole(role);
                            String providerName = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianName, LOGGER);
                            String providerIen = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianCode, LOGGER);
                            String providerDateTimeString = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianDateTime, LOGGER);
                            String providerSignature = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianSignature, LOGGER);
                            String providerTaxonomyCode = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianTaxonomyCode, LOGGER);
                            String providerType = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianProviderType, LOGGER);
                            String providerClassification = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianClassification, LOGGER);
                            String providerSpecialization = RPCUtil.evaluateNodeToString(clinicianNode, exprClinicianSpecialization, LOGGER);

                            if (!StringUtils.isEmpty(providerName)) {
                                provider.setName(providerName);
                            }
                            if (!StringUtils.isEmpty(providerIen)) {
                                document.setUserIen(providerIen);
                                provider.setCode(providerIen);
                            }

                            Calendar providerDate = null;
                            if (!StringUtils.isEmpty(providerDateTimeString)) {
                                providerDate = Timestamp.toDateTime(providerDateTimeString);
                                provider.setDateTime(providerDate);
                            }
                            if (!StringUtils.isEmpty(providerSignature)) {
                                provider.setSignatureBlock(providerSignature);
                            }
                            if (!StringUtils.isEmpty(providerTaxonomyCode)) {
                                provider.setTaxonomyCode(providerTaxonomyCode);
                            }
                            if (!StringUtils.isEmpty(providerType)) {
                                provider.setProviderType(providerType);
                            }
                            if (!StringUtils.isEmpty(providerClassification)) {
                                provider.setClassification(providerClassification);
                            }
                            if (!StringUtils.isEmpty(providerSpecialization)) {
                                provider.setSpecialization(providerSpecialization);
                            }

                            document.setProvider(provider);
                        }
                    }
                }
            }

            fillCommonDataBeanValues(document);
            return document;
        }
    }

    private Consult[] parseConsultsFromVPR(String response) throws RPCException {
        Consult[] rtc;
        if ("".equals(response)) {
            rtc = new Consult[0];
        } else {
            StringBuilder sbld = new StringBuilder(response);
            sbld.insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    
            ArrayList<Consult> consultList = new ArrayList<Consult>();
    
            final String XPATH_CONSULT = "//consult";
    
            Document document = null;
            XPathExpression exprConsult = null;
    
            try {
                document = parseXMLDocument(sbld.toString());
                XPathFactory xPathFactory = XPathFactory.newInstance();
                XPath xpath = xPathFactory.newXPath();
                exprConsult = xpath.compile(XPATH_CONSULT);
            } catch (Exception e) {
                throw new RPCException(e);
            }
    
            NodeList consultNodeList = null;
            try {
                consultNodeList = (NodeList) exprConsult.evaluate(document, XPathConstants.NODESET);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }
    
            List<VPRConsultParser> consultParserList = new ArrayList<VPRConsultParser>();
            int consultNodeListLength = consultNodeList.getLength();
            for (int consultNum = 0; consultNum < consultNodeListLength; consultNum++) {
                Node consultNode = consultNodeList.item(consultNum);
                consultNode.getParentNode().removeChild(consultNode);
    
                consultParserList.add(new VPRConsultParser(consultNode));
            }
    
            ExecutorService executorService = Executors.newFixedThreadPool(5);
            List<Future<Consult>> futures = null;
            try {
                futures = executorService.invokeAll(consultParserList, 2, TimeUnit.MINUTES);
    
                if (futures == null) {
                    throw new RPCException("Failed to parse consult VPR results : consult node parser futures list is null.");
                }
            } catch (InterruptedException e) {
                RPCUtil.logErrorMessage(LOGGER, e);
                throw new RPCException(e);
            } finally {
                executorService.shutdown();
            }
    
            Iterator<Future<Consult>> futureIterator = futures.iterator();
    
            while (futureIterator.hasNext()) {
                try {
                    Future<Consult> future = futureIterator.next();
                    if (future.isCancelled()) {
                        throw new RPCException("Failed to parse VPR Consult Data");
                    } else {
                        Consult consult = future.get();
    
                        // exclude null results
                        if (consult != null) {
                            consultList.add(consult);
                        }
                    }
                } catch (InterruptedException ie) {
                    RPCUtil.logErrorMessage(LOGGER, ie);
                } catch (ExecutionException ee) {
                    RPCUtil.logErrorMessage(LOGGER, ee);
                }
            }
    
            rtc = consultList.toArray(new Consult[0]);
        }
        return rtc;
    }

    private class VPRConsultParser implements Callable<Consult> {
        private Node consultNode;

        public VPRConsultParser(Node consultNode) {
            this.consultNode = consultNode;
        }

        public Consult call() throws Exception {
//            final String XPATH_FACILITY_CODE = "./facility/@code";
//            final String XPATH_FACILITY_NAME = "./facility/@name";
            final String XPATH_ID = "./id/@value";
//            final String XPATH_NAME = "./name/@value";
            final String XPATH_ORDER_ID = "./orderID/@value";
            final String XPATH_PROCEDURE = "./procedure/@value";
            final String XPATH_REQUESTED = "./requested/@value";
//            final String XPATH_RESULT = "./result/@value";
            final String XPATH_SERVICE = "./service/@value";
            final String XPATH_STATUS = "./status/@value";
            final String XPATH_TYPE = "./type/@value";
            final String XPATH_DOCUMENTS = "./documents";
            final String XPATH_DOCUMENT = "./document";
            final String XPATH_DOCUMENT_ID = "./@id";
            final String XPATH_DOCUMENT_LOCAL_TITLE = "./@localTitle";
            final String XPATH_DOCUMENT_NATIONAL_TITLE = "./@nationalTitle";
//            final String XPATH_DOCUMENT_CONTENT = "./content";

            final String XPATH_PROVIDER_CODE = "./provider/@code";
            final String XPATH_PROVIDER_NAME = "./provider/@name";
            final String XPATH_PROVIDER_ROLE = "./provider/@role";
            final String XPATH_PROVIDER_TAXOMOMY_CODE = "./provider/@taxonomyCode";
            final String XPATH_PROVIDER_PROVIDER_TYPE = "./provider/@providerType";
            final String XPATH_PROVIDER_CLASSIFICATION = "./provider/@classification";
            final String XPATH_PROVIDER_SPECIALIZATION = "./provider/@specialization";
            final String XPATH_PROVIDER_SERVICE = "./provider/@service";
            final String XPATH_PROVIDER_OFFICE_PHONE = "./provider/@officePhone";
            final String XPATH_PROVIDER_ANALOG_PAGER = "./provider/@analogPager";
            final String XPATH_PROVIDER_FAX = "./provider/@fax";
            final String XPATH_PROVIDER_EMAIL = "./provider/@email";
//            final String XPATH_PROVIDER_PRIMARY = "./provider/@primary";

            XPathExpression exprId = null;
            XPathExpression exprService = null;
            XPathExpression exprStatus = null;
//                    exprFacilityCode = null, exprFacilityName = null,
//            XPathExpression exprName = null;
            XPathExpression exprType = null;
            XPathExpression exprOrderId = null;
            XPathExpression exprProcedure = null;
            XPathExpression exprRequested = null;
            XPathExpression exprProviderCode = null;
            XPathExpression exprProviderName = null;
            XPathExpression exprProviderRole = null;
            XPathExpression exprProviderTaxonomyCode = null;
            XPathExpression exprProviderProviderType = null;
            XPathExpression exprProviderClassification = null;
            XPathExpression exprProviderSpecialization = null;
//                    exprProviderPrimary = null,
            XPathExpression exprProviderService = null;
            XPathExpression exprProviderOfficePhone = null;
            XPathExpression exprProviderAnalogPager = null;
            XPathExpression exprProviderFax = null;
            XPathExpression exprProviderEmail = null;
            XPathExpression exprDocuments = null;
            XPathExpression exprDocument = null;
            XPathExpression exprDocumentId = null;
            XPathExpression exprDocumentLocalTitle = null;
            XPathExpression exprDocumentNationalTitle = null;

            XPathFactory xPathFactory = XPathFactory.newInstance();
            XPath xpath = xPathFactory.newXPath();

            try {
//                exprFacilityCode = xpath.compile(XPATH_FACILITY_CODE);
//                exprFacilityName = xpath.compile(XPATH_FACILITY_NAME);
                exprId = xpath.compile(XPATH_ID);
//                exprName = xpath.compile(XPATH_NAME);
                exprOrderId = xpath.compile(XPATH_ORDER_ID);
                exprProcedure = xpath.compile(XPATH_PROCEDURE);
                exprRequested = xpath.compile(XPATH_REQUESTED);
                exprService = xpath.compile(XPATH_SERVICE);
                exprStatus = xpath.compile(XPATH_STATUS);
                exprType = xpath.compile(XPATH_TYPE);

                exprProviderCode = xpath.compile(XPATH_PROVIDER_CODE);
                exprProviderName = xpath.compile(XPATH_PROVIDER_NAME);
                exprProviderRole = xpath.compile(XPATH_PROVIDER_ROLE);
                exprProviderTaxonomyCode = xpath.compile(XPATH_PROVIDER_TAXOMOMY_CODE);
                exprProviderProviderType = xpath.compile(XPATH_PROVIDER_PROVIDER_TYPE);
                exprProviderClassification = xpath.compile(XPATH_PROVIDER_CLASSIFICATION);
                exprProviderSpecialization = xpath.compile(XPATH_PROVIDER_SPECIALIZATION);
                exprProviderService = xpath.compile(XPATH_PROVIDER_SERVICE);
                exprProviderOfficePhone = xpath.compile(XPATH_PROVIDER_OFFICE_PHONE);
                exprProviderAnalogPager = xpath.compile(XPATH_PROVIDER_ANALOG_PAGER);
                exprProviderFax = xpath.compile(XPATH_PROVIDER_FAX);
                exprProviderEmail = xpath.compile(XPATH_PROVIDER_EMAIL);
//                exprProviderPrimary = xpath.compile(XPATH_PROVIDER_PRIMARY);

                exprDocuments = xpath.compile(XPATH_DOCUMENTS);
                exprDocument = xpath.compile(XPATH_DOCUMENT);
                exprDocumentId = xpath.compile(XPATH_DOCUMENT_ID);
                exprDocumentLocalTitle = xpath.compile(XPATH_DOCUMENT_LOCAL_TITLE);
                exprDocumentNationalTitle = xpath.compile(XPATH_DOCUMENT_NATIONAL_TITLE);
            } catch (XPathExpressionException e) {
                throw new RPCException(e);
            }

            Consult consult = new Consult();
            fillCommonDataBeanValues(consult);

//            String facilityCode = RPCUtil.evaluateNodeToString(consultNode, exprFacilityCode);
//            String facilityName = RPCUtil.evaluateNodeToString(consultNode, exprFacilityName);
            String id = RPCUtil.evaluateNodeToString(consultNode, exprId, LOGGER);
//            String name = RPCUtil.evaluateNodeToString(consultNode, exprName, logger);
            String orderId = RPCUtil.evaluateNodeToString(consultNode, exprOrderId, LOGGER);
            String procedure = RPCUtil.evaluateNodeToString(consultNode, exprProcedure, LOGGER);
            String requestedString = RPCUtil.evaluateNodeToString(consultNode, exprRequested, LOGGER);
            String service = RPCUtil.evaluateNodeToString(consultNode, exprService, LOGGER);
            String status = RPCUtil.evaluateNodeToString(consultNode, exprStatus, LOGGER);
            String type = RPCUtil.evaluateNodeToString(consultNode, exprType, LOGGER);

            String providerCode = RPCUtil.evaluateNodeToString(consultNode, exprProviderCode, LOGGER);
            String providerName = RPCUtil.evaluateNodeToString(consultNode, exprProviderName, LOGGER);
            String providerRole = RPCUtil.evaluateNodeToString(consultNode, exprProviderRole, LOGGER);
            String providerTaxonomyCode = RPCUtil.evaluateNodeToString(consultNode, exprProviderTaxonomyCode, LOGGER);
            String providerProviderType = RPCUtil.evaluateNodeToString(consultNode, exprProviderProviderType, LOGGER);
            String providerClassification = RPCUtil.evaluateNodeToString(consultNode, exprProviderClassification, LOGGER);
            String providerSpecialization = RPCUtil.evaluateNodeToString(consultNode, exprProviderSpecialization, LOGGER);
            String providerService = RPCUtil.evaluateNodeToString(consultNode, exprProviderService, LOGGER);
            String providerOfficePhone = RPCUtil.evaluateNodeToString(consultNode, exprProviderOfficePhone, LOGGER);
            String providerAnalogPager = RPCUtil.evaluateNodeToString(consultNode, exprProviderAnalogPager, LOGGER);
            String providerFax = RPCUtil.evaluateNodeToString(consultNode, exprProviderFax, LOGGER);
            String providerEmail = RPCUtil.evaluateNodeToString(consultNode, exprProviderEmail, LOGGER);

//            String providerPrimary = RPCUtil.evaluateNodeToString(consultNode, exprProviderPrimary, logger);

            Node documentsNode = RPCUtil.evaluateNodeToNode(consultNode, exprDocuments, LOGGER);
            if (documentsNode != null) {
                NodeList documentNodeList = RPCUtil.evaluateNodeToNodeList(documentsNode, exprDocument, LOGGER);

                if (documentNodeList != null) {
                    int documentNodeListLength = documentNodeList.getLength();
                    List<EncounterDocument> documentList = new ArrayList<EncounterDocument>();
                    for (int documentNum = 0; documentNum < documentNodeListLength; documentNum++) {
                        Node documentNode = documentNodeList.item(documentNum);
                        documentNode.getParentNode().removeChild(documentNode);

                        EncounterDocument document = new EncounterDocument();
                        fillCommonDataBeanValues(document);

                        String documentId = RPCUtil.evaluateNodeToString(documentNode, exprDocumentId, LOGGER);
                        String documentLocalTitle = RPCUtil.evaluateNodeToString(documentNode, exprDocumentLocalTitle, LOGGER);
                        String documentNationalTitle = RPCUtil.evaluateNodeToString(documentNode, exprDocumentNationalTitle, LOGGER);

                        document.setId(documentId);
                        document.setLocalTitle(documentLocalTitle);
                        document.setNationalTitle(documentNationalTitle);

                        documentList.add(document);
                    }

                    if (!documentList.isEmpty()) {
                        consult.setDocuments(documentList.toArray(new EncounterDocument[0]));
                    }
                }
            }

            Calendar requested = null;
            if (requestedString != null) {
                requested = Timestamp.toDateTime(requestedString);
            }
            consult.setId(id);
            consult.setOrderId(orderId);
            consult.setProcedureConsult(procedure);
            consult.setRequestDate(requested);
            consult.setService(service);
            consult.setStatus(status);
            consult.setConsultType(type);

            Provider provider = new Provider();
            provider.setName(providerName);
            provider.setRole(providerRole);
            provider.setCode(providerCode);
            provider.setTaxonomyCode(providerTaxonomyCode);
            provider.setProviderType(providerProviderType);
            provider.setClassification(providerClassification);
            provider.setSpecialization(providerSpecialization);
            provider.setService(providerService);
            provider.setPhone(providerOfficePhone);
            provider.setAnalogPager(providerAnalogPager);
            provider.setFax(providerFax);
            provider.setEmail(providerEmail);
//            if (providerPrimary != null && "1".equals(providerPrimary)) {
//                provider.setPrimary(true);
//            }
            consult.setProvider(provider);

            return consult;
        }
    }

}
