/**
 * Janus 4.0f (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.AdmissionsDao;
import gov.va.med.vds.patient.dao.rpc.AdmissionsRPC;
import gov.va.med.vds.vistabeans.AdmissionDetail;
import gov.va.med.vds.vistabeans.Diagnosis;
import gov.va.med.vds.vistabeans.Encounter;
import gov.va.med.vds.vistabeans.PatientAdmission;
import gov.va.med.vds.vistabeans.Procedure;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class AdmissionsDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(AdmissionsDataService.class.getName());

    public List<PatientAdmission> getPatientAdmissions(QueryBean queryBean) throws VDSException
    {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> admsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_ADMISSIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_ADMISSIONS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = admsMap.entrySet();
            List<PatientAdmission> admList = new ArrayList<PatientAdmission>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object adms = entry.getValue();
                if (adms instanceof Exception) {
                    adms = generateConnectionUnavailableBeanPatientAdmission(entry.getKey(), (Exception) adms);
                }
                admList.addAll(Arrays.asList((PatientAdmission[]) adms));
            }

            msq = setupVPRContextMultiSourceQuery(queryBean);
            admsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_ADMISSIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_ADMISSIONS_ENCOUNTERS, new Object[]{queryBean});
            entries = admsMap.entrySet();
            ArrayList<Encounter> admEncList = new ArrayList<Encounter>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object admEncs = entry.getValue();
                if (admEncs instanceof Exception) {
                    admEncs = generateConnectionUnavailableBeanPatientAdmission(entry.getKey(), (Exception) admEncs);
                }
                admEncList.addAll(Arrays.asList((Encounter[]) admEncs));
            }

            AdmissionsRPC.addDataFromEncountersToAdmissions(admList.toArray(new PatientAdmission[0]), admEncList);

            return admList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public AdmissionDetail getAdmissionDetails(QueryBean queryBean) throws VDSException {
        validateQueryBean(queryBean);

        RPCConnection cxn = null;

        try {
//            Patient patient = queryBean.getPatient();
            String itemId = queryBean.getItemId();
            String siteCode = queryBean.getRecordSite().getSiteCode();
            if (itemId == null || itemId.length() <= 0) {
                throw new IllegalArgumentException("Item Id cannot be null or empty");
            } else if (StringUtils.isEmpty(siteCode)) {
                throw new IllegalArgumentException("Missing record site code");
            }
            cxn = setupSingleRemoteSourceQuery(queryBean);
            AdmissionsDao admDao = new AdmissionsDao(cxn);
            AdmissionDetail detail = admDao.getAdmissionDetails(queryBean);

            return detail;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<Diagnosis> getAdmissionDiagnosis(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> diagsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_ADMISSIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_ADMISSION_DIAGNOSIS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = diagsMap.entrySet();
            List<Diagnosis> diagList = new ArrayList<Diagnosis>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object diags = entry.getValue();
                if (diags instanceof Exception) {
                    diags = generateConnectionUnavailableBeanDiags(entry.getKey(), (Exception) diags);
                }
                diagList.addAll(Arrays.asList((Diagnosis[]) diags));
            }

            return diagList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<Procedure> getAdmissionProcedures(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> procMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_ADMISSIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_ADMISSION_PROCEDURES, new Object[]{queryBean});


            Set<Map.Entry<Site, Object>> entries = procMap.entrySet();
            List<Procedure> procList = new ArrayList<Procedure>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object procs = entry.getValue();
                if (procs instanceof Exception) {
                    procs = generateConnectionUnavailableBeanProcs(entry.getKey(), (Exception) procs);
                }
                procList.addAll(Arrays.asList((Procedure[]) procs));
            }
            return procList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private PatientAdmission[] generateConnectionUnavailableBeanPatientAdmission(Site site, Exception except) {
       
        PatientAdmission adm = new PatientAdmission();
        adm.setDrg(CONNECTION_UNAVAILABLE);
        adm.setCodingCompleteDate(Utils.backDate().toGregorianCalendar());
        adm.setDispositionDate(Utils.backDate().toGregorianCalendar());
        adm.setDrgDescription(CONNECTION_UNAVAILABLE);
        adm.setWard(CONNECTION_UNAVAILABLE);
        adm.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            adm.setDrg(UNABLE_TO_RETRIEVE);
            adm.setDrgDescription(UNABLE_TO_RETRIEVE);
        } else  {
            LOGGER.error(except.getMessage(), except);        //log if any other exception type
        }

        return new PatientAdmission[] {adm};
    }

    private Diagnosis[] generateConnectionUnavailableBeanDiags(Site site, Exception except) {
       
        Diagnosis diag = new Diagnosis();
        diag.setCode(CONNECTION_UNAVAILABLE);
        diag.setDate(Utils.backDate().toGregorianCalendar());
        diag.setDescription(CONNECTION_UNAVAILABLE);
        diag.setSite(site);
        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            diag.setDescription(UNABLE_TO_RETRIEVE);
            diag.setCode(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);       //log if any other exception type
        }
        return new Diagnosis[] {diag};
    }

    private Procedure[] generateConnectionUnavailableBeanProcs(Site site, Exception except) {
        Procedure proc = new Procedure();
        proc.setCode(CONNECTION_UNAVAILABLE);
        proc.setDate(Utils.backDate().toGregorianCalendar());
        proc.setDescription(CONNECTION_UNAVAILABLE);
        proc.setSite(site);
        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            proc.setDescription(UNABLE_TO_RETRIEVE);
            proc.setCode(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);       //log if any other exception type
        }
        return new Procedure[] {proc};
    }
}
