/*
 * Janus 4.0 (c)
 * Copyright (c) 2014 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.InsuranceDao;
import gov.va.med.vds.vistabeans.InsuranceBean;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class InsuranceDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(InsuranceDataService.class.getName());

    public List<InsuranceBean> getPatientInsurance(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> insuranceMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_INSURANCE,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_INSURANCE, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = insuranceMap.entrySet();
            List<InsuranceBean> insuranceList = new ArrayList<InsuranceBean>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object insurance = entry.getValue();
                if (insurance instanceof Exception) {
                    insurance = generateConnectionUnavailableBean(entry.getKey(), (Exception) insurance);
                }
                insuranceList.addAll(Arrays.asList((InsuranceBean[]) insurance));
            }

            return insuranceList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<InsuranceBean> getPatientInsuranceForSite(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        if (queryBean.getRecordSite().getSiteCode() == null || queryBean.getRecordSite().getSiteCode().isEmpty())
        {
            throw new IllegalArgumentException("record site code cannot be null.");
        }

        RPCConnection cxn = null;

        try {
            cxn = setupVPRContextSingleRemoteSourceQuery(queryBean);

            InsuranceDao insuranceDao = new InsuranceDao(cxn);
            List<InsuranceBean> insuranceList = new ArrayList<InsuranceBean>();
            InsuranceBean[] policies = insuranceDao.getPatientInsurance(queryBean);

            insuranceList.addAll(Arrays.asList((InsuranceBean[]) policies));
            return insuranceList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private InsuranceBean[] generateConnectionUnavailableBean(Site site, Exception except)
    {
        InsuranceBean insurance = new InsuranceBean();
        insurance.setInsuranceType(CONNECTION_UNAVAILABLE);
        insurance.setSubscriberId(CONNECTION_UNAVAILABLE);
        insurance.setEffectiveDate(Utils.backDate().toGregorianCalendar());
        insurance.setSite(site);

        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            insurance.setInsuranceType(UNABLE_TO_RETRIEVE);
            insurance.setSubscriberId(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);        //log if any other exception type
        }
        
        return new InsuranceBean[]{insurance};
    }
}
