/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.vds.patient.dataservice;

import gov.va.med.rpc.RPCConnection;
import gov.va.med.rpc.RPCException;
import gov.va.med.rpc.StringUtils;
import gov.va.med.vds.base.dao.UnsupportedRPCMethodException;
import gov.va.med.vds.base.dataservice.DataServiceBase;
import gov.va.med.vds.base.dataservice.VDSException;
import gov.va.med.vds.base.dataservice.VistaRPCConnectionManager;
import gov.va.med.vds.patient.dao.MedicationsDao;
import gov.va.med.vds.vistabeans.FreeTextReport;
import gov.va.med.vds.vistabeans.Medication;
import gov.va.med.vds.vistabeans.QueryBean;
import gov.va.med.vds.vistabeans.Site;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 *
 */
public class MedicationsDataService extends DataServiceBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(MedicationsDataService.class.getName());

    public List<Medication> getPatientMedications(QueryBean queryBean) throws VDSException
    {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupVPRContextMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> medsMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_MEDICATIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_MEDICATIONS, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = medsMap.entrySet();
            List<Medication> medsList = new ArrayList<Medication>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object meds = entry.getValue();
                if (meds instanceof Exception) {
                    meds = generateConnectionUnavailableBean(entry.getKey(), (Exception) meds);
                }
                medsList.addAll(Arrays.asList((Medication[]) meds));
            }

            return medsList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    public List<Medication> getPatientMedicationsNVA(QueryBean queryBean) throws VDSException {
        validateQueryBeanDateRange(queryBean);

        VistaRPCConnectionManager msq = null;

        try {
            msq = setupMultiSourceQuery(queryBean);
            LinkedHashMap<Site, Object> medsNVAMap = VistaRPCConnectionManager.execute(msq.getConnectionTable(),
                    VistaRPCConnectionManager.EXECUTION_CLASS_MEDICATIONS,
                    VistaRPCConnectionManager.EXECUTION_METHOD_PATIENT_MEDICATIONS_NVA, new Object[]{queryBean});

            Set<Map.Entry<Site, Object>> entries = medsNVAMap.entrySet();
            List<Medication> medsNVAList = new ArrayList<Medication>();

            for (Map.Entry<Site, Object> entry : entries) {
                Object meds = entry.getValue();
                if (meds instanceof Exception) {
                    meds = generateConnectionUnavailableBean(entry.getKey(), (Exception) meds);
                }
                medsNVAList.addAll(Arrays.asList((Medication[]) meds));
            }

            return medsNVAList;
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (msq != null) {
                try {
                    msq.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

//    public Prescription getPrescriptionFills(QueryBean queryBean) throws VDSException
//    {
//        queryParamsCheck(queryBean);
//
//        try
//        {
//            String itemId = queryBean.getItemId();
//            String siteCode = queryBean.getRecordSiteCode();
//            if (itemId == null || itemId.length() <= 0)
//            {
//                throw new IllegalArgumentException("Item Id cannot be null or empty");
//            }
//            else if (StringUtils.isNullOrEmpty(siteCode))
//            {
//                throw new IllegalArgumentException("Missing record site code");
//            }
//            RPCConnection cxn = setupSingleRemoteSourceQuery(queryBean);
//            MedicationsDao medsDao = new MedicationsDao(cxn);
//            Prescription prescription = medsDao.getPrescriptionFills(queryBean);
//            cxn.disconnect();
//            return prescription;
//        }
//        catch (Exception ex)
//        {
//
//            throw new VDSException(ex);
//        }
//    }

    public FreeTextReport getMedicationDetail(QueryBean queryBean) throws VDSException
    {
        validateQueryBean(queryBean);

        RPCConnection cxn = null;

        try {
            String itemId = queryBean.getItemId();
            String siteCode = queryBean.getRecordSite().getSiteCode();
            if (itemId == null || itemId.length() <= 0) {
                throw new IllegalArgumentException("Item Id cannot be null or empty");
            } else if (StringUtils.isEmpty(siteCode)) {
                throw new IllegalArgumentException("Missing record site code");
            }
            cxn = setupSingleRemoteSourceQuery(queryBean);
            MedicationsDao medsDao = new MedicationsDao(cxn);
            return medsDao.getMedicationDetail(queryBean);
        } catch (Exception ex) {
            throw new VDSException(ex);
        } finally {
            if (cxn != null) {
                try {
                    cxn.disconnect();
                } catch (RPCException e) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private Medication[] generateConnectionUnavailableBean(Site site, Exception except)
    {
        
        Medication med = new Medication();
        med.setDrugName(CONNECTION_UNAVAILABLE);
        med.setOrderDate(Utils.backDate().toGregorianCalendar());
        med.setLastFilledDate(Utils.backDate().toGregorianCalendar());
        med.setSite(site);
        if (except.getCause() instanceof UnsupportedRPCMethodException) {
            med.setDrugName(UNABLE_TO_RETRIEVE);
        } else {
            LOGGER.error(except.getMessage(), except);        //log if any other exception type
        }

        return new Medication[] {med};
    }
}
