/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.vds.vistabeans;

import java.util.Calendar;
import java.util.List;


/**
 * This class represents a Patient Admission.  This class has properties to represent procedures and diagnosis for an admission.  These attributes are only populated when requesting details for an admission and will not be populated when requesting a list/array of admissions.
 */
public class PatientAdmission extends DataBean {
    /**
     * Registration Number for this admission.
     */
    private String registerNumber;
    /**
     * Medical center division associated with this admission.
     */
    private String division;
    /**
     * Ward where the patient was last assigned for this admission.
     */
    private String ward;

    /**
     * Ward Id (IEN) of the patient location
     */
    private String wardId;
    /**
     * Date of  the admission.
     */
    private Calendar admissionDate;
    /**
     * Date the patient was discharged.
     */
    private Calendar dispositionDate;
    /**
     * Admitting user associated with this admission.
     */
    private Provider admittingProvider;
    /**
     * Primary DRG code associated with this admission.
     */
    private String drg;
    /**
     * Primary DRG description associated with this admission.
     */
    private String drgDescription;
    /**
     * Internal Id of the admission. This property will be used to retrieve details
     * of an admission.
     */
    private String admissionId;
    /**
     * Internal visit id associated with this admission.  Used to retrieve visit diagnosis
     * and/or procedures.
     */
    private String visitId;
    /**
     * Calendar coding was complete for this admission.
     */
    private Calendar codingCompleteDate;
    /**
     * Status of admission.
     */
    private String recordStatus;
    /**
     * Location of patient within the ward
     */
    private String roomBed;

    private String userIen;
    private Provider provider;
    private String admittingUserIen;

    private String specialty;
    private List<EncounterDocument> encounterDocuments;
    private Code reason;
    private String reasonNarrative;

    private String dischargeSummaryId;

    public PatientAdmission() {
        super();
    }
    
    public PatientAdmission(PatientAdmission adm) {
        super();
        registerNumber = adm.getRegisterNumber();
        division = adm.getDivision();
        ward = adm.getWard();
        admissionDate = adm.getAdmissionDate();
        admissionId = adm.getAdmissionId();
        admittingProvider = adm.getAdmittingProvider();
        codingCompleteDate = adm.getCodingCompleteDate();
        dispositionDate = adm.getDispositionDate();
        drg = adm.getDrg();
        drgDescription = adm.getDrgDescription();
        recordStatus = adm.getRecordStatus();
        visitId = adm.getVisitId();
    }

    /**
     * Registration Number of admission.
     */
    public String getRegisterNumber() {
        return registerNumber;
    }

    public void setRegisterNumber(String newRegisterNumber) {
        registerNumber = newRegisterNumber;
    }

    /**
     * Medical Center Division associated with the admission.
     */
    public String getDivision() {
        return division;
    }

    public void setDivision(String newDivision) {
        division = newDivision;
    }

    /**
     * Ward location name of admission.
     */
    public String getWard() {
        return ward;
    }

    public void setWard(String newWard) {
        ward = newWard;
    }

    /**
     * Date of  admission.
     */
    public Calendar getAdmissionDate() {
        return admissionDate;
    }

    public void setAdmissionDate(Calendar newAdmissionDate) {
        admissionDate = newAdmissionDate;
    }

    /**
     * Date of  discharge.
     */
    public Calendar getDispositionDate() {
        return dispositionDate;
    }

    public void setDispositionDate(Calendar newDispositionDate) {
        dispositionDate = newDispositionDate;
    }

    /**
     * Admitting User for the admission.
     */
    public Provider getAdmittingProvider() {
        return admittingProvider;
    }

    public void setAdmittingProvider(Provider admittingProvider) {
        this.admittingProvider = admittingProvider;
    }

    /**
     * Sets the admitting prov name. If the Provider has not yet been set, creates a new Provider.
     * Only the name will be set and
     *
     * @param provider The Provider's name
     */
    public void setAdmittingProvider(String provider) {
        this.admittingProvider = new Provider();
        this.admittingProvider.setName(provider);
    }

    /**
     * Primary DRG code associated with the admission.
     */
    public String getDrg() {
        return drg;
    }

    public void setDrg(String newDrg) {
        drg = newDrg;
    }

    /**
     * Primary DRG Description
     */
    public String getDrgDescription() {
        return drgDescription;
    }

    public void setDrgDescription(String newDrgDescription) {
        drgDescription = newDrgDescription;
    }

    /**
     * Internal id of admission. Used to retrieve admission details.
     */
    public String getAdmissionId() {
        return admissionId;
    }

    public void setAdmissionId(String newAdmissionId) {
        admissionId = newAdmissionId;
    }

    /**
     * Internal ID of visit associated with the admission.  Used to retieve discharge
     * summary of the admission.
     */
    public String getVisitId() {
        return visitId;
    }

    public void setVisitId(String newVisitId) {
        visitId = newVisitId;
    }

    /**
     * Calendar coding was completed for the admission.
     */
    public Calendar getCodingCompleteDate() {
        return codingCompleteDate;
    }

    public void setCodingCompleteDate(Calendar newCodingCompleteDate) {
        codingCompleteDate = newCodingCompleteDate;
    }

    /**
     * Record status of admission.
     */
    public String getRecordStatus() {
        return recordStatus;
    }

    public void setRecordStatus(String newRecordStatus) {
        recordStatus = newRecordStatus;
    }

    public String getRoomBed() {
        return roomBed;
    }

    public void setRoomBed(String roomBed) {
        this.roomBed = roomBed;
    }

    public String getWardId() {
        return wardId;
    }

    public void setWardId(String wardId) {
        this.wardId = wardId;
    }

    public String getAdmittingUserIen() {
        return admittingUserIen;
    }

    public void setAdmittingUserIen(String admittingUserIen) {
        this.admittingUserIen = admittingUserIen;
    }

    public String getUserIen() {
        return userIen;
    }

    public void setUserIen(String userIen) {
        this.userIen = userIen;
    }

    public Provider getProvider() {
        return provider;
    }

    public void setProvider(Provider provider) {
        this.provider = provider;
    }

    /**
     * Sets the provider name. If the Provider has not yet been set, creates a new Provider.
     * Only the name will be set
     *
     * @param provider The Provider's name
     */
    public void setProvider(String provider) {
        this.provider = new Provider();
        this.provider.setName(provider);
    }

    public String getSpecialty() {
        return specialty;
    }

    public void setSpecialty(String specialty) {
        this.specialty = specialty;
    }

    public List<EncounterDocument> getEncounterDocuments() {
        return encounterDocuments;
    }

    public void setEncounterDocuments(List<EncounterDocument> encounterDocuments) {
        this.encounterDocuments = encounterDocuments;
    }

    public Code getReason() {
        return reason;
    }

    public void setReason(Code reason) {
        this.reason = reason;
    }

    public String getReasonNarrative() {
        return reasonNarrative;
    }

    public void setReasonNarrative(String reasonNarrative) {
        this.reasonNarrative = reasonNarrative;
    }

    public String getDischargeSummaryId() {
        return dischargeSummaryId;
    }

    public void setDischargeSummaryId(String dischargeSummaryId) {
        this.dischargeSummaryId = dischargeSummaryId;
    }
}