/*
 * Janus 4.0 (c)
 * Copyright (c) 2014 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.hdd;

import gov.va.med.foundations.utilities.ARMLog;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.util.DataSourceFactory;
import gov.va.med.vds.webservice.Code;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Cache all terminology mapping codes into HashMap in memory
 * User: Yanhan Nie
 * Date: 4/9/14
 */
public class CodeMappingsCache {

    private static final Map<String, Code> MAPPINGS = new HashMap<String, Code>();
    private static boolean isLoaded = false;

    private static final Logger LOGGER = LoggerFactory.getLogger(CodeMappingsCache.class);

    static {
        try {
            loadDataFromDatabase();
        } catch (JMeadowsException e) {
            LOGGER.error(e.getMessage());
        }
    }

    private CodeMappingsCache() {
    }

    /**
     * Load mapping data from SQL Server
     * This should be loaded only once when application starts
     */
    public static void loadDataFromDatabase() throws JMeadowsException {
        if (!isLoaded || getSize() == 0) {
            LOGGER.info("call loadDataFromDatabase");

            Connection con = null;
            CallableStatement cstmt = null;
            ResultSet rset = null;

            try {
                // Load the SQLServerDriver class, build the
                // connection string, and get a connection

                con = DataSourceFactory.createDataSource().getConnection();

                LOGGER.info("Connected.");

                cstmt = con.prepareCall("{call mapCodeListAll()}");

                rset = cstmt.executeQuery();

                //iterate through mapped codes rows
                while (rset.next()) {

                    String name = rset.getString("name");

                    //skip and log any code mappings containing an error
                    if (name != null && name.contains("^")) {
                        LOGGER.info("Name is invalid: {}", ARMLog.cleanString(name));
                        continue;
                    }
                    String customMap = rset.getString("customMap");
                    String sourceCodeStr = rset.getString("sourceCode");
                    String targetCodeStr = rset.getString("targetCode");


                    SourceCodeMapKey key = new SourceCodeMapKey();
                    key.setCustomMap(customMap);
                    key.setSourceCode(sourceCodeStr);

                    Code code = new Code();
                    code.setCode(targetCodeStr);
                    code.setDisplay(name);

                    MAPPINGS.put(key.toString(), code);

                }
                isLoaded = true;

            } catch (SQLException e) {
                LOGGER.error("mappingsCache exception: " + e.getMessage());
                throw new JMeadowsException(e);
            } finally {
                try {
                    if (rset != null) {
                        rset.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }

                try {
                    if (cstmt != null) {
                        cstmt.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }

                try {
                    if (con != null) {
                        con.close();
                    }
                } catch (SQLException e) {
                    LOGGER.error(e.getMessage());
                }
            }
        }
    }


    /**
     * Get code by providing a SourceCodeMapKey object
     * @param source
     * @return
     */
    public static Code getCode(SourceCodeMapKey source) {
        try {
            loadDataFromDatabase();
        } catch (JMeadowsException e) {
            LOGGER.error(e.getMessage());
        }
        return MAPPINGS.get(source.toString());
    }


    /**
     * Get code
     * @param customMap
     * @param sourceCode
     * @param targetSystem
     * @return
     */
    public static Code getCode(String customMap, String sourceCode, String targetSystem) {
        try {
            loadDataFromDatabase();
        } catch (JMeadowsException e) {
            LOGGER.error(e.getMessage());
        }
        Code code;
        SourceCodeMapKey key = new SourceCodeMapKey();
        key.setCustomMap(customMap);
        key.setSourceCode(sourceCode);
        code = MAPPINGS.get(key.toString());
        if (code != null) {
            code.setSystem(targetSystem);

        }
        return code;
    }

    /**
     * Get size of the HashMap
     * @return
     */
    public static int getSize() {
        return MAPPINGS.size();
    }
}
