/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartAllergy;
import gov.va.med.vds.webservice.Allergy;
import gov.va.med.vds.webservice.AllergyDetail;
import gov.va.med.vds.webservice.Site;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Patient allergy data access object
 */
public class AllergyDao extends DaoBase {

    /**
     * Retrieves a list of patient allergies from all available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return A list of patient allergies.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Allergy> getPatientAllergies(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatAllergies", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
                includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientAllergies")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapAllergiesFromBHIEV4")
                                .location(bhieLoc)
                                .build()
                );

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientAllergies")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapAllergiesFromBHIEV4")
                                    .location(bhieLoc)
                                    .build()
                    );
                }


            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientAllergies")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        SmartAllergy smartAllergy = new SmartAllergy();

        Object[] algObjs = MultiLocationQuery.execute(
                configs, smartAllergy, queryBean);

        List<Allergy> allergies = null;

        if (algObjs != null) {
            allergies = Arrays.asList((Allergy[]) algObjs);

            //map allergy concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, allergies, CodingSystem.UMLS);
        }

        return allergies;
    }

    /**
     * Retrieves a list of patient allergies from all available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @return A list of patient allergies.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Allergy> getDODVLERPatientAllergies(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDODVLERAllergies", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {

                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
//                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
//                includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getVLERPatientAllergies")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("mapAllergiesFromBHIEV4")
                                .location(bhieLoc)
                                .build()
                );
            }
        }

        SmartAllergy smartAllergy = new SmartAllergy();

        Object[] algObjs = MultiLocationQuery.execute(
                configs, smartAllergy, queryBean);

        List<Allergy> allergies = null;

        if (algObjs != null) {
            allergies = Arrays.asList((Allergy[]) algObjs);

            //map allergy concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, allergies, CodingSystem.UMLS);
        }

        return allergies;
    }

    /**
     * Retrieves a patient allergy details from the provided record location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return AllergyDetail containing allergy report text.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public AllergyDetail getAllergyDetail(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("AllergyDetail", queryBean);
        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        AllergyDetail allergyDetail = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite,
                            queryBean.getPatient().getVistaSites())) {

                allergyDetail = sm_vistaHandler.getAllergyDetail(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return allergyDetail;
    }

}
