/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.foundations.utilities.SSLDebugUtilities;
import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.HDDDao;
import gov.va.med.jmeadows.dao.util.DataSourceFactory;
import gov.va.med.jmeadows.dataservicehandler.BHIERelayServiceHandler;
import gov.va.med.jmeadows.dataservicehandler.MVIServiceHandler;
import gov.va.med.jmeadows.dataservicehandler.PDWSServiceHandler;
import gov.va.med.jmeadows.dataservicehandler.VistaDataServiceHandler;
import gov.va.med.jmeadows.dataservicehandler.VistaDataVLERServiceHandler;
import gov.va.med.jmeadows.site.Sites;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vds.webservice.User;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.sql.DataSource;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Data Access Object abstract class.
 */
public abstract class DaoBase {
    protected static final String PROVIDER_QUERY_NOT_SUPPORTED_DOD = "Provider data is not yet supported for DoD users.";

    /**
     * Data Access Object Configurations *
     */
    protected static final AppConfig APP_CONFIG = AppConfig.getInstance();
    protected static final String BHIE_ERROR_SITE_MONIKER = APP_CONFIG.getBHIEErrorSiteMoniker();
    protected static final String BHIE_ERROR_SITE_NAME = APP_CONFIG.getBHIEErrorSiteName();
    protected static final String BHIE_ERROR_SITE_CODE = APP_CONFIG.getBHIEErrorSiteCode();
    protected static final String IDENTIFIER_UNAVAILABLE_FLAG = APP_CONFIG.getIdentifierUnavailableFlag();
    protected static final String SYSTEM_TYPE_CHCS = APP_CONFIG.getSystemTypeCHCS();
    protected static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    protected static final String SYSTEM_TYPE_VISTA = APP_CONFIG.getSystemTypeVISTA();
    protected static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
    protected static final String DOD_EDIPN_OID = APP_CONFIG.getDODEdipnOid();
    protected static final String LOCAL_VA_IEN_OID = APP_CONFIG.getLocalVAIenOid();
    protected static final String VISTA_ICN_OID = APP_CONFIG.getVistaIcnOid();
    protected static final String BHIE_CUSTODIAN_OID_AHLTA = APP_CONFIG.getBHIECustodianOidAHLTA();
    protected static final String SYSTEM_TYPE_DOD_ADAPTER = APP_CONFIG.getSystemTypeDODAdapter();
    protected static final String VISTA_WS_URL = APP_CONFIG.getVistaWsURL();
    protected static final String CHCSWSURL = APP_CONFIG.getCHCSWSURL();
    protected static final String BHIE_RELAY_SERVICE_URL = APP_CONFIG.getBHIERelayServiceURL();
    protected static final String PIXURL = APP_CONFIG.getPIXURL();
    protected static final String MVIURL = APP_CONFIG.getMVIURL();
    protected static final String ERROR_MSG_CONNECTION_UNAVAILABLE = APP_CONFIG.getErrorMsgConnectionUnavailable();
    protected static final String BHIE_CUSTODIAN_OID_VA = APP_CONFIG.getBHIECustodianOidVA();
    protected static final String BHIE_SOURCE_STATUS_REPORT_FLAG = APP_CONFIG.getDODAdapterSourceStatusReportFlag();
    protected static final String BHIE_CUSTODIAN_OID_TMDS = APP_CONFIG.getBHIECustodianOidTMDS();
    protected static final String BHIE_CUSTODIAN_OID_ESS = APP_CONFIG.getBHIECustodianOidESS();
    protected static final String BHIE_CUSTODIAN_OID_CHCS = APP_CONFIG.getBHIECustodianOidCHCS();
    protected static final String BHIE_CUSTODIAN_OID_SHARE = APP_CONFIG.getBHIECustodianOidSHARE();
    protected static final String BHIE_CUSTODIAN_OID_FHIE = APP_CONFIG.getBHIECustodianOidFHIE();
    protected static final String ERROR_MSG_EDR_DOCS = APP_CONFIG.getErrorMsgEDRDocs();
    protected static final boolean TRUST_ALL_SSL_HOSTNAME_AND_CERTS = APP_CONFIG.isTrustAllSSLHostnameAndCerts();

    //initialize webservice handlers
    protected static final VistaDataServiceHandler sm_vistaHandler = new VistaDataServiceHandler(VISTA_WS_URL);
    protected static final VistaDataVLERServiceHandler sm_vistaVLERHandler = new VistaDataVLERServiceHandler(VISTA_WS_URL);
//    protected static final CHCSServiceHandler sm_chcsHandler = new CHCSServiceHandler(CHCSWSURL);
    protected static final BHIERelayServiceHandler sm_bhieRelayServiceHandler = new BHIERelayServiceHandler(BHIE_RELAY_SERVICE_URL);
//    protected static final PIXManagerDataServiceHandler sm_pixDataServiceHandler = new PIXManagerDataServiceHandler(PIXURL);
    protected static final PDWSServiceHandler sm_pdwsHandler = new PDWSServiceHandler();
    protected static final MVIServiceHandler sm_mviHandler = new MVIServiceHandler(MVIURL);

    protected static final DataSource sm_dbConnectionPool = DataSourceFactory.createDataSource();
    protected static final HDDDao hddDao = new HDDDao(sm_dbConnectionPool);
    protected static final DataSource sm_dbConnectionPoolEncrypted = DataSourceFactory.createDataSourceEncrypted();

    //reference to Sites database
    protected static final Sites SITES = Sites.getInstance();

    private static final Logger LOGGER = LoggerFactory.getLogger(DaoBase.class);


    static {

        if (TRUST_ALL_SSL_HOSTNAME_AND_CERTS) {
            LOGGER.info("Relaxing SSL X509 certificate and hostname verification requirements.");
            SSLDebugUtilities.trustAllHostnames();
            SSLDebugUtilities.trustAllHttpsCertificates();
        }
    }

    /**
     * QueryBean parameter enumeration. For use with validateQueryBean helper method
     */
    protected static enum QueryBeanParams {
        USER, PATIENT, RECORD_SITE_CODE, ITEM_ID, DATE_RANGE, STATUS, ACTIVE
    }

    /**
     * Helper method that throws an IllegalArgumentException if any of the specified QueryBeanParams values
     * are missing from the provided QueryBean instance.
     *
     * @param queryBean      QueryBean to check.
     * @param checkParamArgs QueryBeanParameters to check.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    protected void validateQueryBean(JMeadowsQuery queryBean, QueryBeanParams... checkParamArgs) {
        if (queryBean == null) {
            throw new IllegalArgumentException("QueryBean param is null or empty");
        }

        if (checkParamArgs == null) {
            throw new IllegalArgumentException("Missing required params");
        }

        List<QueryBeanParams> checkParams = Arrays.asList(checkParamArgs);

        boolean checkProvider = checkParams.contains(QueryBeanParams.USER);
        boolean checkPatient = checkParams.contains(QueryBeanParams.PATIENT);
        boolean checkRecordSite = checkParams.contains(QueryBeanParams.RECORD_SITE_CODE);
        boolean checkItemId = checkParams.contains(QueryBeanParams.ITEM_ID);
        boolean checkDateRange = checkParams.contains(QueryBeanParams.DATE_RANGE);
        boolean checkStatus = checkParams.contains(QueryBeanParams.STATUS);
        boolean checkActive = checkParams.contains(QueryBeanParams.ACTIVE);

        StringBuilder errorMsgSB = new StringBuilder("QueryBean is missing required parameter(s): ");

        ArrayList<String> missingParams = new ArrayList<String>();

        if (checkProvider && (queryBean.getUser() == null)) {
            missingParams.add("user");
        } else if (checkProvider && (queryBean.getUser().getUserIen() == null
                || queryBean.getUser().getUserIen().isEmpty())) {
            missingParams.add("user user id");
        } else if (checkProvider && (queryBean.getUser().getHostSite() == null)) {
            missingParams.add("user login site");
        }

        if (checkPatient && (queryBean.getPatient() == null)) {
            missingParams.add("patient");
        } else if (checkPatient) {        //throw error if patient missing all identifiers
            boolean hasEDIPI = StringUtils.isNotBlank(queryBean.getPatient().getEDIPI());
            boolean hasICN = StringUtils.isNotBlank(queryBean.getPatient().getICN());
            boolean hasVistaIENs =
                    queryBean.getPatient().getPatientIens() != null && !queryBean.getPatient().getPatientIens().isEmpty()
                            && queryBean.getPatient().getVistaSites() != null && !queryBean.getPatient().getVistaSites().isEmpty();

            if (!hasEDIPI && !hasICN && !hasVistaIENs) {
                missingParams.add("patient identifiers");
            }
        }

        if (checkItemId && (queryBean.getItemId() == null || queryBean.getItemId().isEmpty())) {
            missingParams.add("itemId");
        }

        if (checkRecordSite) {
            if (StringUtils.isBlank(queryBean.getRecordSiteCode())) {
                missingParams.add("recordSite");
            } else if (SITES.getSite(queryBean.getRecordSiteCode()) == null) {
                missingParams.add("Record site (" + queryBean.getRecordSiteCode() + ") not supported");
            }
        }

        if (checkDateRange && (queryBean.getStartDate() == null || queryBean.getEndDate() == null)) {
            missingParams.add("startDate");
            missingParams.add("endDate");
        }

        if (checkStatus && (queryBean.getStatus() == null || queryBean.getStatus().isEmpty())) {
            missingParams.add("status");
        }

        if (checkActive && queryBean.getActive() == null) {
            missingParams.add("active");
        }

        int missingParamsSize = missingParams.size();
        for (int i = 0; i < missingParamsSize; i++) {
            errorMsgSB.append(missingParams.get(i));

            if (i < (missingParamsSize - 1)) {
                errorMsgSB.append(", ");
            }
        }

        if (missingParamsSize > 0) {
            throw new IllegalArgumentException(errorMsgSB.toString());
        }
    }

    /**
     * Helper method that determines if patient identifier was flagged as unavailable by the patient select method.
     * A patient identifier is marked as unavailable due to a PIX Service failure.
     *
     * @param patientId
     * @return true if patient identifier was flagged as unavailable by the patient select method.
     */
    protected boolean isPatientIdentifierUnavailable(String patientId) {
        return IDENTIFIER_UNAVAILABLE_FLAG.equalsIgnoreCase(patientId);
    }


    /**
     * Audits a query.
     *
     * @param inqueryName Name of query to audit.
     * @param qbean        Query data that is sent to the audit database.
     * @throws JMeadowsException if a serious error occurs.
     */
    protected void auditQuery(String inqueryName, JMeadowsQuery qbean) throws JMeadowsException {
        try {
            auditQuery(inqueryName,"", qbean);
        } catch (Exception e) {
            throw new JMeadowsException(e);
        }
    }

    /**
     * Audits a query.
     *
     * @param inqueryName Name of query to audit.
     * @param info misc information to audit.
     * @param qbean        Query data that is sent to the audit database.
     * @throws JMeadowsException if a serious error occurs.
     */
    protected void auditQuery(String inqueryName, String info, JMeadowsQuery qbean) throws JMeadowsException {
        String queryName = inqueryName;
        String siteCode = null;
        String userIen = null;
        String userNPI = null;
        String userName = null;
        String patId = null;
        String emailAddress = null;
        String cardId = null;

        String requestingApp = qbean.getRequestingApp();
        User user = qbean.getUser();
        Patient patient = qbean.getPatient();

        String sDate = "";
        String eDate = "";

        if (user != null) {
            siteCode = (String) user.getHostSite().getSiteCode();
            userIen = user.getUserIen();
            emailAddress = user.getEmail();
            cardId = user.getCardId();

            //append providerIen to userIen if one exists
            if (StringUtils.isNotBlank(user.getProviderIen())) {
                userIen += "^" + user.getProviderIen();
            }

            userNPI = user.getNPI();
            userName = user.getName();
        }

        if (patient != null) {
            patId = patient.getEDIPI();
        }

        if (qbean.getStartDate() != null) {
            sDate = Utils.formatSQLDateTime(qbean.getStartDate().getTime());
        }
        if (qbean.getEndDate() != null) {
            eDate = Utils.formatSQLDateTime(qbean.getEndDate().getTime());
        }

        if (StringUtils.isNotBlank(qbean.getItemId())) {
            info += "^" + qbean.getItemId();
        }

        if (StringUtils.isNotBlank(qbean.getRecordSiteCode())) {
            Site site = getSite(qbean.getRecordSiteCode());

            if (site != null) {
                info += "^" + site.getAgency() + "^" + site.getMoniker();
            }
        }

        try {
            AuditDao auditDao = new AuditDao();
            auditDao.setAudit(siteCode, userIen, userNPI, userName,
                    patId, queryName, requestingApp, sDate, eDate, info, cardId, emailAddress);
        } catch (Exception e) {
            throw new JMeadowsException(e);
        }
    }

    /**
     * Audits a query.
     *
     * @param inqueryName Name of query to audit.
     * @param qbean        Query data that is sent to the audit database.
     * @throws JMeadowsException if a serious error occurs.
     */
    protected void auditQueryEncrypted(String inqueryName, JMeadowsQuery qbean) throws JMeadowsException {
        String queryName = inqueryName;
        String siteCode = null;
        String userIen = null;
        String userNPI = null;
        String userName = null;
        String patId = null;
        String info = "";

        String requestingApp = qbean.getRequestingApp();
        User user = qbean.getUser();
        Patient patient = qbean.getPatient();

        String sDate = "";
        String eDate = "";

        if (user != null) {
            siteCode = (String) user.getHostSite().getSiteCode();
            userIen = user.getUserIen();

            //append providerIen to userIen if one exists
            if (StringUtils.isNotBlank(user.getProviderIen())) {
                userIen += "^" + user.getProviderIen();
            }

            userNPI = user.getNPI();
            userName = user.getName();
        }

        if (patient != null) {
            patId = patient.getEDIPI();
        }

        if (qbean.getStartDate() != null) {
            sDate = Utils.formatSQLDateTime(qbean.getStartDate().getTime());
        }
        if (qbean.getEndDate() != null) {
            eDate = Utils.formatSQLDateTime(qbean.getEndDate().getTime());
        }

        if (StringUtils.isNotBlank(qbean.getItemId())) {
            info += "^" + qbean.getItemId();
        }

        if (StringUtils.isNotBlank(qbean.getRecordSiteCode())) {
            Site site = getSite(qbean.getRecordSiteCode());

            if (site != null) {
                info += "^" + site.getAgency() + "^" + site.getMoniker();
            }
        }

        try {
            AuditDao auditDao = new AuditDao();
            auditDao.setAuditEncrypted(siteCode, userIen, userNPI, userName,
                    patId, queryName, requestingApp, sDate, eDate, info, user.getCardId(), user.getEmail());
        } catch (Exception e) {
            throw new JMeadowsException(e);
        }
    }

    protected Site getSite(String siteCode) {
        return SITES.getSite(siteCode);
    }
}
