/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartMedication;
import gov.va.med.vds.webservice.FreeTextReport;
import gov.va.med.vds.webservice.Medication;
import gov.va.med.vds.webservice.Site;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Patient medications data access object
 */
public class MedicationsDao extends DaoBase {

    /**
     * Retrieves a list of patient medications from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * status (possible options: O - outpatient, I - inpatient, defaults to "O"
     * active (defaults to 1)
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return List of patient medications.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Medication> getPatientMedications(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatMedications", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        String medType = queryBean.getStatus();

        if (medType == null) {
            medType = "O";
        }

        if (queryBean.getActive() == null) {
            queryBean.setActive(1);
        }

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                if ("I".equals(medType)) {
                    PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                    List<String> includedSources = new ArrayList<String>();
                    includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                    gov.va.med.bhie.QueryBean bhieQueryBean =
                            BHIERelayBeanFactory.create(queryBean,  includedSources, null);

                    configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientMedications")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("getInpatientMedicationsFromBHIE")
                                .location(bhieLoc)
                                .build()
                    );
                } else {
                    PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                    List<String> includedSources = new ArrayList<String>();
                    includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
                    includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                    gov.va.med.bhie.QueryBean bhieQueryBean =
                            BHIERelayBeanFactory.create(queryBean, includedSources, null);

                    configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_bhieRelayServiceHandler)
                                .dataServiceMethod("getPatientMedications")
                                .dataServiceMethodArgs(bhieQueryBean)
                                .smartBeanMappingMethod("getOutpatientMedicationsFromBHIE")
                                .location(bhieLoc)
                                .build()
                    );
                }
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientMedications")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        SmartMedication smartMedication = new SmartMedication();

        Object[] medObjs = MultiLocationQuery.execute(
                configs, smartMedication, queryBean);

        List<Medication> medications = null;

        if (medObjs != null) {
            medications = Arrays.asList((Medication []) medObjs);

            //map medication concept codes
            ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, medications, CodingSystem.RXNORM);
        }

        return medications;
    }

    /**
     * Retrieves medication record report from the provided VistA location.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return Inpatient medication record report.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public FreeTextReport getMedicationDetail(JMeadowsQuery queryBean) throws JMeadowsException {

        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.PATIENT,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("MedicaitonDetail", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        FreeTextReport medDetail = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            // get vista data
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite,
                            queryBean.getPatient().getVistaSites())) {

                medDetail = sm_vistaHandler.getMedicationDetail(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return medDetail;
    }
}
