/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 * Honorable Senator Daniel K. Inouye
 * VA Pacific Islands Health Care System
 * Tripler Army Medical Center
 */


package gov.va.med.jmeadows.dao.patient;

import gov.va.med.bhie.DemographicV4;
import gov.va.med.bhie.RecordIdentifierV4;
import gov.va.med.bhie.V4DataRecord;
import gov.va.med.bhie.V4ResultCollection;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.beans.PDWSQueryBean;
import gov.va.med.jmeadows.dao.beans.ResponsePatientQuery;
import gov.va.med.jmeadows.dao.beans.ResponsePatientSelect;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.UserLocation;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.heartbeat.ServiceErrors;
import gov.va.med.jmeadows.smartbeans.patient.SmartDemographics;
import gov.va.med.jmeadows.smartbeans.patient.SmartDemographicsDetail;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.PatientDemographics;
import gov.va.med.vds.webservice.PatientDemographicsDetail;
import gov.va.med.vds.webservice.QueryBean;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vds.webservice.User;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Patient data access object.
 */
public class PatientDao extends DaoBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(PatientDao.class);


    /**
     * Performs a patient lookup against the provided site location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * recordSiteCode - site to perform search against
     * itemId - patient name
     *
     * @return A list of patients that match search criteria.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Patient> lookupPatient(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean,
                QueryBeanParams.USER,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        String requestingApp = queryBean.getRequestingApp();

//        Patient bean = null;
//        org.hui.chcs.QueryBean qb_chcs;

        UserLocation loc = LocationHelper.getUserLocation(queryBean.getUser());

        AuditDao auditDao = new AuditDao();

        User user = queryBean.getUser();

        String userIen = user.getUserIen();

        //append providerId to userIen if one exists
        if (StringUtils.isNotBlank(user.getProviderIen())) {
            userIen += "^" + user.getProviderIen();
        }

        auditDao.setAudit((String) queryBean.getUser().getHostSite().getSiteCode(),
                userIen,
                "",
                //queryBean.getUser().getUserNPI(),
                queryBean.getUser().getName(),
                queryBean.getItemId(),
                "PatientLookup", requestingApp, "", "");

        List<Patient> pats = null;

        if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {

            pats = sm_vistaHandler.lookupPatient(VistaBeanFactory.create(queryBean));

        }

        if (pats != null && queryBean.getMax() != null
                && queryBean.getMax() < pats.size()) {
            pats = pats.subList(0, queryBean.getMax());
        }

        return pats;
    }

    /**
     * Performs a patient lookup against the DMDC Patient Discovery Web Service (PDWS).
     *
     * @param queryBean PDWS Query parameters.
     * @return A list of patients that match search criteria.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public ResponsePatientQuery lookupPatientPDWS(PDWSQueryBean queryBean) throws JMeadowsException {
/*      TODO: PDWSQueryBean parameter check
        validateQueryBean(queryBean,
                QueryBeanParams.USER, QueryBeanParams.DATE_RANGE,
                QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);
*/
        AuditDao auditDao = new AuditDao();
        User user = queryBean.getUser();
        String siteId = "";
        String userIen = "";
        String userNPI = "";
        String userName = "";
        String info = "";
        String requestingApp = "";

        siteId = user.getHostSite().getSiteCode();
        userIen = user.getUserIen();
        userNPI = user.getNPI();
        userName = user.getName();
        info = queryBean.getPatientId()
                + "^" + queryBean.getLastName()
                + "^" + queryBean.getFirstName()
                + "^" + queryBean.getDob()
                + "^" + queryBean.getGender();
        requestingApp = queryBean.getRequestingApp();

        auditDao.setAudit(siteId, userIen, userNPI, userName, queryBean.getPatientId(), "lookupPatientPDWS", requestingApp, "", "", info, user.getCardId(), user.getEmail());

        ResponsePatientQuery responsePatientQuery = null;
        try {
            responsePatientQuery = sm_pdwsHandler.lookupPatient(queryBean);

            if (!responsePatientQuery.isSuccess()) {
                auditDao.setAudit(siteId, userIen, userNPI, userName, queryBean.getPatientId(), "Error: PDWS Service", requestingApp, "", "", info, user.getCardId(), user.getEmail());
                LOGGER.error(responsePatientQuery.getErrorMsg());
            }
        } catch (Exception e) {
            // Audit error
            auditDao.setAudit(siteId, userIen, userNPI, userName, queryBean.getPatientId(), "Error: PDWS Service", requestingApp, "", "", info, user.getCardId(), user.getEmail());
            throw new JMeadowsException(e);
        }

        return responsePatientQuery;
    }

    /**
     * Performs patient selection against the MVI service using either EDIPI, ICN or VA IEN.
     *
     * MVI service-based patient selection details:
     *
     *      MVI service based patient selection will query the MVI service to
     *      retrieve available corresponding VA identifiers.
     *
     *      The following Patient object fields if the MVI service-based selection method is utilized:
     *          edipi
     *          icn
     *          patientIens
     *
     *       The edipi, icn or va ien will be set to the itemId parameter.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * itemId - The edipi, icn or va ien identifier of the patient to select
     * recordSiteCode - the site code of the identifier if it's a va ien
     *
     * @return The selected patient response.
     * @throws JMeadowsException if there is an error.
     */
    public ResponsePatientSelect selectPatientMVI(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.ITEM_ID);

        auditQuery("SelectPatientMVI", queryBean);

        LOGGER.debug("selectPatientMVI::Performing MVI based patient select");

        ResponsePatientSelect result = sm_mviHandler.getPatientIds(queryBean, SITES.getSiteMap());

        // Check whether patient is a sensitive patient or not
        if ((result != null) && (result.getPatient() != null)) {
            Patient patient = result.getPatient();

            //if DoD adapter is up, get ahlta Unit Number
            if (!ServiceErrors.getInstance().getServiceErrors().contains("BHIERelayService") && StringUtils.isNotBlank(patient.getEDIPI())) {
                try {
                    gov.va.med.bhie.QueryBean bhieQueryBean = new gov.va.med.bhie.QueryBean();

                    bhieQueryBean.setUserId(queryBean.getUser().getUserIen());
                    StringBuilder patientId = new StringBuilder();
                    patientId.append("EDIPI:");
                    patientId.append(patient.getEDIPI());
                    bhieQueryBean.setPatientId(patientId.toString());

                    V4ResultCollection patientDemographics = sm_bhieRelayServiceHandler.getPatientDemographics(bhieQueryBean);
                    V4DataRecord dataRecord = patientDemographics.getDataRecordList();
                    StringBuilder fmpssn = new StringBuilder();
                    if (dataRecord != null && dataRecord.getDemographics() != null) {
                        for (DemographicV4 demo : dataRecord.getDemographics()) {
                            if (demo != null) {
                                if (demo.getPatientId() != null) {
                                    Iterator<RecordIdentifierV4> iter = demo.getPatientId().iterator();
                                    String ahltaUnitNum = "";
                                    while (iter.hasNext()) {
                                        RecordIdentifierV4 recid = iter.next();
                                        if ("2.16.840.1.113883.3.42.126.100001.2".equals(recid.getAssigningAuthority())) {
                                            ahltaUnitNum = recid.getId();
                                            break;
                                        }
                                    }
//                                  String ahltaUnitNum = demo.getPatientDemographics().getUnitNumber();
                                    if (ahltaUnitNum != null && !ahltaUnitNum.isEmpty()) {
                                        patient.setAhltaUnitNumber(ahltaUnitNum);
                                    }
                                }

                                // Save fmp/ssn to add to PDWS demographics (person retrieve) call
                                fmpssn.setLength(0);
                                if (StringUtils.isNotBlank(demo.getFmp())) {
                                    fmpssn.append(demo.getFmp());
                                }
                                fmpssn.append("/");
                                if ((demo.getSponsor() != null) && StringUtils.isNotBlank(demo.getSponsor().getSsn())) {
                                    fmpssn.append(demo.getSponsor().getSsn());
                                }
                                patient.setFMP(fmpssn.toString());
                            }
                        }
                    }

                } catch (JMeadowsException e) {
                    LOGGER.info("Could not reach the BHIE service: ", e.getLocalizedMessage());
                }
            }

            result.setPatient(patient);

            // Patient has VA IENs
            if (patient.getPatientIens() != null && !patient.getPatientIens().isEmpty()) {
                QueryBean qbean = new QueryBean();

                qbean.setUser(queryBean.getUser());
                qbean.setPatient(patient);

                boolean isSensitive;
                try {
                    isSensitive = sm_vistaHandler.isSensitivePatient(qbean);
                } catch (JMeadowsException e) {
                    LOGGER.error(e.getMessage(), e);
                    isSensitive = false;
                }
                patient.setSensitive(isSensitive);
                if (!isSensitive) {
                    try {
                        sm_vistaHandler.selectPatient(qbean);
                    } catch (JMeadowsException e) {
                        LOGGER.error(e.getMessage(), e);
                    }
                }
            }
        }

        // Add user's selected patient to recently viewed patients table
        if ((queryBean.getPatient() != null) && StringUtils.isNotBlank(queryBean.getPatient().getEDIPI())) {
            UserDao userDao = new UserDao();
            userDao.addRecentlyViewedPatient(queryBean.getUser().getUserId(), queryBean.getPatient());
        }

        // Add JLV usage for outside providers
        if ((queryBean.getUser() != null) && "OP".equalsIgnoreCase(queryBean.getUser().getAgency())) {
            UserDao userDao = new UserDao();
            userDao.addAssignedPatientUsage(queryBean, queryBean.getUser().getUserIen(), queryBean.getUser().getProviderFlag());
        }

        return result;
//        return handlePixServiceBasedPatientSelect(queryBean);
    }

    /**
     * Performs patient selection to VistA for a sensitive patient after user has accepted sensitive patient warning.
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void selectPatientForVASensitive(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("selectPatientForVASensitive", queryBean);

        QueryBean qbean = new QueryBean();

        qbean.setUser(queryBean.getUser());
        qbean.setPatient(queryBean.getPatient());

        sm_vistaHandler.selectPatient(qbean);
    }

    /**
     * Performs audit on VA user who accepts to access a DoD-only patient record
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void auditVARestrictedAccess(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQueryEncrypted("auditVARestrictedAccess", queryBean);
    }

    /**
     * Performs audit on VA user who accepts to access a sensitive patient record
     *
     * @param queryBean Query parameters.
     *
     * {@link JMeadowsQuery} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * @throws JMeadowsException if there is an error.
     */
    public void auditSensitiveRecordAccess(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        StringBuilder auditTextSB = new StringBuilder(queryBean.getItemId());
        queryBean.setItemId(auditTextSB.toString());

        auditQueryEncrypted("sensitiveRecordAccess", queryBean);
    }


    /**
     * Retrieves a list of patient demographics from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     *
     * @return A list of patient demographics.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDemographics> getPatientDemographics(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDemographics", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {
            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                // Leaving below code in case BDA demographics needs to be called -Jimi
                List<String> includedSources = new ArrayList<String>();
//                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

                if (queryBean.isIncludeVLER()) {
                    configs.add(
                            new PatientMultiLocQueryCfgBuilder()
                                    .dataServiceHandler(sm_bhieRelayServiceHandler)
                                    .dataServiceMethod("getVLERPatientDemographics")
                                    .dataServiceMethodArgs(bhieQueryBean)
                                    .smartBeanMappingMethod("mapPatientDemographicsFromBHIE")
                                    .location(bhieLoc)
                                    .build()
                    );
                }

                PDWSQueryBean pdwsQueryBean = new PDWSQueryBean();
                pdwsQueryBean.setUser(queryBean.getUser());
                pdwsQueryBean.setPatientIdType("EDIPI");
                pdwsQueryBean.setPatientId(queryBean.getPatient().getEDIPI());
                pdwsQueryBean.setInquiryType("PR");
                pdwsQueryBean.setRequestingApp(queryBean.getRequestingApp());

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_pdwsHandler)
                                .dataServiceMethod("lookupPatient")
                                .dataServiceMethodArgs(pdwsQueryBean)
                                .smartBeanMappingMethod("mapPatientDemographicsFromPDWS")
                                .location(bhieLoc)
                                .build()
                );

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientDemographics")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        Object[] demographicObjs = MultiLocationQuery.execute(
                configs, new SmartDemographics(), queryBean);

        List<PatientDemographics> demographics = null;

        if (demographicObjs != null) {
            demographics = new ArrayList(Arrays.asList((PatientDemographics[]) demographicObjs));

            ConceptCodeMapper.getInstance().mapDemographics(hddDao, demographics);

            Patient patient = queryBean.getPatient();
            String edipi = "";
            String fmp = "";
            if (patient != null) {
                if (StringUtils.isNotBlank(patient.getEDIPI())) {
                    edipi = patient.getEDIPI();
                }
                // PDWS does not return FMP and sponsor SSN, therefore FMP and sponsor SSN are retrieved from the
                // BDA demographics call during MVI patient select when querying for the patient's AHLTA unit number
                if (StringUtils.isNotBlank(patient.getFMP())) {
                    fmp = patient.getFMP(); // Format: FMP/SponsorSSN (FMP 'forward slash' Sponsor SSN)
                }
            }

            boolean dodDemoExists = false;
            for (PatientDemographics patDemo : demographics) {
                if ("DOD".equalsIgnoreCase(patDemo.getSite().getAgency())) {
                    dodDemoExists = true;
                    patDemo.setFMP(fmp);
                }
                // Set patient EDIPI
                patDemo.setEDIPI(edipi);
            }

            if (StringUtils.isNotBlank(edipi) && !dodDemoExists) {

                Site dodSite = new Site();
                dodSite.setAgency(AGENCY_DOD);
                dodSite.setSiteCode(AGENCY_DOD);
                dodSite.setName(AGENCY_DOD);
                dodSite.setMoniker(AGENCY_DOD);
                dodSite.setAgency(AGENCY_DOD);

                PatientDemographics dodDemo = new PatientDemographics();
                dodDemo.setName(ERROR_MSG_CONNECTION_UNAVAILABLE);
                dodDemo.setAddress1("");
                dodDemo.setAddress2("");
                dodDemo.setAge("");
                dodDemo.setCity("");
                dodDemo.setClinic("");
                dodDemo.setDob("");
                dodDemo.setGender("");
                dodDemo.setPrimaryProvider("");
                dodDemo.setPhone1("");
                dodDemo.setPhone2("");
                dodDemo.setSSN("");
                dodDemo.setState("");
                dodDemo.setZipCode("");
                dodDemo.setCodeGreen("0");
                dodDemo.setSite(dodSite);

                demographics.add(dodDemo);
            }
        }
        return demographics;
    }

    /**
     * Retrieves a list of patient demographics details from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     *
     * @return A list of patient demographics details.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<PatientDemographicsDetail> getPatientDemographicsDetail(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatDemographicsDetail", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, false);

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                //SB: Exclude data range to retrieve entire data set
                bhieQueryBean.setStartDate(null);
                bhieQueryBean.setEndDate(null);

//                configs.add(
//                        new PatientMultiLocQueryCfgBuilder()
//                                .dataServiceHandler(sm_bhieRelayServiceHandler)
//                                .dataServiceMethod("getPatientDemographics")
//                                .dataServiceMethodArgs(bhieQueryBean)
//                                .smartBeanMappingMethod("mapPatientDemographicsDetailsFromBHIE")
//                                .location(bhieLoc)
//                                .build()
//
//                );
                PDWSQueryBean pdwsQueryBean = new PDWSQueryBean();
                pdwsQueryBean.setUser(queryBean.getUser());
                pdwsQueryBean.setPatientIdType("EDIPI");
                pdwsQueryBean.setPatientId(queryBean.getPatient().getEDIPI());
                pdwsQueryBean.setInquiryType("PR");
                pdwsQueryBean.setRequestingApp(queryBean.getRequestingApp());

                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_pdwsHandler)
                                .dataServiceMethod("lookupPatient")
                                .dataServiceMethodArgs(pdwsQueryBean)
                                .smartBeanMappingMethod("mapPatientDemographicsDetailsFromPDWS")
                                .location(bhieLoc)
                                .build()
                );
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                        new PatientMultiLocQueryCfgBuilder()
                                .dataServiceHandler(sm_vistaHandler)
                                .dataServiceMethod("getPatientDemographicsDetail")
                                .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                                .location(loc)
                                .build()
                );
            }
        }

        Object[] demographicDetailObjs = MultiLocationQuery.execute(
                configs, new SmartDemographicsDetail(), queryBean);

        List<PatientDemographicsDetail> demographicsDetail = null;

        if (demographicDetailObjs != null) {
            demographicsDetail = Arrays.asList((PatientDemographicsDetail[]) demographicDetailObjs);
        }

        return demographicsDetail;
    }
}
