/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.patient;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.hdd.CodingSystem;
import gov.va.med.jmeadows.dao.hdd.ConceptCodeMapper;
import gov.va.med.jmeadows.dao.util.BHIERelayBeanFactory;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.PatientLocation;
import gov.va.med.jmeadows.dao.util.PatientMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.patient.SmartRadiologyExam;
import gov.va.med.vds.webservice.RadiologyReport;
import gov.va.med.vds.webservice.Site;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Radiology data access object.
 */
public class RadiologyDao extends DaoBase {

    /**
     * Retrieves a list of patient radiology reports from available patient locations.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     *
     * Optional:
     * startDate (defaults to T-3 months)
     * endDate   (defaults to T).
     *
     * @return A list of patient radiology reports.
     * @throws JMeadowsException if a serious error occurs.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<RadiologyReport> getPatientRads(JMeadowsQuery queryBean) throws JMeadowsException
    {
        List<RadiologyReport> output = new ArrayList<RadiologyReport>();
        validateQueryBean(queryBean, QueryBeanParams.USER, QueryBeanParams.PATIENT);

        auditQuery("PatRads", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean, true);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null) {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -3, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        for (PatientLocation loc : locs) {

            if (SYSTEM_TYPE_DOD_ADAPTER.equalsIgnoreCase(loc.getSystemType())) {
                PatientLocation bhieLoc = LocationHelper.toBHIEPatientLocation(loc);

                List<String> includedSources = new ArrayList<String>();
                includedSources.add(APP_CONFIG.getBHIECustodianOidAHLTA());
                includedSources.add(APP_CONFIG.getBHIECustodianOidSHARE());

                gov.va.med.bhie.QueryBean bhieQueryBean =
                        BHIERelayBeanFactory.create(queryBean, includedSources, null);

                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_bhieRelayServiceHandler)
                            .dataServiceMethod("getPatientRadiologyReports")
                            .dataServiceMethodArgs(bhieQueryBean)
                            .smartBeanMappingMethod("mapRadiologyExamFromBHIE")
                            .location(bhieLoc)
                            .build()
                );

            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())) {
                configs.add(
                    new PatientMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod("getPatientRads")
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(loc)
                            .build()
                );
            }
        }

        Object[] radObjs = MultiLocationQuery.execute(
                configs, new SmartRadiologyExam(), queryBean);

        List<RadiologyReport> radExams = null;

        if (radObjs != null) {
            List<RadiologyReport> vaList = new ArrayList<RadiologyReport>();
            List<RadiologyReport> dodList = new ArrayList<RadiologyReport>();

            radExams = Arrays.asList((RadiologyReport []) radObjs);
            for (RadiologyReport rad : radExams) {
                if (AGENCY_DOD.equalsIgnoreCase(rad.getSite().getAgency())) {
                    dodList.add(rad);
                } else {
                    vaList.add(rad);
                }
            }

            if (!dodList.isEmpty()) {
                ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, dodList, CodingSystem.RADIOLOGY);
                output.addAll(dodList);
            }

            if (!vaList.isEmpty()) {
                ConceptCodeMapper.getInstance().mapConceptCodes(hddDao, vaList, CodingSystem.RADIOLOGY);
                output.addAll(vaList);
            }

        }

        return output;
    }

    /**
     * Retrieves a patient radiology report from the provided record location.
     *
     * @param queryBean Query parameters.
     *
     * {@link QueryBean} required parameters:
     *
     * {@link gov.va.med.vds.webservice.User}
     * {@link gov.va.med.vds.webservice.Patient}
     * recordSiteCode
     * itemId
     *
     * @return RadiologyReport bean.
     * @throws JMeadowsException if there is an error.
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public RadiologyReport getRadiologyReport(JMeadowsQuery queryBean) throws JMeadowsException {
        validateQueryBean(queryBean,
            QueryBeanParams.USER, QueryBeanParams.PATIENT,
            QueryBeanParams.RECORD_SITE_CODE, QueryBeanParams.ITEM_ID);

        auditQuery("RadiologyReport", queryBean);

        List<PatientLocation> locs = LocationHelper.getPatientLocations(queryBean);

        RadiologyReport radReport = null;

        Site recordSite = getSite(queryBean.getRecordSiteCode());

        for (PatientLocation loc : locs) {

            //get vista data
            if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(loc.getSystemType())
                    && LocationHelper.vistaLocationsContainSiteCode(recordSite,
                            queryBean.getPatient().getVistaSites())) {

                radReport = sm_vistaHandler.getRadiologyReport(VistaBeanFactory.create(queryBean));
                break;
            }
        }

        return radReport;

    }
}
