/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.provider;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.patient.DaoBase;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.ProviderMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.UserLocation;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.provider.SmartConsult;
import gov.va.med.vds.webservice.Consult;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 *
 * @author DNS   brunos
 */
public class ProviderConsultsDao extends DaoBase {

    /**
     * Gets a list of all consults which have been requested by the user
     * passed in the queryBean. All statuses will be returned.
     * @param queryBean requires a user, startDate and endDate. Start and
     * end dates tell the underlying processing routines when to start and stop
     * searching. The wider the date range the longer it takes to process.
     * @return a list of all consults requested
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Consult> getProviderConsultsRequested(JMeadowsQuery queryBean) throws JMeadowsException {
        List<Consult> rtc = null;
        validateQueryBean(queryBean, QueryBeanParams.USER);

        auditQuery("ProviderConsultsRequested", queryBean);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null)
        {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -2, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {
            String systemType = location.getSystemType();
            String method = "getProviderConsultsRequested";
            if (SYSTEM_TYPE_CHCS.equalsIgnoreCase(systemType)) {
                rtc = Arrays.asList((Consult[])
                        new SmartConsult().generateConnectionUnavailableBean(location, PROVIDER_QUERY_NOT_SUPPORTED_DOD));
                break;
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod(method)
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(location)
                            .build()
                );
            }
        }
        if (rtc == null) {
            Object[] consultObjs = MultiLocationQuery.execute(
                    configs, new SmartConsult(), queryBean);
    
            List<Consult> consultsRequested = null;
    
            if (consultObjs != null) {
                consultsRequested = Arrays.asList((Consult []) consultObjs);
            }
    
            rtc = consultsRequested;
        }
        return rtc;
    }

    /**
     * Gets a list of all consults which a user must complete. Ie,
     * user X orders an echocardiogram; user Y receives the consult
     * to be completed. This method enables a user (in this example Y) to
     * see what consults need to be completed on a longitudinal patient domain
     * rather than a patient by patient basis.
     * @param queryBean requires user, startDate and endDate. Start and end
     * dates tell the underlying processing routines when to start and
     * stop searching. The wider the date range the longer it takes to process.
     * @return a list of consults that a user has received for a given
     * user and date range.
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Consult> getProviderConsultsReceived(JMeadowsQuery queryBean) throws JMeadowsException
    {
        List<Consult> rtc = null;
         
        validateQueryBean(queryBean, QueryBeanParams.USER);

        auditQuery("ProviderConsultsReceived", queryBean);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null)
        {
            queryBean.setStartDate(Utils.getCal(Utils.getRelativeDate(0, -2, 0))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        //build QueryLocation configs
        List<MultiLocationQueryConfig> configs = new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {
            String systemType = location.getSystemType();
            String method = "getProviderConsultsReceived";
            if (SYSTEM_TYPE_CHCS.equalsIgnoreCase(systemType)) {
                rtc = Arrays.asList((Consult[])
                        new SmartConsult().generateConnectionUnavailableBean(location, PROVIDER_QUERY_NOT_SUPPORTED_DOD));
                break;
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod(method)
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(location)
                            .build()
                );
            }
        }

        if (rtc == null) {
            Object[] consultObjs = MultiLocationQuery.execute(
                    configs, new SmartConsult(), queryBean);
    
            List<Consult> consultsReceived = null;
    
            if (consultObjs != null) {
                consultsReceived = Arrays.asList((Consult []) consultObjs);
            }
    
            rtc = consultsReceived;
        }
        return rtc;
    }
    
}
