/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.provider;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.patient.DaoBase;
import gov.va.med.jmeadows.dao.util.LocationHelper;
import gov.va.med.jmeadows.dao.util.MultiLocationQuery;
import gov.va.med.jmeadows.dao.util.MultiLocationQueryConfig;
import gov.va.med.jmeadows.dao.util.ProviderMultiLocQueryCfgBuilder;
import gov.va.med.jmeadows.dao.util.UserLocation;
import gov.va.med.jmeadows.dao.util.VistaBeanFactory;
import gov.va.med.jmeadows.smartbeans.provider.SmartOrder;
import gov.va.med.vds.webservice.Order;
import gov.va.med.DNS   .common.Utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 *
 * @author DNS   brunos
 */
public class ProviderOrdersDao extends DaoBase {

    /**
     * Gets a list of orders pending which have been placed by the user
     * passed in the queryBean.
     * @param queryBean requires user, startDate and endDate. Start and end
     * dates tell the underlying processing routines when to start and stop
     * searching. The wider the date range the longer it takes to process.
     * @return a list of orders pending for a given user and date range.
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Order> getProviderOrdersPending(JMeadowsQuery queryBean) throws JMeadowsException {
        List<Order> rtc = null;
        validateQueryBean(queryBean, QueryBeanParams.USER);

        auditQuery("ProviderOrdersPending", queryBean);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null)
        {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -1, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        List<MultiLocationQueryConfig> configs =
                new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {
            String systemType = location.getSystemType();
            String method = "getProviderOrdersPending";
            if (SYSTEM_TYPE_CHCS.equalsIgnoreCase(systemType)) {
                rtc = Arrays.asList((Order[])
                        new SmartOrder().generateConnectionUnavailableBean(location, PROVIDER_QUERY_NOT_SUPPORTED_DOD));
                break;
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod(method)
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(location)
                            .build()
                );
            }
        }

        if (rtc == null) {
            Object[] orderObj = MultiLocationQuery.execute(configs,
                    new SmartOrder(), queryBean);
    
            List<Order> orders = null;
    
            if (orderObj != null) {
                orders = Arrays.asList((Order []) orderObj);
            }
    
            rtc = orders; 
        }
        return rtc;
    }

     /**
     * Gets a list of orders which have results which have been placed by the
     * user passed in the queryBean.
     * @param queryBean requires user, startDate and endDate. Start and end
     * dates tell the underlying processing routines when to start and stop
     * searching. The wider the date range the longer it takes to process.
     * @return a list of orders which have results for a given user and
     * date range. For each order, the actual result is also included.
     * @throws gov.va.med.jmeadows.common.JMeadowsException when an error occurs in the retrieval process
     * @throws IllegalArgumentException if required parameters are missing or invalid
     */
    public List<Order> getProviderOrdersResulted(JMeadowsQuery queryBean) throws JMeadowsException {
        List<Order> rtc = null;
        validateQueryBean(queryBean, QueryBeanParams.USER);

        auditQuery("ProviderOrdersResulted", queryBean);

        if (queryBean.getStartDate() == null || queryBean.getEndDate() == null)
        {
            queryBean.setStartDate(Utils.getCalendar(Utils.pushDateToStartOfDay(Utils.getDate(Utils.getRelativeDate(0, -2, 0))))); //(years,months,days)
            queryBean.setEndDate(Utils.getCalendar(Utils.pushDateToEndOfDay(new java.util.Date())));
        }

        List<MultiLocationQueryConfig> configs =
                new ArrayList<MultiLocationQueryConfig>();

        List<UserLocation> locations;

        locations = LocationHelper.getUserLocations(queryBean.getUser());

        for (UserLocation location : locations) {
            String systemType = location.getSystemType();
            String method = "getProviderOrdersResulted";
            if (SYSTEM_TYPE_CHCS.equalsIgnoreCase(systemType)) {
                rtc = Arrays.asList((Order[])
                        new SmartOrder().generateConnectionUnavailableBean(location, PROVIDER_QUERY_NOT_SUPPORTED_DOD));
                break;
            } else if (SYSTEM_TYPE_VISTA.equalsIgnoreCase(systemType)) {
                configs.add(
                    new ProviderMultiLocQueryCfgBuilder()
                            .dataServiceHandler(sm_vistaHandler)
                            .dataServiceMethod(method)
                            .dataServiceMethodArgs(VistaBeanFactory.create(queryBean))
                            .location(location)
                            .build()
                );
            }
        }

        if (rtc == null) {
            Object[] orderObjs = MultiLocationQuery.execute(configs,
                    new SmartOrder(), queryBean);
    
            List<Order> orders = null;
    
            if (orderObjs != null) {
                orders = Arrays.asList((Order []) orderObjs);
            }
    
            rtc = orders;
        }
        return rtc;
    }
    
}
