/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.util;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * DataQuery uses reflection to execute a method on a given class object.
 * @author DNS   brunos
 */
public class DataQuery implements Serializable {
    /**
     * Class UID.
     */
    private static final long serialVersionUID = -4672649255420220521L;

    private Object classObj;
    private String methodName;
    private Object[] args;

    /**
     * DataQuery constructor
     * @param classObj Object to invoke
     * @param methodName Method to invoke
     * @param args Arguments to pass method.
     */
    public DataQuery(Object classObj, String methodName, Object[] args) {
        this.classObj = classObj;
        this.methodName = methodName;
        if (args == null) {
            this.args = null;
        } else {
            this.args = args.clone();
        }
    }

    public String toString() {
        StringBuilder sbuild = new StringBuilder();
        sbuild.append("{");
        sbuild.append("clasObj: ");
        if (classObj != null) {
            sbuild.append(classObj.getClass().getName());
        }
        sbuild.append(", methodName: ");
        sbuild.append(methodName);
        sbuild.append(", args length: ");
        if (args != null) {
            sbuild.append(args.length);
        }
        sbuild.append("}");

        return sbuild.toString();
    }

    /**
     * Uses reflection invoke a method on the classObj member variable
     * with given arguments.
     * @return Result of invoked method.
     */
    public Object execute() {
        Object rtc;
        if (classObj == null) {
            rtc = null;
        } else {
            Object result;
    
            try {
                Class cls = classObj.getClass();
                Class[] paramTypes = new Class[args.length];
                for (int i = 0; i < args.length; i++) {
                    paramTypes[i] = args[i].getClass();
                }
    
                Method method = getMethod(cls, methodName, paramTypes);
                result = method.invoke(classObj, args);
            } catch (NoSuchMethodException e) {
                result = e;
            } catch (InvocationTargetException e) {
                result = e;
            } catch (IllegalAccessException e) {
                result = e;
            }
    
            rtc = result;
        }
        return rtc;
    }

    /**
     * Searches for and returns method within given class that matches
     * name and parameter type list. This method is used in lieu of
     * Class.getMethod() because it matches parameter types based on
     * assignability instead of type name.
     * (i.e. can pass parameters that are a subclass of those
     * defined in method signature).
     *
     * @param theClass Class to get method from.
     * @param methodName Name of method to get.
     * @param paramTypes List of parameter types.
     * @return Method if found.
     * @throws NoSuchMethodException If no method is found.
     */
    private Method getMethod(Class theClass, String methodName, Class[] paramTypes) throws NoSuchMethodException {
        Method[] methods = theClass.getMethods();

        for (Method method : methods) {
            if (method.getName().equals(methodName)) {
                Class[] methodParamTypes = method.getParameterTypes();
                if (methodParamTypes.length != paramTypes.length) {
                    continue;
                }

                boolean isParamMatch = true;
                for (int i = 0; i < paramTypes.length; i++) {
                    Class param = methodParamTypes[i];
                    if (!param.isAssignableFrom(paramTypes[i])) {
                        isParamMatch = false;
                    }
                }

                if (isParamMatch) {
                    return method;
                }
            }
        }

        throw new NoSuchMethodException();
    }
}
