/*
 * Janus 4.0 (c)
 * Copyright (c) 2014 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dao.util;

import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.heartbeat.ServiceErrors;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vds.webservice.User;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class LocationHelper {
    private static final AppConfig APP_CONFIG = AppConfig.getInstance();

    /**
     * True if vista locations list contains site code.
     *
     * @param site       Site to match.
     * @param vistaSites List of vista location site codes (found within patient bean)
     * @return True if vista locations list contains site code.
     */
    public static boolean vistaLocationsContainSiteCode(Site site, List<Site> vistaSites) {
        boolean rtc = false;
        if (vistaSites != null) {
            for (Site vistaSite : vistaSites) {
                if (vistaSite.getSiteCode().equals(site.getSiteCode())) {
                    rtc = true;
                    break;
                }
            }
        }
        return rtc;
    }

    /**
     * Retrieves locations associated with a patient.
     *
     * Includes the DoD Adapter location if connectivity to the adapter service is down.
     *
     * @param qbean QueryBean containing a patient bean.
     * @return A list of patient locations.
     */
    public static List<PatientLocation> getPatientLocations(JMeadowsQuery qbean) {
        return getPatientLocations(qbean, false);
    }

    /**
     * Retrieves locations associated with a patient.
     *
     * @param qbean QueryBean containing a patient bean.
     * @param excludeDoDAdapterIfUnavailable
     *              Exclude the DoD Adapter location if connectivity to the adapter service is down.
     * @return A list of patient locations.
     */
    public static List<PatientLocation> getPatientLocations(JMeadowsQuery qbean, boolean excludeDoDAdapterIfUnavailable) {

        if (qbean.getPatient() == null) {
            throw new IllegalArgumentException("Patient is null");
        }

        boolean isDoDAdaptorConnectivity = true;

        // Non-VA providers are currently limited to see only VA patient data.
        // TODO: Use OP data type flags instead of checking agency.
        if (qbean != null && qbean.getUser() != null && StringUtils.equalsIgnoreCase(qbean.getUser().getAgency(), "OP")) {
            isDoDAdaptorConnectivity = false;
        }

        if (excludeDoDAdapterIfUnavailable && ServiceErrors.getInstance().getServiceErrors().contains("BHIERelayService")) {
            isDoDAdaptorConnectivity = false;
        }

        Patient patient = qbean.getPatient();

        List<PatientLocation> patLocs = new ArrayList<PatientLocation>();

        if (isDoDAdaptorConnectivity && StringUtils.isNotBlank(patient.getEDIPI())) {
            PatientLocation dodLoc = new PatientLocation();
            dodLoc.setSystemType(APP_CONFIG.getSystemTypeDODAdapter());
            dodLoc.setPatient(patient);

            patLocs.add(dodLoc);
        }

        //add Vista system location if patient has vista sites and an ICN or IEN list
        if ((patient.getVistaSites() != null && !patient.getVistaSites().isEmpty())
                && (StringUtils.isNotBlank(patient.getICN())
                || (patient.getPatientIens() != null && !patient.getPatientIens().isEmpty())))
        {
            PatientLocation vaLoc = new PatientLocation();
            vaLoc.setSystemType(APP_CONFIG.getSystemTypeVISTA());
            vaLoc.setPatient(patient);

            patLocs.add(vaLoc);
        }

        if (patLocs.size() < 1) {
            throw new RuntimeException("Patient could not be correlated to a known facility location.");
        }

        return patLocs;
    }

    public static List<UserLocation> getUserLocations(User user) {
        return Arrays.asList(getUserLocation(user));
    }

    public static UserLocation getUserLocation(User user) {
        UserLocation userLocation = new UserLocation();
        if (APP_CONFIG.getAgencyDOD().equalsIgnoreCase(user.getHostSite().getAgency())) {
            userLocation.setSystemType(APP_CONFIG.getSystemTypeCHCS());
        } else if (APP_CONFIG.getAgencyVA().equalsIgnoreCase(user.getHostSite().getAgency())) {
            userLocation.setSystemType(APP_CONFIG.getSystemTypeVISTA());
        } else {
            throw new IllegalArgumentException("User does not belong to a supported agency!");
        }

        userLocation.setUser(user);

        return userLocation;
    }

    /**
     * Helper function that clones PatientLocation bean parameter and changes the site fields to BHIE site fields.
     *
     * @param loc Location to clone.
     * @return A BHIE PatientLocation
     * @throws gov.va.med.jmeadows.common.JMeadowsException if a serious error occurs.
     */
    public static PatientLocation toBHIEPatientLocation(PatientLocation loc) throws JMeadowsException {
        try {
            PatientLocation patientLocation = new PatientLocation();
            patientLocation.setSystemType(APP_CONFIG.getSystemTypeDODAdapter());
            patientLocation.setPatient(loc.getPatient());

            return patientLocation;
        } catch (Exception e) {
            throw new JMeadowsException(e);
        }
    }
}
