/**
 * Janus 4.0 (c)
 * Copyright (c) 2011 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *     Honorable Senator Daniel K. Inouye
 *     VA Pacific Islands Health Care System
 *     Tripler Army Medical Center
 */
 
 
package gov.va.med.jmeadows.dao.util;

import gov.va.med.jmeadows.common.AppConfig;
import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.smartbeans.patient.SmartBean;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author DNS   brunos
 */
public class MultiLocationQuery {
    private final static int MAX_THREAD_POOL_SIZE = 7;

    private final static AppConfig APP_CONFIG = AppConfig.getInstance();
//    private static final String AGENCY_VA = APP_CONFIG.getAgencyVA();
//    private static final String AGENCY_DOD = APP_CONFIG.getAgencyDOD();
    private static final long THREAD_TIMEOUT_MS = APP_CONFIG.getThreadTimeoutMS();
//    private static final String SITE_UNKNOWN = "Unknown";

    private final static Logger LOGGER = LoggerFactory.getLogger(MultiLocationQuery.class);


    /**
     * Executes multi-location data queries.
     * @param queryLocCfgs List of location query configurations.
     * @param smartBean SmartBean shared by the query configurations.
     * @param queryBean Webservice querybean
     * @return A list of databean results.
     */
    public static Object[] execute(List<MultiLocationQueryConfig> queryLocCfgs,
            SmartBean smartBean, JMeadowsQuery queryBean) throws JMeadowsException {
        Object[] rtc;
        if (queryLocCfgs == null || queryLocCfgs.isEmpty()) {
            LOGGER.debug("List of query configs if null or empty");

            //bail if query locations configs is null or empty
            rtc = null;
        } else {
    
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug("queryLocCfgs ({})", queryLocCfgs.size());
            }
    
            //setup thread and thread execution arrays
            List<MultiLocationQueryThread> dataQueryThreads = new ArrayList<MultiLocationQueryThread>();
    
            //generate query threads for each location
            for (MultiLocationQueryConfig config : queryLocCfgs) {
                dataQueryThreads.add(new MultiLocationQueryThread(config, smartBean));
            }
    
            LinkedHashMap<Location, Object> dataBeansMap = new LinkedHashMap<Location, Object>();
    
            int threadPoolSize = MAX_THREAD_POOL_SIZE;
            if (queryLocCfgs.size() < MAX_THREAD_POOL_SIZE) {
                threadPoolSize = queryLocCfgs.size();
            }
            LOGGER.debug("threadPoolSize: {}, maxThreadPoolSize: {}", threadPoolSize, MAX_THREAD_POOL_SIZE);
    
            ExecutorService executorService = Executors.newFixedThreadPool(threadPoolSize);
    
            try {
                List<Future<Object>> futures = executorService.invokeAll(dataQueryThreads, THREAD_TIMEOUT_MS, TimeUnit.MILLISECONDS);
    
                //futures are returned in the same sequential order as produced by the iterator for the given task list.
                Iterator<MultiLocationQueryThread> queryThreadItr = dataQueryThreads.iterator();
                Iterator<Future<Object>> futuresItr = futures.iterator();
    
                while (queryThreadItr.hasNext() && futuresItr.hasNext()) {
                    MultiLocationQueryThread queryThread = queryThreadItr.next();
                    Future<Object> future = futuresItr.next();
                    handleQuery(queryThread, future, dataBeansMap, smartBean);
                }
    
                //init return object array
                Object[] returnVal = null;
    
                //use smart bean to combine data from each location
                for (Object entry : dataBeansMap.entrySet()) {
                    Object[] dataBeanList = (Object[]) ((Map.Entry) entry).getValue();
                    returnVal = smartBean.combineData(returnVal, dataBeanList, queryBean.getSortBy());
                }
    
                //Limit returnVal by max, if max is set
                if (returnVal == null || queryBean.getMax() == null) {
                    rtc = returnVal;
                } else {
                    if (returnVal.length <= queryBean.getMax()) {
                        rtc = returnVal;
                    } else {
                        Object[] maxRetVal = new Object[queryBean.getMax()];
                        System.arraycopy(returnVal, 0, maxRetVal, 0, queryBean.getMax());
                        rtc = maxRetVal;
                    }
                }
    
            } catch (InterruptedException e) {
                throw new JMeadowsException(e);
            } finally {
                executorService.shutdown();
            }
        }
        return rtc;
    }

    private static void handleQuery(MultiLocationQueryThread queryThread, Future<Object> future, LinkedHashMap<Location, Object> dataBeansMap, SmartBean smartBean) throws InterruptedException {
        Object[] result = null;

        try {
            if (future.isCancelled()) {
                MultiLocationQueryConfig queryConfig = queryThread.getQueryConfig();
                result = smartBean.generateConnectionUnavailableBean(
                        queryConfig.getLocation(),
                        queryConfig.getConnectionErrorMsg());

                //log connection timeout
                StringBuilder errorMsg = new StringBuilder();
                errorMsg.append("MultiLocationQueryThread timeout. Thread wait time exceeded ");
                errorMsg.append(THREAD_TIMEOUT_MS);
                errorMsg.append("ms");
                LOGGER.error(errorMsg.toString(), new JMeadowsException(errorMsg.toString()));
            } else {
                result = (Object[]) future.get();
            }

            MultiLocationQueryConfig config = queryThread.getQueryConfig();

            //add result to databean location map
            if (result != null) {
                Object[] newResult;
                //if previous result exists for location, combine it with returning result set

                Location locationKey;

                locationKey = config.getLocation();

                Object[] prevResult = (Object[]) dataBeansMap.get(locationKey);

                if (prevResult == null) {
                    newResult = result;
                } else {
                    ArrayList<Object> resultList = new ArrayList<Object>();
                    resultList.addAll(Arrays.asList(prevResult));
                    resultList.addAll(Arrays.asList(result));
                    newResult = resultList.toArray(new Object[resultList.size()]);
                }
                dataBeansMap.put(locationKey, newResult);
            }
        } catch (ExecutionException ex) {
            MultiLocationQueryConfig config = queryThread.getQueryConfig();

            result = smartBean.generateConnectionUnavailableBean(
                    config.getLocation(),
                    config.getConnectionErrorMsg());

            Location locationKey;
            locationKey = config.getLocation();

            dataBeansMap.put(locationKey, result);

            LOGGER.error(ex.getMessage(), ex);
        }
    }
}
