/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.dataservicehandler;

import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.jmeadows.dao.beans.ExternalID;
import gov.va.med.jmeadows.dao.beans.JMeadowsQuery;
import gov.va.med.jmeadows.dao.beans.ResponsePatientSelect;
import gov.va.med.jmeadows.dao.hdd.VLERSiteCache;
import gov.va.med.jmeadows.dao.patient.AuditDao;
import gov.va.med.mvi.ActClassControlAct;
import gov.va.med.mvi.CD;
import gov.va.med.mvi.CS;
import gov.va.med.mvi.CommunicationFunctionType;
import gov.va.med.mvi.EntityClassDevice;
import gov.va.med.mvi.II;
import gov.va.med.mvi.MCCIMT000100UV01Device;
import gov.va.med.mvi.MCCIMT000100UV01Receiver;
import gov.va.med.mvi.MCCIMT000100UV01Sender;
import gov.va.med.mvi.MCCIMT000300UV01Acknowledgement;
import gov.va.med.mvi.MCCIMT000300UV01AcknowledgementDetail;
import gov.va.med.mvi.ObjectFactory;
import gov.va.med.mvi.PRPAIN201309UV02;
import gov.va.med.mvi.PRPAIN201309UV02QUQIMT021001UV01ControlActProcess;
import gov.va.med.mvi.PRPAIN201310UV02;
import gov.va.med.mvi.PRPAIN201310UV02MFMIMT700711UV01Subject1;
import gov.va.med.mvi.PRPAMT201304UV02Patient;
import gov.va.med.mvi.PRPAMT201307UV02ParameterList;
import gov.va.med.mvi.PRPAMT201307UV02PatientIdentifier;
import gov.va.med.mvi.PRPAMT201307UV02QueryByParameter;
import gov.va.med.mvi.TS;
import gov.va.med.mvi.VAIdM;
import gov.va.med.mvi.VAIdMPort;
import gov.va.med.mvi.XActMoodIntentEvent;
import gov.va.med.vds.webservice.Code;
import gov.va.med.vds.webservice.KeyValuePair;
import gov.va.med.vds.webservice.Patient;
import gov.va.med.vds.webservice.Site;
import gov.va.med.vds.webservice.User;

import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.StringTokenizer;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;

public class MVIServiceHandler extends DataServiceHandler {
    private static final Logger LOGGER = LoggerFactory.getLogger(MVIServiceHandler.class);
    private static int lastGoodURLindex = 0;
    private VAIdM mviService;
    private JAXBContext jaxbContext = null;
    

    public MVIServiceHandler(String url) {
        super();

        serviceURL = url;
        serviceName = "MVI_Service";
        serviceNamespace = "http://webservice.mviws.med.DNS   ";
        qServiceName = new QName(serviceNamespace, serviceName);

        mviService = new VAIdM();
    }

    private VAIdMPort getMVIPort() {
        VAIdMPort port = mviService.getVAIdMPort();

        // serviceURL can be a set of URLs to try like: URL;URL;URL
        String currentURL = serviceURL;
        if (serviceURL != null && serviceURL.indexOf(';') > -1) {
            String[] urls = serviceURL.split(";");
            if (lastGoodURLindex < urls.length) {
                currentURL = urls[lastGoodURLindex];
            } else {
                currentURL = urls[0];
                lastGoodURLindex = 0;
            }
        }
        BindingProvider prov = (BindingProvider) port;
        prov.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, currentURL);
        prov.getRequestContext().put("com.sun.xml.internal.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.internal.ws.request.timeout", serviceRequestTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.connect.timeout", serviceConnectionTimeoutMS);
        prov.getRequestContext().put("com.sun.xml.ws.request.timeout", serviceRequestTimeoutMS);

        // SOAP message logging - Enable to show logging in output window
        if ("test".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()) || "development".equalsIgnoreCase(APP_CONFIG.getAppEnvironment()))
        {
            prov.getBinding().getHandlerChain().add(new SOAPLoggingHandler());
            List<Handler> handlerList = prov.getBinding().getHandlerChain();
            handlerList.add(new SOAPLoggingHandler());
            prov.getBinding().setHandlerChain(handlerList);
        }

        return port;
    }

    public ResponsePatientSelect getPatientIds(JMeadowsQuery queryBean, Map<String, Site> siteMap) throws JMeadowsException
    {
        String errorMessage = "";
        boolean done = false;

        ResponsePatientSelect responsePatientSelect = new ResponsePatientSelect();
        List<ExternalID> externalIDs = new ArrayList<ExternalID>();

        if (jaxbContext == null) {
            try {
                jaxbContext = JAXBContext.newInstance(PRPAIN201309UV02.class);
            } catch (JAXBException e) {
                jaxbContext = null;
                incLastGoodURLindex();

                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";

                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);

                logAndAuditException(e, errorMessage, queryBean, false);

                done = true;
            }
        }
        if (!done) {
            try {
                // Generate PRPAIN201309UV02 request message
                PRPAIN201309UV02 request = generateRequest(queryBean);
    
                // Marshall PRPAIN201309UV02 request into XML
                StringWriter stringWriter = new StringWriter();
                jaxbContext.createMarshaller().marshal(request, stringWriter);
    
                // Send MVI request message and get a response message
                Element element = (Element) getMVIPort().prpaIN201309UV02(stringWriter.toString());
    
                // Unmarshall element into PRPAIN201310UV02 response
                PRPAIN201310UV02 response = jaxbContext.createUnmarshaller().unmarshal(element, PRPAIN201310UV02.class).getValue();
    
                // Parse response into patient IDs map
                List<String> patientIds = parseResponseToPatientIds(response);
    
                Patient patient = new Patient();
    
                Iterator<String> iterator = patientIds.iterator();
                while (iterator.hasNext()) {
                    // Corresponding IDs format: ID^TYPE^SOURCE^ISSUER^STATUS
                    StringTokenizer tokenizer = new StringTokenizer(iterator.next(), "^");
                    if (tokenizer.countTokens() == 5) // According to format, there should only be 5 tokens
                    {
                        String id = tokenizer.nextToken();
                        String type = tokenizer.nextToken();
                        String source = tokenizer.nextToken();
                        String issuer = tokenizer.nextToken();
                        String status = tokenizer.nextToken();
    
                        // Makes sure each relevant token is not null or blank and status is "A" for active
                        if (StringUtils.isNotBlank(id) && StringUtils.isNotBlank(type)
                                && StringUtils.isNotBlank(source) && StringUtils.isNotBlank(issuer)
                                && StringUtils.isNotBlank(status)) {
                            if ("NI".equalsIgnoreCase(type)) {
                                // National patient identifier
                                if ("200DOD".equalsIgnoreCase(source) && "USDOD".equalsIgnoreCase(issuer) && "A".equalsIgnoreCase(status))
                                {
                                    // DoD EDIPI
                                    if ("EDIPI".equalsIgnoreCase(queryBean.getStatus())) {
                                        id = queryBean.getItemId();
                                    }
                                    patient.setEDIPI(id);
                                } else if ("200M".equalsIgnoreCase(source) && "USVHA".equalsIgnoreCase(issuer) && "P".equalsIgnoreCase(status)) {
                                    // VA ICN
                                    patient.setICN(id);
                                } else {
                                    externalIDs.add(new ExternalID(id, type, source, issuer));
                                }
                            } else if ("PI".equalsIgnoreCase(type)) {
                                // Local facility patient identifier
                                if ("USVHA".equalsIgnoreCase(issuer) && "A".equalsIgnoreCase(status)) {
                                    // VA DFN (IEN)
                                    if (siteMap != null) {
                                        Site site = siteMap.get(source);
                                        if (site == null) {
                                            LOGGER.debug("Site with site code " + source + " not found in sites map");
                                        } else {
                                            patient.getVistaSites().add(site);
    
                                            // Add patient's VA IEN and corresponding site code to key-value pair list
                                            KeyValuePair kvp = new KeyValuePair();
                                            kvp.setKey(source);
                                            kvp.setValue(id);
                                            patient.getPatientIens().add(kvp);
                                        }
                                    }
                                }  else {
                                    externalIDs.add(new ExternalID(id, type, source, issuer));
                                }
                            }
                        }
                    } else {
                        LOGGER.debug("Invalid corresponding ID string format");
                    }
                }
    
                // VA user restrict flag is set. Check to see if VA user is selecting a DoD-only patient
                if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserRestrict()) 
                        && APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getHostSite().getAgency()) && patient.getICN() == null 
                        && patient.getPatientIens().isEmpty()) {
                    responsePatientSelect.setVARestricted(true);
    
                    if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserAllowRestrictAccess())) {
                        responsePatientSelect.setAllowVAAccess(true);
                    }
                }
    
                externalIDs = getVLERSiteNames(externalIDs);
    
                responsePatientSelect.setSuccess(true);
                responsePatientSelect.setPatient(patient);
                responsePatientSelect.setExternalIDs(externalIDs);
            } catch (JAXBException e) {
                incLastGoodURLindex();
                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";
    
                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);
    
                logAndAuditException(e, errorMessage, queryBean, false);
            } catch (JMeadowsException e) {
                // If VA user, they do not need to see the error message. GUI checks for empty string and skips message.
                if (APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getAgency()))
                {
                    errorMessage = "";
                } else {
                    errorMessage = "This patient is currently not known to the VA Master Veterans Index, and therefore no VA"
                            + " data will be displayed.<br/><br/>If this patient is a veteran, please notify the help desk of"
                            + " this issue.";
                }
    
                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);
    
                logAndAuditException(e, errorMessage, queryBean, false);
            } catch (Exception e) {
                incLastGoodURLindex();
                errorMessage = "VA Patient Identity Service is unavailable. VA data may not display.";
    
                responsePatientSelect.setSuccess(false);
                responsePatientSelect.setErrorMsg(errorMessage);
    
                logAndAuditException(e, errorMessage, queryBean, true);
                // Commenting out below line will not pass fortify(?) scan, but leaving it in breaks functionality
                // to continue showing DoD patient data when there is an issue with the MVI service. Revisit. -Jimi
    //            throw new JMeadowsException("MVI Application Error: " + errorMessage);
            }
    
            // MVI was unsuccessful, at least keep original identifier
            if (!responsePatientSelect.isSuccess()) {
    
                Patient patient = new Patient();
    
                String type = queryBean.getStatus();
                String id = queryBean.getItemId();
                String source = queryBean.getRecordSiteCode();
    
                if ("EDIPI".equalsIgnoreCase(type)) {
                    patient.setEDIPI(id);
                } else if ("ICN".equalsIgnoreCase(type)) {
                    patient.setICN(id);
                } else if ("VAIEN".equalsIgnoreCase(type) && siteMap != null) {
                    Site site = siteMap.get(source);
                    if (site != null) {
                        patient.getVistaSites().add(site);
    
                        // Add patient's VA IEN and corresponding site code to key-value pair list
                        KeyValuePair kvp = new KeyValuePair();
                        kvp.setKey(source);
                        kvp.setValue(id);
                        patient.getPatientIens().add(kvp);
                    }
                }
    
                // VA user restrict flag is set. Check to see if VA user is selecting DoD-only patient
                if ("TRUE".equalsIgnoreCase(APP_CONFIG.getVaUserRestrict()) 
                        && APP_CONFIG.getAgencyVA().equalsIgnoreCase(queryBean.getUser().getHostSite().getAgency()) 
                        && patient.getICN() == null && patient.getPatientIens().isEmpty()) {
                    responsePatientSelect.setVARestricted(true);
                    responsePatientSelect.setAllowVAAccess(true);
                }
    
                responsePatientSelect.setErrorMsg(errorMessage);
                responsePatientSelect.setPatient(patient);
            }
        }
        return responsePatientSelect;
    }
    
    private static void incLastGoodURLindex() {
        ++lastGoodURLindex;
    }
    

    private PRPAIN201309UV02 generateRequest(JMeadowsQuery queryBean) {
        ObjectFactory factory;
        PRPAIN201309UV02 request;
        MCCIMT000100UV01Receiver receiver;
        MCCIMT000100UV01Device device;
        MCCIMT000100UV01Sender sender;
        PRPAIN201309UV02QUQIMT021001UV01ControlActProcess controlActProcess;
        PRPAMT201307UV02QueryByParameter queryByParameter;
        PRPAMT201307UV02ParameterList parameterList;
        PRPAMT201307UV02PatientIdentifier patientIdentifier;
        SimpleDateFormat simpleDateFormat;
        II ii;
        TS ts;
        CS cs;
        CD cd;
        Date date;
        String status;
        String oid;
        String id;
        String type;
        String source;
        String issuer;

        factory = new ObjectFactory();
        request = factory.createPRPAIN201309UV02();
        date = new Date();

        //id
        simpleDateFormat = new SimpleDateFormat("yyyyMMdd", Locale.getDefault());
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("MCID_JANUS_DEV_T1_" + simpleDateFormat.format(date));
        request.setId(ii);

        //creationTime
        ts = factory.createTS();
        simpleDateFormat = new SimpleDateFormat("yyyyMMddhhmmss", Locale.getDefault());
        ts.setValue(simpleDateFormat.format(date));
        request.setCreationTime(ts);

        //interactionId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.1.6");
        ii.setExtension("PRPA_IN201309UV02");
        request.setInteractionId(ii);

        //processingCode
        cs = factory.createCS();
        if ("production".equalsIgnoreCase(APP_CONFIG.getAppEnvironment())) {
            cs.setCode("P");
        } else {
            cs.setCode("T");
        }
        request.setProcessingCode(cs);

        //processingModeCode
        cs = factory.createCS();
        cs.setCode("T");
        request.setProcessingModeCode(cs);

        //acceptAckCode
        cs = factory.createCS();
        cs.setCode("AL");
        request.setAcceptAckCode(cs);

        //receiver
        receiver = factory.createMCCIMT000100UV01Receiver();
        request.getReceiver().add(receiver);

        //receiver/typeCode
        receiver.setTypeCode(CommunicationFunctionType.RCV);

        //receiver/device
        device = factory.createMCCIMT000100UV01Device();
        receiver.setDevice(device);

        //receiver/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //receiver/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //receiver/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension("200M");
        device.getId().add(ii);

        //sender
        sender = factory.createMCCIMT000100UV01Sender();
        request.setSender(sender);

        //sender/typeCode
        sender.setTypeCode(CommunicationFunctionType.SND);

        //sender/device
        device = factory.createMCCIMT000100UV01Device();
        sender.setDevice(device);

        //sender/device/classCode
        device.setClassCode(EntityClassDevice.DEV);

        //sender/device/determinerCode
        device.setDeterminerCode("INSTANCE");

        //sender/device/id
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.4.349");
        ii.setExtension(APP_CONFIG.getMVISndDevId());
        device.getId().add(ii);

        //controlActProcess
        controlActProcess = factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcess();
        request.setControlActProcess(controlActProcess);

        //controlActProcess/classCode
        controlActProcess.setClassCode(ActClassControlAct.CACT);

        //controlActProcess/moodCode
        controlActProcess.setMoodCode(XActMoodIntentEvent.EVN);

        //controlActProcess/code
        cd = factory.createCD();
        cd.setCode("PRPA_TE201309UV02");
        controlActProcess.setCode(cd);

        //controlActProcess/codeSystem
        cd.setCodeSystem("2.16.840.1.113883.1.6");

        //controlActProcess/queryByParameter
        queryByParameter = factory.createPRPAMT201307UV02QueryByParameter();
        controlActProcess.setQueryByParameter(factory.createPRPAIN201309UV02QUQIMT021001UV01ControlActProcessQueryByParameter(queryByParameter));

        //controlActProcess/queryByParameter/queryId
        ii = factory.createII();
        ii.setRoot("2.16.840.1.113883.3.933");
        ii.setExtension("33452");
        queryByParameter.setQueryId(ii);

        //controlActProcess/queryByParameter/statusCode
        cs = factory.createCS();
        cs.setCode("new");
        queryByParameter.setStatusCode(cs);

        //controlActProcess/queryByParameter/responsePriorityCode
        cs = factory.createCS();
        cs.setCode("I");
        queryByParameter.setResponsePriorityCode(cs);

        //controlActProcess/queryByParameter/parameterList
        parameterList = factory.createPRPAMT201307UV02ParameterList();
        queryByParameter.setParameterList(parameterList);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier
        patientIdentifier = factory.createPRPAMT201307UV02PatientIdentifier();
        parameterList.getPatientIdentifier().add(patientIdentifier);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/value
        ii = factory.createII();
        status = queryBean.getStatus();
        if ("EDIPI".equalsIgnoreCase(status)) {
            // Get identifiers by DoD EDIPI
            oid = "2.16.840.1.113883.3.364";
            id = queryBean.getItemId();
            type = "NI";
            source = "200DOD";
            issuer = "USDOD";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            LOGGER.debug("Getting corresponding patient identifiers via EDIPI");
        } else if ("ICN".equalsIgnoreCase(status)) {
            // Get identifiers by VA ICN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "NI";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type);

            LOGGER.debug("Getting corresponding patient identifiers via ICN");
        } else if ("VAIEN".equalsIgnoreCase(status)) {
            // Get identifiers by VA IEN
            oid = "2.16.840.1.113883.4.349";
            id = queryBean.getItemId();
            type = "PI";
            source = queryBean.getRecordSiteCode();
            issuer = "USVHA";

            ii.setRoot(oid);
            ii.setExtension(id + "^" + type + "^" + source + "^" + issuer);

            LOGGER.debug("Getting corresponding patient identifiers via VA IEN");
        }
//        else
//        {
//            logger.error("Could not query with identifier number type: " + status);
//        }
        patientIdentifier.getValue().add(ii);

        //controlActProcess/queryByParameter/parameterList/patientIdentifier/semanticsText
        patientIdentifier.setSemanticsText("Patient.id");

        return request;
    }

    private List<String> parseResponseToPatientIds(PRPAIN201310UV02 response) throws JMeadowsException
    {
        Iterator<MCCIMT000300UV01Acknowledgement> acknowledgementIterator;
        MCCIMT000300UV01Acknowledgement acknowledgement;
        Iterator<MCCIMT000300UV01AcknowledgementDetail> acknowledgementDetailIterator;
        MCCIMT000300UV01AcknowledgementDetail acknowledgementDetail;
        Iterator<PRPAIN201310UV02MFMIMT700711UV01Subject1> subjectIterator;
        PRPAIN201310UV02MFMIMT700711UV01Subject1 subject;
        PRPAMT201304UV02Patient patient;
        Iterator<II> iiIterator;
        II ii;

        List<String> patientIds = new ArrayList<String>();
        String errorMessage = "No Acknowledgement detail was provided in MVI repsonse.";

        // Check for errors
        acknowledgementIterator = response.getAcknowledgement().iterator();
        while (acknowledgementIterator.hasNext()) {
            acknowledgement = acknowledgementIterator.next();
            if (!"AA".equalsIgnoreCase(acknowledgement.getTypeCode().getCode())) {
                acknowledgementDetailIterator = acknowledgement.getAcknowledgementDetail().iterator();
                while (acknowledgementDetailIterator.hasNext()) {
                    acknowledgementDetail = acknowledgementDetailIterator.next();
                    errorMessage = acknowledgementDetail.getText();
                    break;
                }
                //melcolm todo: error logs are reporting: "Correlation Does Not Exist:IQI"  filter these out from error log
                //return null?
                throw new JMeadowsException("MVI Application Error: " + errorMessage);
            }
        }

        // Parse out patient identifiers
        subjectIterator = response.getControlActProcess().getSubject().iterator();
        while (subjectIterator.hasNext()) {
            subject = subjectIterator.next();
            patient = subject.getRegistrationEvent().getSubject1().getPatient();

            iiIterator = patient.getId().iterator();
            while (iiIterator.hasNext()) {
                ii = iiIterator.next();
                patientIds.add(ii.getExtension());
            }
        }

        return patientIds;
    }

    private void logAndAuditException(Throwable e, String auditMsg, JMeadowsQuery qb, boolean isAudit)
    {
        AuditDao auditDao = new AuditDao();
//        StringWriter sw = new StringWriter();
//        PrintWriter pw = new PrintWriter(sw);

        // Log error
        if (e != null) {
//            e.printStackTrace(pw);
//            logger.error(sw.toString());
            LOGGER.error(e.getMessage(), e);
        }

        // Audit error
        if (isAudit) {
            String requestingApp = qb.getRequestingApp();
            String siteCode = null;
            String userIen = null;
            String userNPI = null;
            String userName = null;
            String patId;
            User user = qb.getUser();

            if (user != null) {
                siteCode = user.getHostSite().getSiteCode();
                userIen = user.getUserIen();

                if (StringUtils.isNotBlank(user.getProviderIen())) {
                    userIen += "^" + user.getProviderIen();
                }

                userNPI = user.getNPI();
                userName = user.getName();
            }

            patId = qb.getItemId();
            if (StringUtils.isNotBlank(qb.getRecordSiteCode())) {
                patId += "^" + qb.getRecordSiteCode();
            }

            try {
                auditDao.setAudit(siteCode, userIen, userNPI, userName, patId, auditMsg, requestingApp, null, null);
            } catch (JMeadowsException exception) {
//                exception.printStackTrace(pw);
//                logger.error(sw.toString());
                if (e != null) {
                    LOGGER.error(e.getMessage(), e);
                }
            }
        }
    }

    private List<ExternalID> getVLERSiteNames(List<ExternalID> ids) {
        for (ExternalID id : ids) {
            if (id.getSource() != null) {
                Code code = VLERSiteCache.getCode("", id.getSource(), "");
                if (code != null && code.getDisplay() != null) {
                    id.setSourceName(code.getDisplay());
                }
            }
        }
        return ids;
    }
}
