/*
 * Janus 4.0 (c)
 * Copyright (c) 2013 Hawaii Resource Group LLC. All Rights Reserved.
 * Developed for the Pacific Telehealth & Technology Hui and the Pacific Joint Information Technology Center
 * Contributors:
 *      Honorable Senator Daniel K. Inouye
 *      VA Pacific Islands Health Care System
 *      Tripler Army Medical Center
 */

package gov.va.med.jmeadows.site;


import gov.va.med.jmeadows.common.JMeadowsException;
import gov.va.med.vds.webservice.Endpoint;
import gov.va.med.vds.webservice.JanusPermission;
import gov.va.med.vds.webservice.Region;
import gov.va.med.vds.webservice.Site;
import net.sf.ehcache.constructs.blocking.CacheEntryFactory;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.sql.DataSource;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class SiteCacheEntryFactory implements CacheEntryFactory {
    private static final Logger LOGGER = LoggerFactory.getLogger(Sites.class);

    private DataSource dataSource;
    private String siteCacheEntryKey;
    private String rpcEndpointEnvironment;
    private String cacheEndpointEnvironment;
    private String shareEndpointEnvironment;
    private String medwebEndpointEnvironment;
    private String ahltaEndpointEnvironment;

    public SiteCacheEntryFactory(DataSource dataSource, String siteCacheEntryKey,
                                 String rpcEndpointEnvironment, String cacheEndpointEnvironment,
                                 String shareEndpointEnvironment, String medwebEndpointEnvironment,
                                 String ahltaEndpointEnvironment) {
        this.dataSource = dataSource;
        this.siteCacheEntryKey = siteCacheEntryKey;
        this.rpcEndpointEnvironment = rpcEndpointEnvironment;
        this.cacheEndpointEnvironment = cacheEndpointEnvironment;
        this.shareEndpointEnvironment = shareEndpointEnvironment;
        this.medwebEndpointEnvironment = medwebEndpointEnvironment;
        this.ahltaEndpointEnvironment = ahltaEndpointEnvironment;
    }

    /**
     *  Creates the cacheEntry for the given cache key.
     */
    public Object createEntry(Object key) throws RuntimeException {
        Object rtc = null;
        if (siteCacheEntryKey.equals(key)) {
            rtc = getSiteMapFromDB();
        }
        return rtc;
    }

    /**
     * Generates a site map from Site database.
     *
     * @return A site map that is derived from Site database.
     */
    private Map<String, Site> getSiteMapFromDB() {
        try {
            Map<String, List<Endpoint>> endpointMap = getEndpointsFromDB();
            Map<Integer, Region> regionMap = getRegionsFromDB();
            Map<String, Site> siteMap = getSiteMapFromDB(endpointMap, regionMap);

            return siteMap;

        } catch (JMeadowsException e) {
            throw new RuntimeException(e);
        }
    }


    /**
     * Generates a site map from Site database.
     *
     * @param endpointMap site endpoints map
     * @param regionMap site region map
     * @return A site map that is derived from Site database.
     * @throws JMeadowsException if an error occurs.
     */
    private Map<String, Site> getSiteMapFromDB(Map<String, List<Endpoint>> endpointMap, Map<Integer, Region> regionMap) throws JMeadowsException {

        Map<String, Site> siteMap = new LinkedHashMap<String, Site>();

        Connection sqlConn = null;

        CallableStatement cstmt = null;
        ResultSet rset = null;

        try {
            sqlConn = dataSource.getConnection();
            cstmt = sqlConn.prepareCall("{call getSites()}");
            rset = cstmt.executeQuery();
            while (rset.next()) {
                Site site = new Site();
                site.setId(rset.getInt("id"));
                site.setSiteCode(rset.getString("sitecode"));
                site.setMoniker(rset.getString("moniker"));
                site.setName(rset.getString("name"));
                site.setAgency(rset.getString("agency"));
                site.setDmisId(rset.getString("DMISID"));
                site.setStatus(rset.getString("status"));
                site.setContainer(rset.getString("container"));

                //set endpoints and regions

                if (endpointMap != null && !endpointMap.isEmpty()) {
                    List<Endpoint> endpoints = site.getEndpoints();
                    List<Endpoint> siteEndpoints = endpointMap.get(site.getSiteCode());
                    if (siteEndpoints != null) {
                        endpoints.addAll(siteEndpoints);
                    }
                }

                if (regionMap != null && !regionMap.isEmpty()) {
                    String regionId = rset.getString("regionId");
                    if (regionId != null && StringUtils.isNotBlank(regionId)) {
                        Region region = regionMap.get(Integer.valueOf(regionId));
                        if (region != null) {
                            site.setRegion(region);
                        }
                    }
                }

                List<JanusPermission> sitePermissions = getSitePermissions(site.getId());
                if (sitePermissions != null) {
                    site.getPermissions().addAll(sitePermissions);
                }

                siteMap.put(site.getSiteCode(), site);
            }

            return siteMap;

        } catch (SQLException e) {
            LOGGER.error("get Sites exception: " + e.getMessage());

            throw new JMeadowsException(e);
        } finally {
            try {
                if (rset != null) {
                    rset.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    private Map<String, List<Endpoint>> getEndpointsFromDB() throws JMeadowsException {

        Map<String, List<Endpoint>> endpointMap = new LinkedHashMap<String, List<Endpoint>>();

        Connection sqlConn = null;

        CallableStatement cstmt = null;
        ResultSet rset = null;

        try {
            sqlConn = dataSource.getConnection();
            cstmt = sqlConn.prepareCall("{call getEndpoints(?,?,?,?,?)}");
            cstmt.setString(1, rpcEndpointEnvironment);
            cstmt.setString(2, cacheEndpointEnvironment);
            cstmt.setString(3, shareEndpointEnvironment);
            cstmt.setString(4, medwebEndpointEnvironment);
            cstmt.setString(5, ahltaEndpointEnvironment);

            rset = cstmt.executeQuery();
            while (rset.next()) {
                int id = rset.getInt("id");
                String protocol = rset.getString("protocol");
                String host = rset.getString("host");
                String env = rset.getString("env");
                String siteCode = rset.getString("sitecode");
                int port = rset.getInt("port");
                int siteId = rset.getInt("siteId");
                String modality = rset.getString("modality");
                String username = rset.getString("uname");
                String password = rset.getString("pword");
                String status = rset.getString("status");
                String aeTitle = rset.getString("aetitle");
                String accessionPrefix = rset.getString("accessionprefix");

                Endpoint endpoint = new Endpoint();
                if (id != 0) {
                    endpoint.setId(id);
                }
                if (siteId != 0) {
                    endpoint.setSiteId(siteId);
                }
                endpoint.setModality(modality);
                endpoint.setSiteCode(siteCode);
                endpoint.setEnvironment(env);
                endpoint.setHost(host);
                if (port != 0) {
                    endpoint.setPort(port);
                }
                endpoint.setProtocol(protocol);
                endpoint.setUsername(username);
                endpoint.setPassword(password);
                endpoint.setStatus(status);
                endpoint.setAeTitle(aeTitle);
                endpoint.setAccessionPrefix(accessionPrefix);

                List<Endpoint> endpoints = endpointMap.get(endpoint.getSiteCode());
                if (endpoints == null) {
                    endpoints = new ArrayList<Endpoint>();
                }

                endpoints.add(endpoint);
                endpointMap.put(endpoint.getSiteCode(), endpoints);
            }

            return endpointMap;

        } catch (SQLException e) {
            LOGGER.error("get Endpoints exception: " + e.getMessage());

            throw new JMeadowsException(e);
        } finally {
            try {
                if (rset != null) {
                    rset.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    private Map<Integer, Region> getRegionsFromDB() throws JMeadowsException {

        Map<Integer, Region> regionMap = new LinkedHashMap<Integer, Region>();

        Connection sqlConn = null;

        CallableStatement cstmt = null;
        ResultSet rset = null;

        try {
            sqlConn = dataSource.getConnection();
            cstmt = sqlConn.prepareCall("{call getRegions()}");

            rset = cstmt.executeQuery();
            while (rset.next()) {
                int id = rset.getInt("id");
                String name = rset.getString("name");


                Region region = new Region();
                if (id != 0) {
                    region.setId(id);
                }
                region.setName(name);

                regionMap.put(region.getId(), region);
            }

            return regionMap;

        } catch (SQLException e) {
            LOGGER.error("get Regions exception: " + e.getMessage());

            throw new JMeadowsException(e);
        } finally {
            try {
                if (rset != null) {
                    rset.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

    private List<JanusPermission> getSitePermissions(int siteId) throws JMeadowsException {

        List<JanusPermission> permissionList = new ArrayList<JanusPermission>();

        Connection sqlConn = null;

        CallableStatement cstmt = null;
        ResultSet rset = null;

        try {
            sqlConn = dataSource.getConnection();
            cstmt = sqlConn.prepareCall("{call getSitePermissions(?)}");
            cstmt.setInt(1, siteId);

            rset = cstmt.executeQuery();
            while (rset.next()) {
                int id = rset.getInt("permid");
                String description = rset.getString("description");


                JanusPermission permission = new JanusPermission();
                permission.setId(Integer.toString(id));
                permission.setDescription(description);

                permissionList.add(permission);
            }

            return permissionList;

        } catch (SQLException e) {
            LOGGER.error("get site permissions exception: " + e.getMessage());

            throw new JMeadowsException(e);
        } finally {
            try {
                if (rset != null) {
                    rset.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (cstmt != null) {
                    cstmt.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }

            try {
                if (sqlConn != null) {
                    sqlConn.close();
                }
            } catch (SQLException e) {
                LOGGER.error(e.getMessage());
            }
        }
    }

}
